/*
 * AbstractDataCollection.java
 */

package de.pdv_systeme.DataCollection;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.jfree.data.time.Day;
import org.jfree.data.time.Millisecond;
import org.jfree.data.time.Minute;
import org.jfree.data.time.RegularTimePeriod;
import org.jfree.data.time.Second;
import org.jfree.util.StringUtils;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesDataItem;

import de.pdv_systeme.util.MyCollections;
import de.pdv_systeme.util.VmsDate;

/**
 * An abstract base class for data collections.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public abstract class AbstractDataCollection implements DataCollection {
    
    /** This architecture's newline character */
    static final String NEWLINE = StringUtils.getLineSeparator();
    
    /** Nested class for the internal storage of a time series of values */
    protected static class DataCollectionTimeSeries extends TimeSeries {
        
        /**
         * Precision of the columns' timesSet.
         * @see #timeBaseClass
         * @see de.pdv_systeme.util.VmsDate
         */
        public static final int PRECISION = VmsDate.PRECISION_SECOND;
        
        /**
         * Base class for the columns' timesSet.
         * @see #PRECISION
         * @see org.jfree.data.time
         */
        private static Class TimeBaseClass;
        
        /** Flag: Is it a primary series? */
        private boolean isPrimary;
        
        /** Minimum value in the list */
        private transient double minValue = Double.NaN;
        
        /** Maximum value in the list */
        private transient double maxValue = Double.NaN;
        
        /** Median value of the list */
        private transient double medianValue = Double.NaN;
        
        /** Item count when min/max value were last calculated */
        private transient int minMaxItemCount = 0;
        
        /** Item count when median value were last calculated */
        private transient int medianItemCount = 0;
        
        /**
         * Get the data collection's time base class
         * based on the {@link #PRECISION}
         * @return the time base class
         * @see org.jfree.data.time
         */
        public static Class timeBaseClass() {
            if (TimeBaseClass == null) {
                if (PRECISION == VmsDate.PRECISION_DATE)
                    TimeBaseClass = Day.class;
                else if (PRECISION == VmsDate.PRECISION_MINUTE)
                    TimeBaseClass = Minute.class;
                else if (PRECISION == VmsDate.PRECISION_SECOND)
                    TimeBaseClass = Second.class;
                else if (PRECISION == VmsDate.PRECISION_MILLISECOND)
                    TimeBaseClass = Millisecond.class;
                else
                    throw new RuntimeException("Invalid PRECISION value");
            }
            return TimeBaseClass;
        }
        
        /**
         * Constructor
         * @param name  the time series name
         */
        DataCollectionTimeSeries(final String name) {
            super(name, timeBaseClass());
            this.isPrimary = true;
        }
        
        /**
         * Constructor
         * @param name  the time series name
         * @param isPrimary flag: is it a primary series?
         */
        DataCollectionTimeSeries(final String name, final boolean isPrimary) {
            super(name, timeBaseClass());
            this.isPrimary = isPrimary;
        }
        
        /**
         * See {@link java.util.ArrayList#ensureCapacity}
         * @param minCapacity the desired minimum capacity
         */
        void ensureCapacity(int minCapacity) {
            //!!! This relies on TimeSeries' implementaion of data as an ArrayList !!!
            ((ArrayList) data).ensureCapacity(minCapacity);
        }
        
        /**
         * Add a value to a time series.
         * @param timestamp  the timestamp
         * @param value  a Double value
         */
        public void addValue(final RegularTimePeriod timestamp, final double value) {
            add(new TimeSeriesDataItem(timestamp, new Double(value)), false);
        }
        
        /**
         * Add a value to a time series.
         * @param timestamp  the timestamp
         * @param value  a Double value
         */
        public void addValue(final RegularTimePeriod timestamp, final Double value) {
            add(new TimeSeriesDataItem(timestamp, value), false);
        }
        
        /**
         * Get smallest value in the collection
         * @return the smallest value (or NaN if the collection is empty)
         */
        public double minimum() {
            initializeMinMax();
            return minValue;
        }
        
        /**
         * Get biggest value in the collection
         * @return the biggest value (or NaN if the collection is empty)
         */
        public double maximum() {
            initializeMinMax();
            return maxValue;
        }
        
        /**
         * Get the median of the collection
         * @return the median (or NaN if the collection is empty)
         */
        public double median() {
            int itemCount = getItemCount();
            if (itemCount == 0)
                return Double.NaN;
            
            if (Double.isNaN(medianValue) || medianItemCount != itemCount) {
                if (minimum() == maximum()) {
                    medianValue = minimum();
                } else {
                    // Get collection of values
                    List values = new ArrayList(itemCount);
                    for (Iterator di = data.iterator(); di.hasNext(); )
                        values.add(((TimeSeriesDataItem)di.next()).getValue());
                    if (System.getProperty("JTLviz.withMedian") != null)
                        medianValue = ((Double) MyCollections.median(values)).doubleValue();
                }
                medianItemCount = itemCount;
            }
            return medianValue;
        }
        
        /** Initialize minValue/maxValue */
        private void initializeMinMax() {
            int itemCount = getItemCount();
            if (itemCount == 0)
                return;
            
            if (Double.isNaN(minValue)
            || Double.isNaN(maxValue)
            || minMaxItemCount != itemCount) {
                Iterator di = data.iterator();
                double value = ((TimeSeriesDataItem)di.next()).getValue().doubleValue();
                minValue = maxValue = value;
                while (di.hasNext()) {
                    value = ((TimeSeriesDataItem)di.next()).getValue().doubleValue();
                    if (value < minValue) minValue = value;
                    if (value > maxValue) maxValue = value;
                }
                minMaxItemCount = itemCount;
            }
        }
    }
    
    /** The data collections' timestamp precision */
    public static final int PRECISION = DataCollectionTimeSeries.PRECISION;
    
    /** The class' format name (to be overwritten by subclasses) */
    protected String format;
    
    /** Nodename of the collecting system */
    protected String nodename;
    
    /** Storage for the data collection description */
    protected String[] info;
    
    /** The data collection (Map of DataCollectionTimeSeries, key is the item name) */
    private Map timeSeriesMap;
    
    /** The data collection item names (original order) */
    private List timeSeriesList;
    
    /** Creates a new instance of AbstractDataCollection */
    protected AbstractDataCollection() {
        format = "Abstract";
        initializeInstance();
    }
    
    /**
     * Write the data collection to a TLC formatted file.
     * @param file the file to write
     * @throws java.io.IOException
     */
    public void writeData(File file)
    throws IOException {
        writeData(file, timeSeriesMap.keySet());
    }
    
    /**
     * Write a subset of the data collection to a TLC formatted file.
     * @param file the file to write
     * @param items the items to write
     * @throws java.io.IOException
     */
    public void writeData(File file, Set items)
    throws IOException {
        FileWriter out = new FileWriter(file);
        writeDataImpl(out, items);
        out.close();
    }
    
    /*
     * Implementation of the DataCollection interface
     */
    /** {@inheritDoc} */
    public String getFormat() {
        return format;
    }
    
    /** {@inheritDoc} */
    public String getNodename() {
        return nodename;
    }
    
    /** {@inheritDoc} */
    public String[] getInfo() {
        if (info == null) {
            return null;
        } else {
            // return a copy of info
            String[] returnInfo = new String[info.length];
            System.arraycopy(info, 0, returnInfo, 0, info.length);
            return returnInfo;
        }
    }
    
    /** {@inheritDoc} */
    public List getAllItemNames() {
        return new ArrayList(timeSeriesMap.keySet());
    }
    
    /** {@inheritDoc} */
    public int getItemCount() {
        return timeSeriesMap.size();
    }
    
    /** {@inheritDoc} */
    public boolean contains(final String name) {
        return timeSeriesMap.containsKey(name);
    }
    
    /** {@inheritDoc} */
    public int getIndex(final String name) {
        if (! contains(name))
            throw new NoSuchElementException(name);
        return timeSeriesList.indexOf(name);
    }
    
    /** {@inheritDoc} */
    public String getName(final int index) {
        if (index < 0 || index >= getItemCount())
            throw new IndexOutOfBoundsException(String.valueOf(index));
        return (String) timeSeriesList.get(index);
    }
    
    /** {@inheritDoc} */
    public String getMedian(final String name) {
        if (! contains(name))
            throw new NoSuchElementException(name);
        return String.valueOf(getTimeSeries(name).median());
    }
    
    /** {@inheritDoc} */
    public int getRecordCount(final String name) {
        return getTimeSeries(name).getItemCount();
    }
    
    /** {@inheritDoc} */
    public boolean isPrimary(String name) {
        return getTimeSeries(name).isPrimary;
    }
    
    /** {@inheritDoc} */
    public TimeSeries getData(final String name) {
        return getTimeSeries(name);
    }
    
    /** {@inheritDoc} */
    public double getDataMin(final String name) {
        return getTimeSeries(name).minimum();
    }
    
    /** {@inheritDoc} */
    public double getDataMax(final String name) {
        return getTimeSeries(name).maximum();
    }
    
    /** {@inheritDoc} */
    public String getDescription(final String name) {
        if (! contains(name))
            throw new NoSuchElementException(name);
        return name;
    }
    
    /**
     * Calculate the correlation coefficient between two data series.
     * <p>
     * Given x1,...,xn and y1,...,yn (where equal indices denote values
     * with the same timestamp), with
     * Sx = SUM(xi), Sy = SUM(yi), Sx2 = SUM(xi**2), Sy2 = SUM(yi**2), Sxy = SUM(xi*yi),
     * it is calculated as
     * r = (n*Sxy-Sx*Sy)/SQRT((n*Sx2-Sx**2)*(n*Sy2-Sy**2)) .
     *
     * @param name1  item name #1
     * @param name2  item name #2
     * @return the correlation (between -1 and 1), or {@link Double#NaN}, if n or the denominator is 0
     * @throws NoSuchElementException thrown if there's no data series of either name
     */
    public double correlation(final String name1, final String name2) {
        TimeSeries ts1 = getTimeSeries(name1);
        TimeSeries ts2 = getTimeSeries(name2);
        TimeSeries refTs = (getRecordCount(name1) < getRecordCount(name2)) ? ts1 : ts2;
        if (refTs.getItemCount() == 0) {
            return Double.NaN;
        }
        double  sum1 = 0.0,
                sum2 = 0.0,
                sum12 = 0.0,
                sqsum1 = 0.0,
                sqsum2 = 0.0;
        int n = 0;
        for (int i = 0; i < refTs.getItemCount(); i++) {
            RegularTimePeriod time = refTs.getTimePeriod(i);
            Number  num1 = ts1.getValue(time),
                    num2 = ts2.getValue(time);
            if (num1 != null && num2 != null) {
                double  d1 = num1.doubleValue(),
                        d2 = num2.doubleValue();
                sum1 += d1;
                sum2 += d2;
                sum12 += d1 * d2;
                sqsum1 += d1 * d1;
                sqsum2 += d2 * d2;
                n++;
            }
        }
        double sq = (n*sqsum1 - sum1*sum1)*(n*sqsum2 - sum2*sum2);
        return (sq <= 0.0) ? Double.NaN : (n*sum12 - sum1*sum2) / Math.sqrt(sq);
    }
    
    /** {@inheritDoc} */
    public void addCalculatedSeries(
            final String name,
            final Operation op,
            final TimeSeries ts1,
            final TimeSeries ts2
            ) {
        DataCollectionTimeSeries newTs = new DataCollectionTimeSeries(name, false);
        TimeSeries refTs = (ts1.getItemCount() < ts2.getItemCount()) ? ts1 : ts2;
        for (int i = 0; i < refTs.getItemCount(); i++) {
            RegularTimePeriod time = refTs.getTimePeriod(i);
            Number
                    num1 = ts1.getValue(time),
                    num2 = ts2.getValue(time);
            if (num1 == null || num2 == null)
                newTs.addValue(time, (Double) null);
            else
                newTs.addValue(time, op.execute(num1.doubleValue(), num2.doubleValue()));
        }
        addTimeSeries(newTs);
    }
    
    /** {@inheritDoc} */
    public void addCalculatedSeries(
            final String name,
            final Operation op,
            final TimeSeries ts1,
            final double num2
            ) {
        DataCollectionTimeSeries newTs = new DataCollectionTimeSeries(name, false);
        TimeSeries refTs = ts1;
        for (int i = 0; i < refTs.getItemCount(); i++) {
            RegularTimePeriod time = refTs.getTimePeriod(i);
            Number num1 = ts1.getValue(time);
            if (num1 == null)
                newTs.addValue(time, (Double) null);
            else
                newTs.addValue(time, op.execute(num1.doubleValue(), num2));
        }
        addTimeSeries(newTs);
    }
    
    /** {@inheritDoc} */
    public void addCalculatedSeries(
            final String name,
            final Operation op,
            final double num1,
            final TimeSeries ts2
            ) {
        DataCollectionTimeSeries newTs = new DataCollectionTimeSeries(name, false);
        TimeSeries refTs = ts2;
        for (int i = 0; i < refTs.getItemCount(); i++) {
            RegularTimePeriod time = refTs.getTimePeriod(i);
            Number num2 = ts2.getValue(time);
            if (num2 == null)
                newTs.addValue(time, (Double) null);
            else
                newTs.addValue(time, op.execute(num1, num2.doubleValue()));
        }
        addTimeSeries(newTs);
    }
    
    /** {@inheritDoc} */
    public void addCalculatedSeries(
            final String name,
            final Operation op,
            final double num1,
            final double num2
            ) {
        DataCollectionTimeSeries newTs = new DataCollectionTimeSeries(name, false);
        if (getItemCount() != 0) {
            TimeSeries refTs = getData(getName(0));
            Double value = op.execute(num1, num2);
            for (int i = 0; i < refTs.getItemCount(); i++) {
                RegularTimePeriod time = refTs.getTimePeriod(i);
                newTs.addValue(time, value);
            }
        }
        addTimeSeries(newTs);
    }
    
    /**
     * Initialize the instance variables
     */
    protected void initializeInstance() {
        timeSeriesMap = new TreeMap();
        timeSeriesList = new ArrayList();
        nodename = null;
        info = null;
    }
    
    /**
     * Add a time series to the collection.
     * @param timeSeries  the time series
     */
    protected void addTimeSeries(TimeSeries timeSeries) {
        String name = (String) timeSeries.getKey();
        timeSeriesMap.put(name, timeSeries);
        timeSeriesList.add(name);
    }
    
    /**
     * Get a time series.
     * @param name  the time series' name
     * @return the time series
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    protected DataCollectionTimeSeries getTimeSeries(final String name) {
        if (! timeSeriesMap.containsKey(name))
            throw new NoSuchElementException(name);
        return (DataCollectionTimeSeries) timeSeriesMap.get(name);
    }
    
    /**
     * Write data in TLC format.
     * @param out FileWriter
     * @param items the items to write
     * @throws java.io.IOException
     */
    private void writeDataImpl(FileWriter out, Set items)
    throws IOException {
        // Write header line
        out.write(getNodename());
        String[] info = getInfo();
        for (int i = 0; i < info.length; i++) {
            out.write(",");
            out.write(info[i]);
        }
        out.write(NEWLINE);
        
        // Write empty start date and time lines
        out.write(NEWLINE);
        out.write(NEWLINE);
        
        // Write line of item names
        out.write("Sample Time");
        for (Iterator si = items.iterator(); si.hasNext(); ) {
            String name = (String) si.next();
            out.write(",");
            out.write(name);
        }
        out.write(NEWLINE);
        
        // To write the data lines, first gather timestamps from all items
        // into an ordered list with no duplicates
        SortedSet timesSet = new TreeSet(); // of RegularTimePeriod
        for (Iterator si = items.iterator(); si.hasNext(); ) {
            String name = (String) si.next();
            TimeSeries ts = getTimeSeries(name);
            for (int i = 0; i < ts.getItemCount(); i++)
                timesSet.add(ts.getDataItem(i).getPeriod());
        }
        
        // Then, write the data line: time stamp and values for all timestamps gathered
        for (Iterator ti = timesSet.iterator(); ti.hasNext(); ) {
            // The timestamps are really end times of the collection intervals,
            // but have been loaded as start times of the RegularTimePeriods.
            RegularTimePeriod period = (RegularTimePeriod) ti.next();
            out.write(VmsDate.toString(period.getStart(), PRECISION));
            for (Iterator si = items.iterator(); si.hasNext(); ) {
                String name = (String) si.next();
                TimeSeries ts = getTimeSeries(name);
                out.write(",");
                Number value = ts.getValue(period);
                if (value != null)
                    out.write(value.toString());
            }
            out.write(NEWLINE);
        }
    }
}
