/*
 * DataCollection.java
 */

package de.pdv_systeme.DataCollection;

import java.util.List;

import org.jfree.data.time.TimeSeries;

/**
 * Interface for data collections
 *
 * @author Martin Vorlnder <mv@pdv-systeme.de>
 */
public interface DataCollection {
    
    /**
     * Return a string identifyng the data type.
     * @return the data type
     */
    String getFormat();
    
    /**
     * Get the collector node name.
     * @return the node name
     */
    String getNodename();
    
    /**
     * Get any info the collection has to offer.
     * @return the info
     */
    String[] getInfo();
    
    /**
     * Get the number of items in the collection.
     * @return the number of items
     */
    int getItemCount();
    
    /**
     * Get all item names.
     * @return all item names
     */
    List getAllItemNames();
    
    /**
     * Test whether the specified item is part of the collection.
     * @param name The item name
     * @return <code>true</code> if a item called <code>name</code>
     * exists in the collection, <code>false</code> otherwise.
     */
    boolean contains(String name);
    
    /**
     * Get the original index of a data series.
     * @param name the data series' name
     * @return the index
     * @throws NoSuchElementException if there's no data series of that name
     */
    public int getIndex(final String name);
    
    /**
     * Get the name for an index.
     * @param index the index
     * @return the name
     * @throws IndexOutOfBoundsException
     */
    public String getName(final int index);
    
    /**
     * Get the median of a data series.
     * @param name the data series' name
     * @return the median
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    public String getMedian(final String name);

    /**
     * Get the number of data records in the specified data series.
     * @param name  The item name
     * @return the number of data records
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    int getRecordCount(String name);
    
    /**
     * Is the specified data series a primary series?
     * @param name  The item name
     * @return the isPrimary attribute
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    boolean isPrimary(String name);
    
    /**
     * Get one data series by item name.
     * @param name  The item name
     * @return the data series
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    TimeSeries getData(String name);
    
    /**
     * Get the minimum value of one data series by column name.
     * @param name  The column name
     * @return the minimum value
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    double getDataMin(String name);
    
    /**
     * Get the maximum value of one data series by item name.
     * @param name  The item name
     * @return the maximum value
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    double getDataMax(String name);
    
    /**
     * Get a descriptive string of one item.
     * @param name  The item name
     * @return the descriptive string
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    String getDescription(String name);
    
    /**
     * Calculate the correlation coefficient between two data series.
     * @param name1  item name #1
     * @param name2  item name #2
     * @return the correlation (between -1 and 1, or {@link Double#NaN})
     * @throws NoSuchElementException thrown if there's no data series of that name
     */
    double correlation(String name1, String name2);
    
    /**
     * Add a calculated series.
     * @param name  item name of the calculated series
     * @param op  operation to perform
     * @param ts1  time series #1
     * @param ts2  time series #2
     * @throws IllegalArgumentException if <code>op</code> isn't a valid operation
     */
    void addCalculatedSeries(String name, Operation op, TimeSeries ts1, TimeSeries ts2);
    
    /**
     * Add a calculated series.
     * @param name  item name of the calculated series
     * @param op  operation to perform
     * @param ts1  time series #1
     * @param num2  double constant
     * @throws IllegalArgumentException if <code>op</code> isn't a valid operation
     */
    void addCalculatedSeries(String name, Operation op, TimeSeries ts1, double num2);
    
    /**
     * Add a calculated series.
     * @param name  item name of the calculated series
     * @param op  operation to perform
     * @param num1  double constant
     * @param ts2  time series #2
     * @throws IllegalArgumentException if <code>op</code> isn't a valid operation
     */
    void addCalculatedSeries(String name, Operation op, double num1, TimeSeries ts2);
    
    /**
     * Add a calculated series.
     * @param name  item name of the calculated series
     * @param op  operation to perform
     * @param num1  double constant
     * @param num2  double constant
     * @throws IllegalArgumentException if <code>op</code> isn't a valid operation
     */
    void addCalculatedSeries(String name, Operation op, double num1, double num2);
    
}
