/*
 * FileDataCollection.java
 */

package de.pdv_systeme.DataCollection;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.LineNumberReader;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.jfree.data.time.RegularTimePeriod;
import org.jfree.data.time.Second;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesDataItem;

import de.pdv_systeme.util.VmsDate;

/**
 * Base class for data collections that are based on files
 *
 * @author Martin Vorlnder <mv@pdv-systeme.de>
 */
public abstract class FileDataCollection extends AbstractDataCollection {
    
    /** The file containing the data */
    protected File file = null;
    
    /** The number of lines in the file */
    protected int lineCount;
    
    /** Flag: is the load to be interrupted */
    protected boolean isInterrupted;
    
    /** The input stream */
    private LineNumberReader in;
    
    /** Flag: has the file been loaded? */
    private boolean isFileRead;
    
    /**
     * Creates a new instance of FileDataCollection
     */
    protected FileDataCollection() {
        super();
        format = "File";
        file = null;
        lineCount = -1;
        in = null;
        isFileRead = false;
    }
    
    /**
     * Create a new instance of FileDataCollection.
     * As <code>FileDataCollection</code> is an abstract class,
     * the constructor is protected.
     * 
     * @param file File object representing the data collection file
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     * @throws java.text.ParseException 
     * @throws java.lang.NumberFormatException 
     * @throws java.lang.IllegalArgumentException 
     */
    protected FileDataCollection(final File file)
    throws FileNotFoundException,
            IOException,
            ParseException,
            NumberFormatException,
            IllegalArgumentException {
        super();
        format = "File";
        this.file = file;
        lineCount = -1;
        in = null;
        isFileRead = false;
        initialize();
    }
    
    /**
     * Test the file format
     * <p>
     * This method must be overloaded by implementations.
     * 
     * @return true iff the file has the correct format
     * @param file File object representing the file
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     */
    public static boolean testFormat(final File file)
    throws FileNotFoundException, IOException {
        throw new RuntimeException("testFormat has not been overloaded");
    }
    
    /**
     * Load the file data
     * @param timeDataCollection  the data collection to provide the timestamps (may be <code>null</code>)
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     * @throws java.text.ParseException 
     * @throws java.lang.InterruptedException 
     * @throws java.lang.NumberFormatException 
     * @throws java.lang.IllegalArgumentException 
     */
    public void loadData(FileDataCollection timeDataCollection)
    throws FileNotFoundException,
            IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException {
        if (file == null)
            throw new IOException("file is null");
        
        // Make sure lineCount is initialized
        // (to faciliate implementations to reserve sufficient storage)
        initializeLinecount();
        
        isInterrupted = false;
        in = new LineNumberReader(new FileReader(file));
        readData(in, timeDataCollection);
        in.close();
        in = null;
        if (isInterrupted) {
            super.initializeInstance();
            lineCount = -1;
            isFileRead = false;
            throw new IOException("load interrupted");
        } else {
            isFileRead = true;
        }
    }
    
    /**
     * Load the file data
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     * @throws java.text.ParseException 
     * @throws java.lang.InterruptedException 
     * @throws java.lang.NumberFormatException 
     * @throws java.lang.IllegalArgumentException 
     */
    public void loadData()
    throws FileNotFoundException,
            IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException {
        loadData(null);
    }

    /**
     * Interrupt a file load in progress.
     */
    public void interruptLoad() {
        isInterrupted = true;
    }
    
    /**
     * Get the number of lines in the file.
     * @return the number of lines
     * @throws java.io.IOException 
     */
    public int getLinecount()
    throws IOException {
        if (file == null)
            throw new IOException("file is null");
        
        if (lineCount == -1)
            initializeLinecount();
        return lineCount;
    }
    
    /**
     * Get the number of lines read of a read operation in progress.
     * @return the current line count, or 0, if no read is in progress
     */
    public int getCurrentLinecount() {
        return (in != null) ? in.getLineNumber() : isFileRead ? lineCount : 0;
    }
    
    /**
     * Get the file name
     * @return  the data collection file's name
     */
    public String getFilename() {
        return file.getPath();
    }
    
    /**
     * Subclasses can override this method to do any initialization
     * before loading data.
     */
    protected void initialize() {
        // empty in the base class
    }
    
    /**
     * Read the file's data records.
     * <p>
     * This method must be overloaded by implementations.
     * 
     * @param in the input stream
     * @param timeDataCollection  the data collection to provide the timestamps (may be <code>null</code>)
     * @throws java.io.IOException 
     * @throws java.text.ParseException 
     * @throws java.lang.InterruptedException 
     * @throws java.lang.NumberFormatException 
     * @throws java.lang.IllegalArgumentException 
     */
    protected abstract void readData(
            LineNumberReader in,
            FileDataCollection timeDataCollection
            )
    throws IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException;
    
    /**
     * Initialize the line count.
     */
    private void initializeLinecount()
    throws IOException {
        lineCount = 0;
        BufferedReader in2 = new BufferedReader(new FileReader(file));
        while (in2.readLine() != null)
            lineCount++;
        in2.close();
    }
    
}
