/*
 * MemoryUsageDataCollection.java
 *
 * based on:
 * --------------------
 * MemoryUsageDemo.java
 * --------------------
 * contained in the JFreeChart demo package
 * (C) Copyright 2002-2005, by Object Refinery Limited.
 */

package de.pdv_systeme.DataCollection;

import java.util.Date;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.Timer;

import org.jfree.data.general.SeriesException;
import org.jfree.data.time.RegularTimePeriod;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesDataItem;

import de.pdv_systeme.DataCollection.AbstractDataCollection.DataCollectionTimeSeries;

/**
 * A collection that displays the current JVM memory usage.
 */
public class MemoryUsageDataCollection extends AbstractDataCollection {
    
    /** A column name */
    private static final String
            FREE_NAME  = "JVM Free Memory (MB)",
            USED_NAME  = "JVM Used Memory (MB)",
            TOTAL_NAME = "JVM Total Memory (MB)";
    
    /** The time series for free memory. */
    private TimeSeries free;
    
    /** The time series for free memory. */
    private TimeSeries used;
    
    /** The time series for total memory used. */
    private TimeSeries total;
    
    /**
     * Creates a new instance of MemoryUsageDataCollection.
     *
     * @param maxAge  the maximum age of values (in seconds)
     * @param interval  the data collection interval (in seconds)
     */
    public MemoryUsageDataCollection(int maxAge, int interval) {
        super();
        format = "Memory";
        nodename = "local";
        info = new String[] {
            "Java vendor : "+System.getProperty("java.vendor"),
            "Java version: "+System.getProperty("java.version"),
            "JVM         : "+System.getProperty("java.vm.vendor")+" "+System.getProperty("java.vm.name"),
            "JVM version : "+System.getProperty("java.vm.version")
        };

        free = new DataCollectionTimeSeries(FREE_NAME);
        free.setMaximumItemAge(1000*maxAge);
        addTimeSeries(free);
        
        used = new DataCollectionTimeSeries(USED_NAME);
        used.setMaximumItemAge(1000*maxAge);
        addTimeSeries(used);
        
        total = new DataCollectionTimeSeries(TOTAL_NAME);
        total.setMaximumItemAge(1000*maxAge);
        addTimeSeries(total);

        // start the data collection
        new DataGenerator(1000*interval).start();
    }
    
    /**
     * The data generator.
     */
    private class DataGenerator extends Timer implements ActionListener {
        
        /** Scaling factor Bytes -> Megabytes */
        private static final double SCALE = 1.0 / 1024.0 / 1024.0;
        
        /**
         * Constructor.
         *
         * @param interval  the interval (in milliseconds)
         */
        DataGenerator(int interval) {
            super(interval, null);
            addActionListener(this);
            setInitialDelay(0);
        }
        
        /**
         * Adds a new free/total memory reading to the dataset.
         *
         * @param event  the action event.
         */
        public void actionPerformed(ActionEvent event) {
            RegularTimePeriod time = RegularTimePeriod.createInstance(
                    DataCollectionTimeSeries.timeBaseClass(),
                    new Date(),
                    RegularTimePeriod.DEFAULT_TIME_ZONE
                    );
            long f = Runtime.getRuntime().freeMemory();
            long t = Runtime.getRuntime().totalMemory();
            try {
                free.add(new TimeSeriesDataItem(time, new Double(f * SCALE)), false);
                used.add(new TimeSeriesDataItem(time, new Double((t - f) * SCALE)), false);
                total.add(new TimeSeriesDataItem(time, new Double(t * SCALE)), false);
            } catch (SeriesException e) {
                // Do nothing
            }
        }
        
    }
    
}
