/*
 * TDCDataCollection.java
 */

package de.pdv_systeme.DataCollection;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.LineNumberReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.MessageFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Pattern;
import java.util.Iterator;
import java.util.NoSuchElementException;

import org.jfree.data.time.RegularTimePeriod;

import de.pdv_systeme.DataCollection.AbstractDataCollection.DataCollectionTimeSeries;
import de.pdv_systeme.util.VmsDate;

/**
 * <code>TDCDataCollection</code> reads a CSV formatted file
 * exported from Enterprise Capacity Planer (ECP) or
 * The Data Collector (TDC) V2
 * 
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class TDCDataCollection
        extends FileDataCollection {
    
    /** Description of a field in the ECP/TDC file */
    private static class FileField {
        private static final String RECORD_TYPE_STR = "Record type";
        String name;
        String description; // or value
        Class fieldClass;
        FileField(String n, String d, Class c) {
            name = n;
            description = d;
            fieldClass = c;
        }
        FileField(String d) {
            name = RECORD_TYPE_STR;
            description = d;
            fieldClass = String.class;
        }
    };
    
    /** Description of a record (line) in the ECP/TDC file */
    private static class FileRecord {
        String description;
        MessageFormat nameFormat;
        FileField[] field;
        FileRecord(String d, MessageFormat n, FileField[] f) {
            description = d;
            nameFormat = n;
            field = f;
        }
    };
    
    /** Header field of ECP/TDC file records */
    private static final FileField
            INTERVAL_START_TIME = new FileField("Interval start time", "Start time (VMS format)", Date.class),
            INTERVAL_END_TIME = new FileField("Interval end time", "End time (VMS format)", Date.class),
            HOST_NAME = new FileField("Host name", "Collecting system name", String.class),
            INTERVAL_LENGTH = new FileField("Interval length", "Time interval (seconds)", Double.class);
    
    /** Index of the header field */
    private static final int
            INTERVAL_START_TIME_INDEX = 0,
            INTERVAL_END_TIME_INDEX   = 1,
            HOST_NAME_INDEX           = 2,
            RECORD_TYPE_INDEX         = 3,
            INTERVAL_LENGTH_INDEX     = 4;
    /** Index of the last header field */
    private static final int
            END_HEADER_INDEX = INTERVAL_LENGTH_INDEX;
    
    /** Index of VMSPRC data field */
    private static final int
            VMSPRC_PID_INDEX         = END_HEADER_INDEX + 1,
            VMSPRC_PROCESSNAME_INDEX = END_HEADER_INDEX + 4,
            VMSPRC_USERNAME_INDEX    = END_HEADER_INDEX + 5,
            VMSPRC_IMAGENAME_INDEX   = END_HEADER_INDEX + 21;

    // TODO: RECORD should really be final. Are there no final objects in Java?
    /** Description of all possible ECP/TDC file records */
    private static /*final*/ FileRecord[] RECORD;

    /** Nested class to store process infrmation */
    private static class ProcInfo {
        // TODO: Store more process data, e.g. image names
        String procName;
        String userName;
        List imgName;
        ProcInfo(String nam, String user) {
            procName = nam;
            userName = user;
            imgName = new ArrayList();
        }
    }

    /**
     * Creates a new instance of TDCDataCollection 
     */
    public TDCDataCollection() {
        super();
        format = "TDC";
    }
    
    /**
     * Create a new instance of TDCDataCollection.
     * @param file File object representing the TDC file
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     * @throws java.text.ParseException 
     * @throws java.lang.NumberFormatException 
     * @throws java.lang.IllegalArgumentException 
     */
    public TDCDataCollection(final File file)
    throws FileNotFoundException,
            IOException,
            ParseException,
            NumberFormatException,
            IllegalArgumentException {
        super(file);
        format = "TDC";
    }
    
    /** {@inheritDoc} */
    protected void initialize() {
        initialize_RECORD();
    }
    
    /* {@inheritDoc} */
    /**
     * Test the file format
     * @return true iff the file has the correct format
     * @param file File object representing the file
     * @throws java.io.FileNotFoundException 
     * @throws java.io.IOException 
     */
    public static boolean testFormat(final File file)
    throws FileNotFoundException, IOException {
        boolean isTDC = false;
        if (file != null) {
            BufferedReader in = new BufferedReader(new FileReader(file));
            // 1st and 2nd field of 1st line must be valid VMS timestamps
            // 3rd and 4th field of 1st line must be strings (hostname & field name)
            String[] line = in.readLine().split(",");
            in.close();
            isTDC = VmsDate.datePattern.matcher(line[0]).matches()
            && VmsDate.datePattern.matcher(line[1]).matches()
            && Pattern.compile("^\\w+$", Pattern.CASE_INSENSITIVE).matcher(line[2]).matches()
            && Pattern.compile("^\\w+$", Pattern.CASE_INSENSITIVE).matcher(line[3]).matches();
        }
        return isTDC;
    }
    
    /** {@inheritDoc} */
    protected void readData(
            LineNumberReader in,
            FileDataCollection timeDataCollection
            )
    throws IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException {
        Map procInfoMap = new TreeMap(); // of ProcInfo
        String line;
        while ((line = in.readLine()) != null && ! isInterrupted) {
            String[] lineItem = line.split(",");
            FileRecord rec = findFileRecord(lineItem[RECORD_TYPE_INDEX]);
            if (rec == null) {
                System.err.println("Unknown record type "+lineItem[RECORD_TYPE_INDEX]+" encountered");
                continue;
            }
            if (rec.nameFormat == null) {
                continue;
            }
            // Set nodename
            if (nodename == null) {
                nodename = lineItem[HOST_NAME_INDEX];
            }
            // Get timestamp
            Date thisTime;
            try {
                thisTime = VmsDate.parse(lineItem[INTERVAL_END_TIME_INDEX]);
            }
            catch (ParseException e) {
                throw new ParseException(
                        "Line "+in.getLineNumber()+": "+
                        lineItem[INTERVAL_END_TIME_INDEX]+", "+e.getMessage(),
                        e.getErrorOffset()
                        );
            }
            RegularTimePeriod timestamp = RegularTimePeriod.createInstance(
                        DataCollectionTimeSeries.timeBaseClass(),
                        thisTime,
                        RegularTimePeriod.DEFAULT_TIME_ZONE
                        );
            // Store data values
            String baseName = rec.nameFormat.format(lineItem);
            for (int itemNum = END_HEADER_INDEX + 1; itemNum < rec.field.length; itemNum++) {
                FileField fld = rec.field[itemNum];
                if (fld.fieldClass != Double.class && fld.fieldClass != Integer.class)
                    continue;

                String name = baseName + fld.name;
                DataCollectionTimeSeries dcts;
                if (contains(name)) {
                    dcts = getTimeSeries(name);
                } else {
                    dcts = new DataCollectionTimeSeries(name);
                    addTimeSeries(dcts);
                }

                try {
                    dcts.addValue(timestamp, Double.valueOf(lineItem[itemNum]));
                }
                catch (NumberFormatException e) {
                    throw new NumberFormatException(
                            "Line "+in.getLineNumber()+", column "+itemNum+": "+
                            e.getMessage()
                            );
                }
            }

            // Special treatment of VMSPRC records
            if (lineItem[RECORD_TYPE_INDEX].equals("VMSPRC")) {
                // Store process info: PID, process name, user name, image(s)
                String pid = lineItem[VMSPRC_PID_INDEX];
                ProcInfo proc;
                if (procInfoMap.containsKey(pid)) {
                    proc = (ProcInfo) procInfoMap.get(pid);
                } else {
                    // remove quotes
                    String procName = lineItem[VMSPRC_PROCESSNAME_INDEX];
                    String userName = lineItem[VMSPRC_USERNAME_INDEX];
                    // remove quotes
                    procName = procName.substring(1, procName.length()-1);
                    userName = userName.substring(1, userName.length()-1);
                    proc = new ProcInfo(procName, userName);
                    procInfoMap.put(pid, proc);
                }
                String imgName = lineItem[VMSPRC_IMAGENAME_INDEX];
                if (imgName.length() > 2) {
                    // remove quotes
                    imgName = imgName.substring(1, imgName.length()-1);
                    if (! proc.imgName.contains(imgName))
                        proc.imgName.add(imgName);
                }
            }
        }
        
        if (isInterrupted) {
            procInfoMap = null;
            throw new InterruptedException();
        }
        
        // Fill info
        List infoList = new ArrayList();
        for (Iterator pi = procInfoMap.entrySet().iterator(); pi.hasNext(); ) {
            Map.Entry entry = (Map.Entry) pi.next();
            String pid = (String) entry.getKey();
            ProcInfo proc = (ProcInfo) entry.getValue();
            infoList.add("PID "+pid+":");
            infoList.add("  Process Name: "+proc.procName);
            if (! proc.userName.equals(""))
                infoList.add("  User Name   : "+proc.userName);
            for (Iterator ii = proc.imgName.iterator(); ii.hasNext(); )
                infoList.add("  Image Name  : "+(String)ii.next());
        }
        info = (String[]) infoList.toArray(new String[infoList.size()]);
    }
    
    /** {@inheritDoc} */
    public String getDescription(final String name) {
        if (! contains(name)) {
            throw new NoSuchElementException(name);
        }
        int delim = name.indexOf(']');
        String itemField = name.substring(delim+1);
        String itemClass = name.substring(0, delim);
        if ((delim = itemClass.indexOf('.')) != -1)
            itemClass = itemClass.substring(0, delim);
        for (int i = 0; i < RECORD.length; i++) {
            if (RECORD[i].nameFormat == null)
                continue;
            String recClass = RECORD[i].nameFormat.toPattern();
            if (recClass.startsWith(itemClass))
                for (int j = 0; j < RECORD[i].field.length; j++)
                    if (RECORD[i].field[j].name.equals(itemField))
                        return name+": "+RECORD[i].field[j].description;
        }
        return name;
    }

    /**
     * Find a FileRecord in RECORD by record type
     * @param recType  the record type
     * @return the FileRecord or <code>null</code> if not found
     */
    private FileRecord findFileRecord(final String recType) {
        for (int i = 0; i < RECORD.length; i++) {
            if (recType.equals(RECORD[i].field[RECORD_TYPE_INDEX].description)) {
                return RECORD[i];
            }
        }
        return null;
    }
    
    /**
     * Initialize RECORD
     */
    private static void initialize_RECORD() {
        RECORD = new FileRecord[] {
            /*
             * For each device recognized by the monitored OpenVMS system, there will be a
             * hardware record at the beginning of the formatted data file.
             */
            new FileRecord(
                    "Hardware Configuration",
                    null, // no message format
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSHWC"),
                        INTERVAL_LENGTH,
                        new FileField("Host controller name", "SCS name of node that device is attached. This is either an OpenVMS node or a hierarchical storage controller", String.class),
                        new FileField("Host controller type", "Designates the type of host controller. The string will be either VAX or ALPHA or IA64", String.class),
                        new FileField("Device name", "If part of an allocation class, the name will be prefixed with the allocation class designation.", String.class),
                        new FileField("Device class code", "Device class code specified in the device UCB field UCB$B_DEVCLASS", Integer.class),
                        new FileField("Device type code", "Device type code specified in the device UCB field UCB$B_DEVTYPE", Integer.class)
            }
            ),
            /*
             * For each interval and each processor there is a CPU statistics data record.
             */
                    new FileRecord(
                    "CPU Statistics",
                    new MessageFormat("[CPU.{5}]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSCPU"),
                        INTERVAL_LENGTH,
                        new FileField("CPU id number", "Hardware ID number of the CPU board", Integer.class),
                        new FileField("Total utilization", "Percentage of CPU time spent outside the idle loop", Double.class),
                        new FileField("Average CPU queue length", "Average number of processes in the COM scheduling state", Double.class),
                        new FileField("User mode utilization", "Percentage of time CPU spent in user mode", Double.class),
                        new FileField("Supervisor mode utilization", "Percentage of time CPU spent in supervisor mode", Double.class),
                        new FileField("Executive mode utilization", "Percentage of time CPU spent in executive mode", Double.class),
                        new FileField("Kernel mode utilization", "Percentage of time CPU spent in kernel mode, in process, context, exclusive of spin lock waits", Double.class),
                        new FileField("Compatibility mode utilization", "Percentage of time CPU spent in PDP-11 compatibility mode (VAX only; Alpha and I64 will be zero)", Double.class),
                        new FileField("MP Sync utilization", "Percentage of time CPU spent spin lock wait", Double.class),
                        new FileField("Interrupt stack utilization", "Percentage of time CPU spent in interrupt stack context, exclusive of spin wait time", Double.class),
            }
            ),
            /*
             * For each interval there will be only one memory statistic data record.
             */
                    new FileRecord(
                    "Memory Statistics",
                    new MessageFormat("[MEM]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSMEM"),
                        INTERVAL_LENGTH,
                        new FileField("Physical memory size", "Total number of physical pages of memory", Double.class),
                        new FileField("System working set", "Number of pages actually allocated to OpenVMS", Double.class),
                        new FileField("Virtual I/O cache size", "Number of pages allocated to the virtual I/O cache", Double.class),
                        new FileField("Average modified page list size", "Average number of pages in modified list", Double.class),
                        new FileField("Average free page list size", "Average number of pages in free list", Double.class),
                        new FileField("Balance set", "Number of processes in the balance set", Double.class),
                        new FileField("Total processes", "Total number of processes in the system", Double.class),
                        new FileField("Interactive processes", "Number of interactive processes in the system", Double.class),
                        new FileField("Network processes", "Number of network processes in the system", Double.class),
                        new FileField("Batch processes", "Number of batch processes in the system", Double.class),
                        new FileField("In-swap rate", "Process in-swaps per second", Double.class),
                        new FileField("Out-swap rate", "Process out-swaps per second", Double.class),
                        new FileField("Header in-swap rate", "Process header in-swaps per second", Double.class),
                        new FileField("Header out-swap rate", "Process header out-swaps per second", Double.class),
                        new FileField("Memory queue", "Average number of processes in COMO scheduling state", Double.class),
                        new FileField("Bytes per page", "Size of page", Double.class),
                        new FileField("Memory Utilization", "Total amount of physical memory used", Double.class),
            }
            ),
            /*
             * For each interval there will be only one paging statistics data record.
             */
                    new FileRecord(
                    "Paging Statistics",
                    new MessageFormat("[PAG]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSPAG"),
                        INTERVAL_LENGTH,
                        new FileField("Total Page Faults", "Total Page Faults per second", Double.class),
                        new FileField("Page read rate", "Page read operations from disk per second", Double.class),
                        new FileField("Pages read rate", "Number of pages read to disk per second", Double.class),
                        new FileField("Page write rate", "Page write operations to disk per second", Double.class),
                        new FileField("Pages written rate", "Number of pages written to disk per second", Double.class),
                        new FileField("Demand zero fault rate", "Page faults per second resolved as demand-zero pages", Double.class),
                        new FileField("Global valid fault rate", "Page faults per second resolved as valid global pages", Double.class),
                        new FileField("System fault rate", "Page faults per second incurred in system address space", Double.class),
                        new FileField("Free list fault rate", "Page faults per second resolved from free list", Double.class),
                        new FileField("Modified list fault rate", "Page faults per second resolved from modified list", Double.class),
                        new FileField("Bad list fault rate", "Page faults per second resolved from the bad list", Double.class),
                        new FileField("Transition fault rate", "Page faults per second resolved from pages currently being written to disk", Double.class),
                        new FileField("Write-in-progress fault rate", "Page faults per second resolved from pages currently being written to disk", Double.class),
                        new FileField("Hard fault rate", "Faults per second resolved from disk", Double.class),
                        new FileField("Soft fault rate", "Faults resolved from memory per second", Double.class),
            }
            ),
            /*
             * For each collection interval, there will be only one general I/O statistics record.
             */
                    new FileRecord(
                    "General I/O Statistics",
                    new MessageFormat("[GIO]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSGIO"),
                        INTERVAL_LENGTH,
                        new FileField("Direct I/O rate", "Direct I/Os per second performed exclusive of page and swap I/O and I/O to mapped image sections", Double.class),
                        new FileField("Buffered I/O rate", "Buffered I/Os per second", Double.class),
                        new FileField("Logical name translation rate", "Logical name translations per second", Double.class),
                        new FileField("Mailbox read rate", "Mailbox read operations per second", Double.class),
                        new FileField("Mailbox write rate", "Mailbox write operations per second", Double.class),
                        new FileField("Window turn rate", "XQP WCB updates per second", Double.class),
                        new FileField("Window hit rate", "XQP successful virtual to logical block maps per second", Double.class),
                        new FileField("Split I/O rate", "XQP mappings per second of single virtual I/O request more than one logical segment", Double.class),
                        new FileField("Erase I/O rate", "Disk erase I/O operations per second", Double.class),
                        new FileField("File open rate", "File open requests per second", Double.class),
                        new FileField("Files open", "Average number of open files", Double.class),
            }
            ),
            /*
             * For each interval there will be one XQP statistics data record.
             */
                    new FileRecord(
                    "XQP Statistics",
                    new MessageFormat("[XQP]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSXQP"),
                        INTERVAL_LENGTH,
                        new FileField("Directory hit rate", "Successful attempts per second to find directory FCBs in the directory cache", Double.class),
                        new FileField("Directory miss rate", "Unsuccessful attempts per second to find directory FCBs in the directory cache", Double.class),
                        new FileField("Quota hit rate", "Successful attempts per second to find entries in the quota cache", Double.class),
                        new FileField("Quota miss rate", "Unsuccessful attempts per second to find entries in the quota cache", Double.class),
                        new FileField("File id hit rate", "Successful attempts per second to find file IDs in the file ID cache", Double.class),
                        new FileField("File id miss rate", "Unsuccessful attempts per second of find file IDs in the file ID cache", Double.class),
                        new FileField("Extent hit rate", "Successful attempts per second to find extents in the extend cache", Double.class),
                        new FileField("Extent miss rate", "Unsuccessful attempts per second to find extents in the extend cache", Double.class),
                        new FileField("File header found rate", "Number of file header items found per second in cache", Double.class),
                        new FileField("File header not found rate", "Number of file header items not found per second in cache", Double.class),
                        new FileField("Directory data hit rate", "Successful attempts per second to find directory data in the directory data cache", Double.class),
                        new FileField("Directory data miss rate", "Unsuccessful attempts per second to find directory data in the directory data cache", Double.class),
                        new FileField("Bitmap hit rate", "Successful attempts per second to find entries in the storage bitmap cache", Double.class),
                        new FileField("Bitmap miss rate", "Unsuccessful attempts per second to find entries in the storage bitmap cache", Double.class),
            }
            ),
            /*
             * For each interval there will be only one lock statistics data record.
             */
                    new FileRecord(
                    "Lock Statistics",
                    new MessageFormat("[LCK]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSLCK"),
                        INTERVAL_LENGTH,
                        new FileField("Total number of locks", "Sample count of number of locks", Double.class),
                        new FileField("Total number of resources", "Sample count of number of resources", Double.class),
                        new FileField("Local enqueue rate", "New local locks requested per second", Double.class),
                        new FileField("Incoming enqueue rate", "New lock requests per second coming from other nodes in the cluster", Double.class),
                        new FileField("Outgoing enqueue rate", "New lock requests per second sent to other nodes of the cluster", Double.class),
                        new FileField("Local conversion rate", "Local lock conversion requests per second", Double.class),
                        new FileField("Incoming conversion rate", "Lock conversion requests per second coming from other nodes in the cluster", Double.class),
                        new FileField("Outgoing conversion rate", "Lock conversion requests per second sent to other nodes of the cluster", Double.class),
                        new FileField("Local dequeue rate", "Local lock releases per second", Double.class),
                        new FileField("Incoming dequeue rate", "Lock releases per second coming from other nodes in the cluster", Double.class),
                        new FileField("Outgoing dequeue rate", "Lock releases per second sent to other nodes in the cluster", Double.class),
                        new FileField("Enqueue wait rate", "Number of locks request per second forced to wait", Double.class),
                        new FileField("Enqueue not queued rate", "Number of lock requests per second not granted and process did not wait", Double.class),
                        new FileField("Local blocking AST rate", "Local blocking ASTs per second", Double.class),
                        new FileField("Incoming blocking AST rate", "Blocking ASTs per second coming from other nodes in the cluster", Double.class),
                        new FileField("Outgoing blocking AST rate", "Blocking ASTs per second sent to other nodes in the cluster", Double.class),
                        new FileField("Incoming deadlock message rate", "Deadlock detection messages per second from other nodes in the cluster", Double.class),
                        new FileField("Outgoing deadlock message rate", "Deadlock detection messages per second sent to other nodes in cluster", Double.class),
                        new FileField("Deadlock search rate", "Deadlock searches per second", Double.class),
                        new FileField("Deadlock find rate", "Deadlock finds per second", Double.class),
                        new FileField("Incoming directory function rate", "Directory function messages per second coming from other nodes in the cluster", Double.class),
                        new FileField("Outgoing directory function rate", "Directory function messages per second sent to other nodes in the cluster", Double.class),
            }
            ),
            /*
             * For each interval, for each mounted disk there will be a disk statistics record.
             */
                    new FileRecord(
                    "Disk Statistics",
                    new MessageFormat("[DSK.{5}]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSDSK"),
                        INTERVAL_LENGTH,
                        new FileField("Disk name", "Disk name string", String.class),
                        new FileField("Total I/O rate", "Total I/Os per second performed to disk", Double.class),
                        new FileField("User I/O rate", "I/Os per second that were neither paging or swapping I/Os", Double.class),
                        new FileField("Page I/O rate", "Paging operation I/Os to disk", Double.class),
                        new FileField("Swap I/O rate", "Swapping operation I/Os to disk", Double.class),
                        new FileField("Read I/Os", "Percentage of I/Os that were read from disk", Double.class),
                        new FileField("Served I/Os", "Percentage of all I/Os that were served to other nodes in their cluster", Double.class),
                        new FileField("Disk utilization", "Percentage of time disk was busy from perspective of this node", Double.class),
                        new FileField("Average queue length", "Time-average number of I/O requests either being performed by the disk or waiting to be processed", Double.class),
                        new FileField("Average service time", "Average time in milliseconds required to perform a disk operation once operation initiated", Double.class),
                        new FileField("Average response time", "Average time in milliseconds required to complete a disk I/O request including both wait time and service time", Double.class),
                        new FileField("Free blocks", "Number of free blocks on disk", Double.class),
                        new FileField("Total blocks", "Total number of blocks on disk", Double.class),
            }
            ),
            /*
             * For each interval, for each local port/ remote SCS node combination
             * there will be an SCS statistics data record.
             */
                    new FileRecord(
                    "SCS Statistics",
                    new MessageFormat("[SCS.{5}]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSSCS"),
                        INTERVAL_LENGTH,
                        new FileField("Local port or remote node", "Concatenated string of local port name and remote SCS node name, separated by a hyphen", String.class),
                        new FileField("Datagram receive rate", "Datagram messages received per second", Double.class),
                        new FileField("Datagram send rate", "Datagram messages sent per second", Double.class),
                        new FileField("Datagram discard rate", "Sent datagram messages discarded per second", Double.class),
                        new FileField("Kbytes mapped rate", "Kilobytes mapped per second by the local node for block transfers", Double.class),
                        new FileField("Kbytes request rate", "Kilobytes read per second by locally initiated block mode transfers", Double.class),
                        new FileField("Kbytes send rate", "Kilobytes written per second by locally initiated block mode transfers", Double.class),
                        new FileField("Seq msg receive rate", "Sequenced messages received per second", Double.class),
                        new FileField("Seq msg send rate", "Sequenced messages sent per second", Double.class),
                        new FileField("Queued buffer descriptor rate", "Queued buffer descriptors per second", Double.class),
                        new FileField("Queued send credit rate", "Send credit waits per second", Double.class),
                        new FileField("Request data rate", "Number of locally initiated block mode reads per second", Double.class),
                        new FileField("Send data rate", "Number of locally initiated block mode writes per second", Double.class),
            }
            ),
            /*
             * For each interval, for each process existing at the end of the interval,
             * there will be a process statistics record.
             */
                    new FileRecord(
                    "Process Statistics",
                    new MessageFormat("[PRC.{5}]"),
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSPRC"),
                        INTERVAL_LENGTH,
                        new FileField("Process ID", "Process ID", String.class /* hex */),
                        new FileField("Process type", "String specifying process types", String.class),
                        new FileField("Process scheduling state", "Scheduling state string of the process at interval end", String.class),
                        new FileField("Process name", "Name of process (enclosed in double quotes)", String.class),
                        new FileField("User name", "User name of process owner (enclosed in double quotes)", String.class),
                        new FileField("Process account name", "Account name under which process runs (enclosed in double quotes)", String.class),
                        new FileField("Base priority", "Base priority", Integer.class),
                        new FileField("Average priority", "Average priority", Integer.class),
                        new FileField("CPU time", "CPU time in seconds attributed the process during the interval", Double.class),
                        new FileField("Direct I/Os", "Direct I/Os attributed to the process during the interval", Double.class),
                        new FileField("Buffered I/Os", "Buffered I/Os attributed to the process during the interval", Double.class),
                        new FileField("Page faults", "Page faults attributed to the process during the interval", Double.class),
                        new FileField("Fault I/Os", "Page fault I/Os attributed to the process during the interval", Double.class),
                        new FileField("Minimum working set size", "Minimum working set size of process in pages during the interval", Double.class),
                        new FileField("Maximum working set size", "Maximum working set size of process in pages during the interval", Double.class),
                        new FileField("Average working set size", "Average working set size of process in pages during the interval", Double.class),
                        new FileField("Average virtual address space", "Average virtual address space size in pages during the interval", Double.class),
                        new FileField("Average process private working set", "Average number of process private pages in working set during the interval", Double.class),
                        new FileField("Average global working set", "Average number of global pages in working set during the interval", Double.class),
                        new FileField("Image activations", "Number of image activation during the interval", Integer.class),
                        new FileField("Current image", "Current image name enclosed in double quotes", String.class),
            }
            ),
            /*
             * For each sysgen parameter, there will be a sysgen parameter statistics data record.
             * This data record will be printed only once at the beginning.
             * (This is a TDC-only metric and is not available with ECP collections.)
             */
                    new FileRecord(
                    "SYSGEN Parameter Statistics",
                    null, // no message format
                    new FileField[] {
                INTERVAL_START_TIME,
                        INTERVAL_END_TIME,
                        HOST_NAME,
                        new FileField("VMSPAR"),
                        INTERVAL_LENGTH,
                        new FileField("Parameter name", "Parameter name", String.class),
                        new FileField("Parameter value", "Value of the corresponding paramter", String.class /* or Integer.class */),
            }
            ),
        };
    }
    
}
