/*
 * TLCDataCollection.java
 */

package de.pdv_systeme.DataCollection;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.LineNumberReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.regex.Pattern;

import org.jfree.data.time.RegularTimePeriod;

import de.pdv_systeme.DataCollection.AbstractDataCollection.DataCollectionTimeSeries;
import de.pdv_systeme.util.VmsDate;
import org.jfree.data.time.TimeSeries;


/**
 * <code>TLCDataCollection</code> reads a TimeLine Collaboration (TLC) formatted file
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class TLCDataCollection
        extends FileDataCollection {
    
    /**
     * Unchecked exception
     * thrown when there's a problem with the header lines
     */
    public static class IncompleteHeaderException extends RuntimeException {
        
        /** Construct a new IncompleteHeaderException. */
        IncompleteHeaderException() { super(); }
        
        /**
         * Construct a new IncompleteHeaderException.
         * @param s  the detail message
         */
        IncompleteHeaderException(String s) { super(s); }
    }
    
    /**
     * Unchecked exception
     * thrown when the number of column headers
     * doesn't match the number of values in a line
     */
    public static class ElementCountMismatchException extends RuntimeException {
        
        /** Construct a new ElementCountMismatchException. */
        ElementCountMismatchException() { super(); }
        
        /**
         * Construct a new ElementCountMismatchException.
         * @param s  the detail message
         */
        ElementCountMismatchException(String s) { super(s); }
    }
    
    /** Nested class holding item descriptions */
    private static class Description {
        Pattern pattern;
        String item;
        String description;
        Description(Pattern p, String i, String d) {
            pattern = p; item = i; description = d;
        }
    }
    
    /**
     * Pattern to sort out Sample Time columns
     * @see #readHeaderImpl
     */
    private static final Pattern sampleTimePattern
            = Pattern.compile("^(\\[[^\\]]+\\])?Sample Time$");
    
    /** String to delimit the column headers */
    private static final String COLUMN_HEADER_START
            = "$$$ START COLUMN HEADERS $$$";
    private static final String COLUMN_HEADER_END
            = "$$$ END COLUMN HEADERS $$$";
    
    /** Flag: Use all timestamps? */
    protected boolean useAllTimestamps;
    
    /** Reference date from header */
    private String refDate;
    
    /** The column headers, as they appear in the file */
    private String[] fileColumn;
    
    /** Flag: is a file column a data (or a sample time) column? */
    private boolean[] isDataColumn;
    
    // TODO: DESCRIPTION should really be final. Are there no final objects in Java?
    /** Description of TLC items */
    private static /*final*/ Description[] DESCRIPTION;
    
    /**
     * Creates a new instance of TLCDataCollection
     */
    public TLCDataCollection() {
        super();
        format = "TLC";
        this.useAllTimestamps = false;
    }
    
    /**
     * Creates a new instance of TLCDataCollection.
     * @param file File object representing the TLC file
     * @throws java.io.FileNotFoundException
     * @throws java.io.IOException
     * @throws java.text.ParseException
     * @throws java.lang.NumberFormatException
     * @throws java.lang.IllegalArgumentException
     * @throws IncompleteHeaderException
     * @throws ElementCountMismatchException
     */
    public TLCDataCollection(final File file)
    throws FileNotFoundException,
            IOException,
            ParseException,
            NumberFormatException,
            IllegalArgumentException,
            IncompleteHeaderException,
            ElementCountMismatchException {
        this(file, false);
    }
    
    /**
     * Creates a new instance of TLCDataCollection.
     * @param file File object representing the TLC file
     * @param useAllTimestamps flag: if a line contains multiple timestamps, use them all?
     * @throws java.io.FileNotFoundException
     * @throws java.io.IOException
     * @throws java.text.ParseException
     * @throws java.lang.NumberFormatException
     * @throws java.lang.IllegalArgumentException
     * @throws IncompleteHeaderException
     * @throws ElementCountMismatchException
     */
    public TLCDataCollection(final File file, final boolean useAllTimestamps)
    throws FileNotFoundException,
            IOException,
            ParseException,
            NumberFormatException,
            IllegalArgumentException,
            IncompleteHeaderException,
            ElementCountMismatchException {
        super(file);
        format = "TLC";
        this.useAllTimestamps = useAllTimestamps;
    }
    
    /** {@inheritDoc} */
    protected void initialize() {
        initialize_DESCRIPTION();
    }
    
    /* {@inheritDoc} */
    /**
     * Test the file format
     * @param file  File object representing the file
     * @return  true iff the file has the correct format
     */
    public static boolean testFormat(final File file)
    throws FileNotFoundException, IOException {
        boolean isTLC = false;
        if (file != null) {
            BufferedReader in = new BufferedReader(new FileReader(file));

            // 1st element of 1st line must be a valid nodename
            String line = in.readLine();
            String nodename = line.split(",")[0];
            isTLC = Pattern.compile("^\\w+$", Pattern.CASE_INSENSITIVE).matcher(nodename).matches();

            // 2nd line...
            line = in.readLine();
            String refDate = "";
            if (line.toUpperCase().compareTo(COLUMN_HEADER_START) == 0) {
                // ...either is an introducer to variable length column name lines
                // Then there must be a delimiting line.
                do {
                    line = in.readLine();
                } while (line.toUpperCase().compareTo(COLUMN_HEADER_END) != 0);
            } else {
                // ... or 1st element must be a valid VMS timestamp
                refDate = line.split(",")[0];
                // -- do not use; e.g. in TLC files save by TLViz, 2nd & 3rd line are empty
                // isTLC = isTLC && VmsDate.datePattern.matcher(refDate).matches();

                // Skip 3rd (times) and 4th (column headers) line
                in.readLine();
                in.readLine();
            }

            // 1st element of 1st data line must be a valid VMS timestamp
            // (may consist of only the time part)
            line = in.readLine();
            try {
                VmsDate.parse(line.split(",")[0].trim(), refDate);
            } catch (ParseException e) {
                isTLC = false;
            }
            in.close();
        }
        return isTLC;
    }
    
    /** {@inheritDoc} */
    protected void readData(
            LineNumberReader in,
            FileDataCollection timeDataCollection
            )
            throws IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException,
            IncompleteHeaderException,
            ElementCountMismatchException {
        readHeaderImpl(in);
        readDataImpl(in, timeDataCollection);
    }
    
    /**
     * Read the TLC header.
     * @param in  the input stream
     * @throws IncompleteHeaderException
     */
    private void readHeaderImpl(LineNumberReader in)
    throws IncompleteHeaderException, ParseException {
        // Read the header
        char[] comment;
        String[] fc = null;
        try {
            comment = in.readLine().toCharArray();
            String line = in.readLine();
            if (line.toUpperCase().compareTo(COLUMN_HEADER_START) == 0) {
                // Read column headers
                ArrayList fcv = new ArrayList();
                while ((line = in.readLine()).toUpperCase().compareTo(COLUMN_HEADER_END) != 0) {
                    fc = line.split(",", -1);
                    for (int i = 0; i < fc.length; i++)
                        fcv.add(fc[i]);
                }
                fc = (String[]) fcv.toArray(new String[fcv.size()]);
                refDate = "";
            } else {
                // Read start dates and times
                String[] date0 = line.split(",");
                refDate = (date0.length > 0) ? date0[0] : "";
                in.readLine(); // String[] time0 = in.readLine().split(",");
                // Read column headers
                fc = in.readLine().split(",", -1);
            }
        } catch (IOException e) {
            throw new IncompleteHeaderException(e.getMessage());
        }
        
        // Init timeSeriesMap and isDataColumn from column headers
        // Trim column headers
        fileColumn = new String[fc.length];
        isDataColumn = new boolean[fc.length];
        for (int i = 0; i < fc.length; i++) {
            String key = fc[i].trim();
            fileColumn[i] = key;
            isDataColumn[i] = ! sampleTimePattern.matcher(key).matches();
            if (isDataColumn[i]) {
                DataCollectionTimeSeries ts = new DataCollectionTimeSeries(key);
                ts.ensureCapacity(lineCount);
                addTimeSeries(ts);
            }
        }
        
        // Initialize info from comment
        // Can't use String.split() because of parenthesized lists of comma-separated items
        List infoList = new ArrayList();
        StringBuffer sb = new StringBuffer(1024);
        boolean inParens = false;
        for (int i = 0; i < comment.length; i++) {
            char ch = comment[i];
            if (ch != ',' || inParens) {
                sb.append(ch);
                if (ch == '(') {
                    inParens = true;
                } else if (ch == ')') {
                    inParens = false;
                }
            } else {
                infoList.add(sb.toString().trim());
                sb.delete(0, sb.length());
            }
        }
        // No info yet? Use what we have...
        if (infoList.size() == 0 && sb.length() > 0)
            infoList.add(sb.toString().trim());

        nodename = (String) infoList.remove(0);
        for (Iterator ii = infoList.iterator(); ii.hasNext(); ) {
            String s = (String) ii.next();
            if (s.equals(nodename))
                ii.remove();
        }
        info = (String[]) infoList.toArray(new String[infoList.size()]);
    }
    
    /**
     * Read the TLC data records.
     * @param in  the input stream
     * @param timeDataCollection  the data collection to provide the timestamps (may be <code>null</code>)
     * @throws ElementCountMismatchException
     */
    private void readDataImpl(
            LineNumberReader in,
            FileDataCollection timeDataCollection
            )
            throws IOException,
            ParseException,
            InterruptedException,
            NumberFormatException,
            IllegalArgumentException,
            ElementCountMismatchException {
        TimeSeries times = null;
        if (timeDataCollection != null) {
            String name = (String) timeDataCollection.getAllItemNames().get(0);
            times = timeDataCollection.getData(name);
        }
        String line;
        RegularTimePeriod timestamp;
        int recordNum = 0;
        while ((line = in.readLine()) != null && ! isInterrupted) {
            String[] lineItem = line.split(",", -1);
            if (lineItem[0].equals("Column Average"))
                continue;
            if (lineItem.length != fileColumn.length) {
                throw new ElementCountMismatchException(
                        "Line "+in.getLineNumber()+": "+
                        lineItem.length+" data points, but "+
                        fileColumn.length+" column names"
                        );
            }
            
            // Store data values
            timestamp = null;
            for (int itemNum = 0; itemNum < lineItem.length; itemNum++) {
                lineItem[itemNum] = lineItem[itemNum].trim();
                if (lineItem[itemNum].length() == 0)
                    continue;
                if (isDataColumn[itemNum]) {
                    String name = fileColumn[itemNum];
                    DataCollectionTimeSeries dcts = getTimeSeries(name);
                    
                    try {
                        dcts.addValue(timestamp, Double.valueOf(lineItem[itemNum]));
                    } catch (NumberFormatException e) {
                        throw new NumberFormatException(
                                "Line "+in.getLineNumber()+", item "+itemNum+": "+
                                lineItem[itemNum]+", "+e.getMessage()
                                );
                    }
                } else { // a timestamp column
                    if (itemNum == 0 || useAllTimestamps) {
                        if (times != null) {
                            timestamp = times.getDataItem(recordNum++).getPeriod();
                        } else {
                            Date timeString;
                            try {
                                timeString = VmsDate.parse(lineItem[itemNum], refDate);
                            } catch (ParseException e) {
                                throw new ParseException(
                                        "Line "+in.getLineNumber()+", item "+itemNum+": "+
                                        lineItem[itemNum]+", "+e.getMessage(),
                                        e.getErrorOffset()
                                        );
                            }
                            timestamp = RegularTimePeriod.createInstance(
                                    DataCollectionTimeSeries.timeBaseClass(),
                                    timeString,
                                    RegularTimePeriod.DEFAULT_TIME_ZONE
                                    );
                        }
                    }
                }
            }
        }
        
        if (isInterrupted) {
            fileColumn = null;
            isDataColumn = null;
            throw new InterruptedException();
        }
    }
    
    /** {@inheritDoc} */
    public String getDescription(final String name) {
        if (! contains(name))
            throw new NoSuchElementException(name);
        
        for (int i = 0; i < DESCRIPTION.length; i++) {
            Description dsc = DESCRIPTION[i];
            if (dsc.pattern.matcher(name).matches())
                return dsc.item+": "+dsc.description;
        }
        return name;
    }
    
    /**
     * Initialize the DESCRIPTIONs
     */
    private static void initialize_DESCRIPTION() {
        DESCRIPTION = new Description[] {
            
            // DECNET: DECnet for OpenVMS statistics
            new Description(
                    Pattern.compile("\\[NET\\.DECN\\]Arriving local packets"),
                    "Arriving Local Packet Rate",
                    "Rate at which local packets are being received."
                    ),
                    new Description(
                    Pattern.compile("\\[NET\\.DECN\\]Departing local packets"),
                    "Departing Local Packet Rate",
                    "Rate at which local packets are being sent."
                    ),
                    new Description(
                    Pattern.compile("\\[NET\\.DECN\\]Arriving transit packets"),
                    "Arriving Transit Packet Rate",
                    "Rate at which transit packets are arriving."
                    ),
                    new Description(
                    Pattern.compile("\\[NET\\.DECN\\]Transit congestion loss"),
                    "Transit Congestion Loss Rate",
                    "Rate of transit congestion loss."
                    ),
                    new Description(
                    Pattern.compile("\\[NET\\.DECN\\]Receive buffer failures"),
                    "Receiver Buffer Failure Rate",
                    "Rate of receiver buffer failures."
                    ),
                    
                    // DISK: Disk I/O statistics
                    new Description(
                    Pattern.compile("\\[MON\\.DISK\\]OpRate"),
                    "I/O Operation Rate",
                    "Rate at which I/O operations occur on each disk. Because this statistic does not provide information about the time required for individual operations, use discretion in interpreting it."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DISK\\]QLen"),
                    "I/O Request Queue Length",
                    "Number of outstanding I/O request packets. Includes the request currently being serviced and those awaiting service. Note that, for greater precision, this item is always sampled at a 1-second interval."
                    ),
                    
                    // DLOCK: Distributed lock management statistics
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]New Locks \\(Local\\)"),
                    "New ENQ Rate (Local)",
                    "Rate of new lock (ENQ) requests that originate and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]New Locks \\(Incoming\\)"),
                    "New ENQ Rate (Incoming)",
                    "Rate of new lock requests that originate on other systems and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]New Locks \\(Outgoing\\)"),
                    "New ENQ Rate (Outgoing)",
                    "Rate of new lock requests that originate on this system and are performed on another system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock conversions \\(Local\\)"),
                    "Converted ENQ Rate (Local)",
                    "Rate of lock (ENQ) conversion requests that originate and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock conversions \\(Incoming\\)"),
                    "Converted ENQ Rate (Incoming)",
                    "Rate of lock conversion requests that originate on other systems and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock conversions \\(Outgoing\\)"),
                    "Converted ENQ Rate (Outgoing)",
                    "Rate of lock conversion requests that originate on this system and are performed on another system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock DEQs \\(local\\)"),
                    "DEQ Rate (Local)",
                    "Rate of unlock (DEQ) requests that originate and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock DEQs \\(incoming\\)"),
                    "DEQ Rate (Incoming)",
                    "Rate of unlock requests that originate on other systems and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Lock DEQs \\(outgoing\\)"),
                    "DEQ Rate (Outgoing)",
                    "Rate of unlock requests that originate on this system and are performed on another system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Local BLKASTs"),
                    "Blocking AST Rate (Local)",
                    "Rate of lock manager blocking ASTs that originate and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Incoming BLKASTs"),
                    "Blocking AST Rate (Incoming)",
                    "Rate of lock manager blocking ASTs that originate on other systems and are performed on this system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Outgoing BLKASTs"),
                    "Blocking AST Rate (Outgoing)",
                    "Rate of lock manager blocking ASTs that originate on this system and are performed on another system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]DIR functions \\(Incoming\\)"),
                    "Directory Function Rate (Incoming)",
                    "Rate of requests for locks being managed by this node"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]DIR functions \\(Outgoing\\)"),
                    "Directory Function Rate (Outgoing)",
                    "Rate of requests for locks being managed by other nodes"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.DLOC\\]Deadlock messages"),
                    "Deadlock Message Rate",
                    "Rate of incoming and outgoing messages required for deadlock detection"
                    ),
                    
                    // FCP: File control primitive statistics
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP calls"),
                    "FCP Call Rate",
                    "Rate of QIO requests received by the file system."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Disk space allocations"),
                    "Allocation Rate",
                    "Rate of calls that caused allocation of disk space."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Files created"),
                    "Create Rate",
                    "Rate at which new files were created."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Read I/Os"),
                    "Disk Read Rate",
                    "Rate of read I/O operations from disk by the file system."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Write I/Os"),
                    "Disk Write Rate",
                    "Rate of write I/O operations to disk by the file system."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP waits"),
                    "Volume Lock Wait Rate",
                    "Rate of entry into a wait state due to contention for a volume synchronization lock. Volume synchronization locks are removed by the XQP during file creation, deletion, extension, and truncation operations."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Cpu"),
                    "CPU Tick Rate",
                    "Rate at which CPU time was used by the file system (in 10-millisecond ticks)."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP faults"),
                    "File System Page Fault Rate",
                    "Rate at which page faults occurred in the file system."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP window misses"),
                    "Window Turn Rate",
                    "Rate of file-map window misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Filename lookups"),
                    "File Lookup Rate",
                    "Rate of file name lookup operations in file directories."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP Files opened"),
                    "File Open Rate",
                    "Rate at which files were opened."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FCP\\]FCP erases"),
                    "Erase Rate",
                    "Rate of erase operations issued by the file system."
                    ),
                    
                    // FILE_SYSTEM_CACHE: File system cache statistics
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Dir FCB cache hit %"),
                    "Directory FCB Hit%",
                    "Percentage of directory file control block hits on the directory cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Dir FCB cache attempts"),
                    "Directory FCB Attempt Rate",
                    "Rate at which attempts were made to find directory file control blocks in the directory cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Dir data cache hit %"),
                    "Directory Data Hit%",
                    "Percentage of directory data hits on the directory cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Dir data cache attempts"),
                    "Directory Data Attempt Rate",
                    "Rate at which attempts were made to find directory data in the directory cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]File Header cache hit %"),
                    "File Header Hit%",
                    "Percentage of file header hits on the file header cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]File Header cache attempts"),
                    "File Header Attempt Rate",
                    "Rate at which attempts were made to find file headers in the file header cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]FID cache hit %"),
                    "File ID Hit%",
                    "Percentage of file identifier hits on the file ID cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]FID cache attempts"),
                    "File ID Cache Attempt Rate",
                    "Rate at which attempts were made to find file identifiers in the file ID cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Extent cache hit %"),
                    "Extent Cache Hit%",
                    "Percentage of appropriate size extent hits on the extent cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Extent cache attempts"),
                    "Extent Cache Attempt Rate",
                    "Rate at which attempts were made to find appropriate size extents in the extent cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Quota cache hit %"),
                    "Quota Cache Hit%",
                    "Percentage of quota entry hits on the quota cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Quota cache attempts"),
                    "Quota Cache Attempt Rate",
                    "Rate at which attempts were made to find entries in the quota cache."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Storage Bitmap hit %"),
                    "Bitmap Cache Hit%",
                    "Percentage of entry hits on the bitmap cache. The percentage value shown is the ratio of hits to the sum of hits plus misses."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.FLSC\\]Storage Bitmap attempts"),
                    "Bitmap Cache Attempt Rate",
                    "Rate at which attempts were made to find entries in the bitmap cache."
                    ),
                    
                    // IO: System I/O statistics
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]Mailbox Write Rate"),
                    "Mailbox Write Rate",
                    "Rate of write-to-mailbox requests received by the system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]Split Transfer Rate"),
                    "Split Transfer Rate",
                    "Rate at which transfers were split into multiple I/Os"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]Logical Name Trans"),
                    "Log Name Translation Rate",
                    "Rate of logical name translations"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]File Open Rate"),
                    "File Open Rate",
                    "Rate at which files were opened"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]Page Reads"),
                    "Page Read Rate",
                    "Rate of pages read from disk as a result of page faults"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]Page Writes"),
                    "Page Write Rate",
                    "Rate of pages written to the page file"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.IO\\]In Swaps"),
                    "Inswap Rate",
                    "Rate at which working sets were read into memory from the swap file"
                    ),
                    
                    // LOCK: Lock management statistics
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]New ENQ rate"),
                    "New ENQ Rate",
                    "Rate of new lock (ENQ) requests (as opposed to conversions)"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Converted ENQ rate"),
                    "Converted ENQ Rate",
                    "Rate of lock (ENQ) conversion requests"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]DEQs"),
                    "DEQ Rate",
                    "Rate of unlock (DEQ) requests"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Blocking AST Rate"),
                    "Blocking AST Rate",
                    "Rate of lock manager blocking ASTs delivered"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]ENQs forced to wait"),
                    "ENQs Forced To Wait Rate",
                    "Rate of occurrence of locks that could not be granted immediately, thus causing a wait"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Enqs not queued"),
                    "ENQs Not Queued Rate",
                    "Rate of occurrence of locks that could not be granted immediately but requested not to be queued, and thus received an error status instead"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Deadlock searches"),
                    "Deadlock Search Rate",
                    "Rate at which a deadlock search was performed"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Deadlocks found"),
                    "Deadlock Find Rate",
                    "Rate at which deadlocks were found"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Locks Total"),
                    "Total Locks",
                    "Total number of locks in the system"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.LOCK\\]Resources Total"),
                    "Total Resources",
                    "Total number of resources in the system"
                    ),
                    
                    // MODES: Time spent in each of the processor modes
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Interrupt State"),
                    "Interrupt Stack (VAX) / Interrupt State (Alpha)",
                    "Time spent on the interrupt stack (on VAX systems) or in an interrupt state on a kernel stack (on Alpha systems)."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Mp Synch"),
                    "MP Synchronization",
                    "Time spent synchronizing multiple CPUs (applicable to multiprocessor systems only)."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Kernel Mode"),
                    "Kernel Mode",
                    "Time spent in kernel mode, but not in an interrupt state."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Exec Mode"),
                    "Executive Mode",
                    "Time spent in executive mode."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Super Mode"),
                    "Supervisor Mode",
                    "Time spent in supervisor mode."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]User Mode"),
                    "User Mode",
                    "Time spent in user mode executing instructions."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODE\\]Idle Time"),
                    "Idle Time",
                    "Time not spent in any of the other modes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* Inter mode"),
                    "Interrupt Stack (VAX) / Interrupt State (Alpha)",
                    "Time spent on the interrupt stack (on VAX systems) or in an interrupt state on a kernel stack (on Alpha systems)."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* MpSync mode"),
                    "MP Synchronization",
                    "Time spent synchronizing multiple CPUs (applicable to multiprocessor systems only)."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* Kernel mode"),
                    "Kernel Mode",
                    "Time spent in kernel mode, but not in an interrupt state."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* Exec mode"),
                    "Executive Mode",
                    "Time spent in executive mode."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* Super mode"),
                    "Supervisor Mode",
                    "Time spent in supervisor mode."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* User mode"),
                    "User Mode",
                    "Time spent in user mode executing instructions."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MODES\\]Cpu .* Idle"),
                    "Idle Time",
                    "Time not spent in any of the other modes."
                    ),
                    
                    // MSCP_SERVER: MSCP server statistics
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP requests"),
                    "Server I/O Request Rate",
                    "The rate at which remote processors request I/O transfers."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP reads"),
                    "Read Request Rate",
                    "The rate at which remote processors request Read I/O transfers."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP writes"),
                    "Write Request Rate",
                    "The rate at which remote processors request Write I/O transfers."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP fragments"),
                    "Extra Fragment Rate",
                    "The rate at which the server issues extra fragments. One or more extra fragments are created when, due to buffering constraints, the MSCP server issues multiple I/Os in order to fulfill a single I/O request."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP splits"),
                    "Fragmented Request Rate",
                    "The rate at which fragmented requests occur. A fragmented request is a transfer request that the server fragments due to buffering constraints."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP buffer waits"),
                    "Buffer Wait Rate",
                    "The rate at which 'buffer waits' occur in the server. A buffer wait occurs when a request must wait for MSCP buffer memory."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.MSCP\\]MSCP size .*"),
                    "Request Size Rates",
                    "A histogram that displays the rate of requests for various block sizes."
                    ),
                    
                    // PAGE: Page management statistics
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Page Write I/O rate"),
                    "Page Write I/O Rate",
                    "Rate of write I/O operations to the page file"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Free List Faults"),
                    "Free List Fault Rate",
                    "Rate at which pages were read from the free-page list as a result of page faults"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Modified List faults"),
                    "Modified List Fault Rate",
                    "Rate of pages read from the modified-page list as a result of page faults"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Demand-zero faults"),
                    "Demand Zero Fault Rate",
                    "Rate at which zero-filled pages were allocated as a result of page faults"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Global Valid Fault Rate"),
                    "Global Valid Fault Rate",
                    "Rate of page faults for pages that are not in the process's working set, but are in physical memory and are indicated as valid pages in the systemwide global page tables"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Write-in-Progress faults"),
                    "Writes In Progress Fault Rate",
                    "Rate of pages read that were in the process of being written back to disk when faulted"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]System faults"),
                    "System Fault Rate",
                    "Rate of page faults for pages in system space"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Free List Size"),
                    "Free List Size",
                    "Number of pages on the free-page list"
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.PAGE\\]Modified List Size"),
                    "Modified List Size",
                    "Number of pages on the modified-page list"
                    ),
                    
                    // RLOCK: Dynamic lock remastering statistics
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Lock trees moved"),
                    "Lock Tree Outbound Rate",
                    "Rate at which lock trees are moved from this node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Lock trees moved \\(higher activity\\)"),
                    "Higher Activity",
                    "Rate for trees moved due to higher locking activity on another node in the cluster."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Lock trees moved \\(LOCKDIRWT\\)"),
                    "Higher LOCKDIRWT",
                    "Rate at which trees are moved to a node with a higher value of the SYSGEN parameter LOCKDIRWT."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Lock trees moved \\(Single\\)"),
                    "Sole Interest",
                    "Rate at which trees are moved to another node because that node is the only one with locks remaining on the tree."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Remaster messages sent"),
                    "Remaster Msg Send Rate",
                    "Rate at which remaster messages are sent from this node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Lock trees \\(Incoming\\)"),
                    "Lock Tree Inbound Rate",
                    "Rate at which trees are moved to this node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.RLCK\\]Remaster messages received"),
                    "Remaster Msg Receive Rate",
                    "Rate at which remaster messages are received on this node."
                    ),
                    
                    // SCS: System Communications Services statistics
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Dg sent to .*"),
                    "Datagram Send Rate",
                    "Rate at which datagrams are sent to another node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Dg recv from .*"),
                    "Datagram Receive Rate",
                    "Rate at which datagrams are received from another node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Dg discarded from .*"),
                    "Datagram Discard Rate",
                    "Rate at which datagrams are discarded."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Msg sent to .*"),
                    "Message Send Rate",
                    "Rate at which sequenced messages are sent to another node. Sequenced messages are exchanged between nodes to communicate with mass storage control protocol (MSCP) disks and the lock manager."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Msg recv from .*"),
                    "Message Receive Rate",
                    "Rate at which sequenced messages are received from another node. Sequenced messages are exchanged between nodes to communicate with MSCP disks and the lock manager."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Send-Data commands sent to .*"),
                    "Send Data Rate",
                    "Rate at which block send datas are initiated on the local node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]SD KBytes sent to .*"),
                    "Kbytes Send Rate",
                    "Rate at which kilobytes are sent, as a result of send datas initiated on the local node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Req-Data commands sent to .*"),
                    "Request Data Rate",
                    "Rate at which request datas are initiated on the local node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]RD KBytes recv from .*"),
                    "Kbytes Request Rate",
                    "Rate at which kilobytes are received, as a result of request datas initiated on the local node."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Block Xfers KB mapped to .*"),
                    "Kbytes Map Rate",
                    "Rate at which kilobytes are mapped for block transfers. This is a rough measure of the data transfer rate between the local node and a remote node. Before any transfer can take place, a buffer must be mapped. The size of the accumulated buffers that were mapped is displayed by the Kbytes Map Rate. If request datas or send datas are initiated on the local or a remote node, then the Kbytes Map Rate reflects the number of kilobytes actually transferred between the two nodes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Send Credits waits to .*"),
                    "Send Credit Queued Rate",
                    "Rate at which connections are queued for a send credit. A connection is queued for a send credit whenever all of the buffers that were allocated by the remote node have been used."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SCS\\]Buffer Descr waits to .*"),
                    "Buffer Descriptor Queued Rate",
                    "Rate at which connections are queued for a buffer descriptor. A connection is queued for a buffer descriptor whenever all of the buffer descriptors have been allocated by the local node."
                    ),
                    
                    // STATES: Number of processes in each of the scheduler states
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]ColPg"),
                    "Collided Page Wait (COLPG)",
                    "Waiting for a faulted page in transition."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Mwait"),
                    "Mutex & Miscellaneous Resource Wait (MWAIT)",
                    "Waiting for the availability of a mutual exclusion semaphore or a dynamic resource."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Cef"),
                    "Common Event Flag Wait (CEF)",
                    "Waiting for a combination of event flags to be set in a common event block."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Pfw"),
                    "Page Fault Wait (PFW)",
                    "Waiting for a page to be read as a result of a page fault; resident processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Lef"),
                    "Local Event Flag Wait (LEF)",
                    "Waiting for one or more local event flags to be posted; resident processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]LefO"),
                    "Local Event Flag (Outswapped) (LEFO)",
                    "Waiting for one or more local event flags to be posted; outswapped processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Hib"),
                    "Hibernate (HIB)",
                    "Hibernating, or process has executed a hibernate request; resident processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]HibO"),
                    "Hibernate (Outswapped) (HIBO)",
                    "Hibernating, or process has executed a hibernate request; outswapped processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Susp"),
                    "Suspended (SUSP)",
                    "Process has executed a suspend request; resident processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]SuspO"),
                    "Suspended (Outswapped) (SUSPO)",
                    "Process has executed a suspend request; outswapped processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Fpg"),
                    "Free Page Wait (FPW)",
                    "Waiting for a free page of memory."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Compute"),
                    "Compute (COM)",
                    "Ready to use the processor; resident processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Como"),
                    "Compute (Outswapped) (COMO)",
                    "Ready to use the processor; outswapped processes."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.STAT\\]Cur"),
                    "Current Process (CUR)",
                    "Using the processor."
                    ),
                    
                    // SYSTEM: Summary of statistics from other classes
                    new Description(
                    Pattern.compile("\\[MON\\.SYST\\]Process Count"),
                    "Process Count",
                    ""
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SYST\\]Page Fault Rate"),
                    "Page Fault Rate",
                    ""
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SYST\\]Page Read I/O rate"),
                    "Page Read I/O Rate",
                    ""
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SYST\\]Direct I/O Rate"),
                    "Direct I/O Rate",
                    ""
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.SYST\\]Buffered I/O Rate"),
                    "Buffered I/O Rate",
                    ""
                    ),
                    
                    // TIMER: Timer Queue Entry (TQE) statistics
                    new Description(
                    Pattern.compile("\\[MON\\.TIMR\\]TQE Total"),
                    "Total TQE Rate",
                    "Total Rate of TQEs processed per second. This statistic is a combined total of the three following TQE rates."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.TIMR\\]TQE SySubs"),
                    "SYSUB TQE Rate",
                    "Rate of SYSUB TQEs processed per second. These system subroutine TQEs represent timer requests made by the OpenVMS operating system."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.TIMR\\]TQE Timers"),
                    "Timer TQE Rate",
                    "Rate of Timer TQEs processed per second. These TQEs represent timer requests made by users through the $SETIMR system service."
                    ),
                    new Description(
                    Pattern.compile("\\[MON\\.TIMR\\]TQE wakeups"),
                    "Wakeup TQE Rate",
                    "Rate of wakeup TQEs processed per second. These TQEs represent timer requests made by users through the $SCHDWK system service."
                    ),
        };
    }
}
