/*
 * AddItemDialog.java
 */

package de.pdv_systeme.JTLviz;

import java.util.Collection;
import java.util.Iterator;

import javax.swing.DefaultListModel;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import de.pdv_systeme.DataCollection.Operation;

/**
 * Select a calculated item to add.
 *
 * @author  <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class AddItemDialog extends JDialog {
    
    /** Action command string and parameter to {@link #getSourceItem}. */
    public static final String SOURCE_ITEM1 = "Source #1";
    public static final String SOURCE_ITEM2 = "Source #2";
    
    /** Parent frame. */
    private JFrame parent;
    
    /** The list model. */
    private DefaultListModel itemListModel;
    
    /** Set to true by the submit button iff all fields have been filled */
    private boolean validItem;
    
    /**
     * Creates new form AddItemDialog
     * @param parent the dialog's parent frame
     * @param items the item list to display
     */
    public AddItemDialog(final JFrame parent, final Collection items) {
        super(parent, true);
        this.parent = parent;
        this.itemListModel = new DefaultListModel();
        this.itemListModel.ensureCapacity(items.size());
        for (Iterator i = items.iterator(); i.hasNext(); )
            this.itemListModel.addElement(i.next().toString());
        this.validItem = false;
        initComponents();
        itemList.setModel(itemListModel);
    }
    
    /**
     * Get the value of a source item
     * @param item the source item tag. One of {@link #SOURCE_ITEM1}, {@link #SOURCE_ITEM2}.
     * @return the source item's value (String, Double, or <code>null</code>)
     */
    public Object getSourceItem(final String item) {
        String contents;
        if (item == SOURCE_ITEM1)
            contents = source1TextField.getText();
        else // item == SOURCE_ITEM2
            contents = source2TextField.getText();
        if (contents.equals("") || itemListModel.contains(contents))
            return contents;
        Double num;
        try {
            num = Double.valueOf(contents);
        } catch (NumberFormatException e) {
            return null;
        }
        return num;
    }
    
    /**
     * Get the selected operation.
     * @return the operation
     * @see Operation
     */
    public Operation getOperation() {
        if (opAddRadioButton.isSelected())
            return Operation.ADD;
        else if (opSubRadioButton.isSelected())
            return Operation.SUBTRACT;
        else if (opMulRadioButton.isSelected())
            return Operation.MULTIPLY;
        else if (opDivRadioButton.isSelected())
            return Operation.DIVIDE;
        else
            return Operation.NONE;
    }
    
    /**
     * Get the the new item name.
     * @return the item name
     */
    public String getNewItemName() {
        return nameTextField.getText();
    }
    
    /**
     * Get the status of the Add Item dialog.
     * @return true iff all fields have been filled
     */
    public boolean isValidItem() {
        return validItem;
    }
    
    /**
     * Check the status if the dialog
     * @return true iff all fields have been filled
     */
    private boolean checkValidItem() {
        return !(
                source1TextField.getText().equals("") ||
                source2TextField.getText().equals("") ||
                getOperation() == Operation.NONE ||
                nameTextField.getText().equals("") ||
                itemListModel.contains(nameTextField.getText())
                );
    }
    
    /**
     * Show a message dialog.
     * @param s  the resource string
     */
    private void showMessage(final String s) {
        JOptionPane.showMessageDialog(
                parent,
                Main.getResourceString(s),
                Main.getResourceString("APP_NAME"),
                JOptionPane.PLAIN_MESSAGE
                );
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        java.awt.GridBagConstraints gridBagConstraints;
        javax.swing.JPanel jPanel1;
        javax.swing.JPanel jPanel2;
        javax.swing.JScrollPane jScrollPane1;
        javax.swing.JLabel nameLabel;

        opButtonGroup = new javax.swing.ButtonGroup();
        jScrollPane1 = new javax.swing.JScrollPane();
        itemList = new javax.swing.JList();
        jPanel1 = new javax.swing.JPanel();
        source1Button = new javax.swing.JButton();
        source1TextField = new javax.swing.JTextField();
        jPanel2 = new javax.swing.JPanel();
        opAddRadioButton = new javax.swing.JRadioButton();
        opSubRadioButton = new javax.swing.JRadioButton();
        opMulRadioButton = new javax.swing.JRadioButton();
        opDivRadioButton = new javax.swing.JRadioButton();
        source2Button = new javax.swing.JButton();
        source2TextField = new javax.swing.JTextField();
        nameLabel = new javax.swing.JLabel();
        nameTextField = new javax.swing.JTextField();
        submitButton = new javax.swing.JButton();

        itemList.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        jScrollPane1.setViewportView(itemList);

        getContentPane().add(jScrollPane1, java.awt.BorderLayout.CENTER);

        jPanel1.setLayout(new java.awt.GridBagLayout());

        source1Button.setText(Main.getResourceString("Select_Source_Item_#1_from_ListBox_on_Left"));
        source1Button.setActionCommand(SOURCE_ITEM1);
        source1Button.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sourceButtonActionPerformed(evt);
            }
        });

        jPanel1.add(source1Button, new java.awt.GridBagConstraints());

        source1TextField.setDisabledTextColor(new java.awt.Color(102, 102, 102));
        source1TextField.setPreferredSize(source1Button.getPreferredSize());
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        jPanel1.add(source1TextField, gridBagConstraints);

        opButtonGroup.add(opAddRadioButton);
        opAddRadioButton.setSelected(true);
        opAddRadioButton.setText(Main.getResourceString("Add"));
        jPanel2.add(opAddRadioButton);

        opButtonGroup.add(opSubRadioButton);
        opSubRadioButton.setText(Main.getResourceString("Subtract"));
        jPanel2.add(opSubRadioButton);

        opButtonGroup.add(opMulRadioButton);
        opMulRadioButton.setText(Main.getResourceString("Multiply"));
        jPanel2.add(opMulRadioButton);

        opButtonGroup.add(opDivRadioButton);
        opDivRadioButton.setText(Main.getResourceString("Divide"));
        jPanel2.add(opDivRadioButton);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        jPanel1.add(jPanel2, gridBagConstraints);

        source2Button.setText(Main.getResourceString("Select_Source_Item_#2_from_ListBox_on_Left"));
        source2Button.setActionCommand(SOURCE_ITEM2);
        source2Button.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sourceButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 3;
        jPanel1.add(source2Button, gridBagConstraints);

        source2TextField.setDisabledTextColor(new java.awt.Color(102, 102, 102));
        source2TextField.setPreferredSize(source1Button.getPreferredSize());
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 4;
        jPanel1.add(source2TextField, gridBagConstraints);

        nameLabel.setText(Main.getResourceString("Name_of_new_Item"));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 5;
        jPanel1.add(nameLabel, gridBagConstraints);

        nameTextField.setPreferredSize(source1Button.getPreferredSize());
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 6;
        jPanel1.add(nameTextField, gridBagConstraints);

        submitButton.setText(Main.getResourceString("Do_It_!"));
        submitButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                submitButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 7;
        jPanel1.add(submitButton, gridBagConstraints);

        getContentPane().add(jPanel1, java.awt.BorderLayout.EAST);

        pack();
    }
    // </editor-fold>//GEN-END:initComponents
    
    private void submitButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_submitButtonActionPerformed
        if (checkValidItem()) {
            if (getSourceItem(SOURCE_ITEM1) == null) {
                showMessage("Input_not_numeric_in_first_field");
            } else if (getSourceItem(SOURCE_ITEM2) == null) {
                showMessage("Input_not_numeric_in_second_field");
            } else {
                hide();
                validItem = true;
            }
        } else {
            showMessage("Please_enter_all_required_fields");
        }
    }//GEN-LAST:event_submitButtonActionPerformed
    
    private void sourceButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sourceButtonActionPerformed
        int selectedIndex = itemList.getSelectedIndex();
        if (selectedIndex != -1) {
            String name = (String) itemListModel.getElementAt(selectedIndex);
            if (evt.getActionCommand() == SOURCE_ITEM1) {
                source1TextField.setText(name);
            } else { // evt.getActionCommand() == SOURCE_ITEM2
                source2TextField.setText(name);
            }
        }
    }//GEN-LAST:event_sourceButtonActionPerformed
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JList itemList;
    private javax.swing.JTextField nameTextField;
    private javax.swing.JRadioButton opAddRadioButton;
    private javax.swing.ButtonGroup opButtonGroup;
    private javax.swing.JRadioButton opDivRadioButton;
    private javax.swing.JRadioButton opMulRadioButton;
    private javax.swing.JRadioButton opSubRadioButton;
    private javax.swing.JButton source1Button;
    private javax.swing.JTextField source1TextField;
    private javax.swing.JButton source2Button;
    private javax.swing.JTextField source2TextField;
    private javax.swing.JButton submitButton;
    // End of variables declaration//GEN-END:variables
    
}
