/*
 * ChartPropertiesDataTable.java
 */

package de.pdv_systeme.JTLviz;

import java.util.Date;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import java.awt.Color;
import java.awt.Rectangle;

import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.JTableHeader;

import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.event.ChartChangeEventType;
import org.jfree.chart.event.PlotChangeEvent;
import org.jfree.chart.event.PlotChangeListener;
import org.jfree.chart.plot.CombinedDomainXYPlot;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;

import de.pdv_systeme.DataCollection.AbstractDataCollection;
import de.pdv_systeme.util.VmsDate;

/**
 * The data table in the chart properties dialog.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class ChartPropertiesDataTable extends JTable {
    
    /**
     * Creates a new instance of ChartPropertiesDataTable
     *
     * @param plot the plot data to be tabulated
     * @see ChartPropertiesDataTableModel
     */
    public ChartPropertiesDataTable(final XYPlot plot) {
        super();
        
        setModel(new ChartPropertiesDataTableModel(plot));
        setAutoResizeMode(AUTO_RESIZE_OFF);
        
        JTableHeader dataTableHeader = getTableHeader();
        dataTableHeader.setReorderingAllowed(false);
        Color headerBg = dataTableHeader.getBackground();
        getColumnModel().getColumn(0).setCellRenderer(new NumberColumnRenderer(headerBg));
        setDefaultRenderer(Date.class, new MyDateRenderer());
        
        setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        setSelectedRow(0);
        
        // TODO: setPreferredWidth of data table columns to longest entry
    }
    
    /**
     * Set the selected row
     * @param row the row to select
     */
    public void setSelectedRow(int row) {
        getSelectionModel().setSelectionInterval(row, row);
        // Scroll the table to make the selected row visible
        int h = getRowHeight(row);
        scrollRectToVisible(new Rectangle(0, row*h, 10, h));
    }
    
    /*
     * Table cell renderers
     */
    /** Table cell renderer for column #0 */
    private static class NumberColumnRenderer extends DefaultTableCellRenderer {
        /** Create a new instance of NumberColumnRenderer */
        public NumberColumnRenderer(final Color bg) {
            super();
            setHorizontalAlignment(JLabel.RIGHT);
            setBackground(bg);
        }
    }
    
    /** Table cell renderer for Date columns */
    private static class MyDateRenderer extends DefaultTableCellRenderer {
        /** {@inheritDoc} */
        public void setValue(final Object value) {
            if (value == null)
                setText("");
            else
                setText(VmsDate.toString(value, AbstractDataCollection.PRECISION));
        }
    }
    
    /**
     * Table model and renderer for the chart properties data table
     */
    private static class ChartPropertiesDataTableModel
            extends AbstractTableModel
            implements PlotChangeListener {
        
        /** The associated plot. */
        private XYPlot plot;
        
        /** The datasets. */
        private List datasets; // of TimeSeries
        
        /** The maximum row count. */
        private int maxRowCount;
        
        /**
         * Creates a new instance of ChartPropertiesDataTableModel
         * @param plot the associated plot
         */
        public ChartPropertiesDataTableModel(final XYPlot plot) {
            if (plot == null)
                throw new IllegalArgumentException("Null 'plot' argument");
            this.plot = plot;
            this.plot.addChangeListener(this);
            datasets = new ArrayList();
            initLocalData();
        }
        
    /*
     * PlotChangeListener methods
     */
        /** {@inheritDoc} */
        public void plotChanged(final PlotChangeEvent event) {
            if (event.getType() == ChartChangeEventType.DATASET_UPDATED) {
                initLocalData();
                fireTableChanged(new TableModelEvent(this));
            }
        }
        
        /*
         * TableModel methods
         */
        /** {@inheritDoc} */
        public int getRowCount() {
            return maxRowCount;
        }
        
        /** {@inheritDoc} */
        public int getColumnCount() {
            return 2 + 2 * datasets.size();
        }
        
        /**
         * Check the row index.
         * @param rowIndex the row index
         * @return the row index
         * @throws IndexOutOfBoundsException if the row index is out of range
         */
        private int boundRow(int rowIndex) {
            if (rowIndex < 0 || rowIndex >= getRowCount()) {
                throw new IndexOutOfBoundsException("Row index out of range");
            }
            return rowIndex;
        }
        
        /**
         * Check the column index.
         * @param columnIndex the column index
         * @return the column index
         * @throws IndexOutOfBoundsException if the column index is out of range
         */
        private int boundColumn(int columnIndex) {
            if (columnIndex < 0 || columnIndex >= getColumnCount()) {
                throw new IndexOutOfBoundsException("Column index out of range");
            }
            return columnIndex;
        }
        
        /**
         * Return the time series for a columnIndex.
         * @param columnIndex  the column number
         * @return the time series
         */
        private TimeSeries getTimeSeries(final int columnIndex) {
            return (TimeSeries) datasets.get((columnIndex - 2) >> 1);
        }
        
        /** {@inheritDoc} */
        public Class getColumnClass(final int columnIndex) {
            if (boundColumn(columnIndex) == 0) {
                return Integer.class;
            } else if (columnIndex == 1) {
                return String.class;
            } else if (columnIndex % 2 == 0) {
                return Date.class;
            } else { // columnIndex > 1 && columnIndex % 2 != 0
                return Double.class;
            }
        }
        
        /** {@inheritDoc} */
        public String getColumnName(final int columnIndex) {
            if (boundColumn(columnIndex) == 0) {
                return Main.getResourceString("Number");
            } else if (columnIndex == 1) {
                return Main.getResourceString("Text");
            } else {
                TimeSeries ts = getTimeSeries(columnIndex);
                if (columnIndex % 2 == 0) {
                    return Main.getResourceString("Timestamp"); // "X";
                } else { // columnIndex % 2 != 0
                    return (String) ts.getKey(); // "Y";
                }
            }
        }
        
        /** {@inheritDoc} */
        public Object getValueAt(final int rowIndex, final int columnIndex) {
            if (boundColumn(columnIndex) == 0) {
                return new Integer(boundRow(rowIndex));
            } else if (columnIndex == 1) {
                return "";
            } else {
                TimeSeries ts = getTimeSeries(columnIndex);
                if (boundRow(rowIndex) >= ts.getItemCount()) {
                    return null;
                } else if (columnIndex % 2 == 0)
                    return ts.getTimePeriod(rowIndex).getStart();
                else // columnIndex % 2 != 0
                    return ts.getValue(rowIndex);
            }
        }
        
        /** {@inheritDoc} */
        public void setValueAt(final Object aValue, final int rowIndex, final int columnIndex) {
            TimeSeries ts = getTimeSeries(boundColumn(columnIndex));
            if (boundRow(rowIndex) >= ts.getItemCount()) {
                throw new IndexOutOfBoundsException();
            }
            
            // TODO: check format and set value
            
            // Notify listeners
            fireTableCellUpdated(rowIndex, columnIndex);
        }
        
        /** {@inheritDoc} */
        public boolean isCellEditable(final int rowIndex, final int columnIndex) {
            return (boundColumn(columnIndex) >= 1);
        }
        
        /** Initialize local data. */
        private void initLocalData() {
            datasets.clear();
            if (plot instanceof CombinedDomainXYPlot) {
                CombinedDomainXYPlot cdplot = (CombinedDomainXYPlot) plot;
                for (Iterator pi = cdplot.getSubplots().iterator(); pi.hasNext(); ) {
                    XYPlot xyplot = (XYPlot) pi.next();
                    for (int i = 0; i < xyplot.getDatasetCount(); i++) {
                        TimeSeriesCollection tsc = (TimeSeriesCollection) xyplot.getDataset(i);
                        datasets.addAll(tsc.getSeries());
                    }
                }
            } else { // plot instanceof XYPlot
                for (int i = 0; i < plot.getDatasetCount(); i++) {
                    TimeSeriesCollection tsc = (TimeSeriesCollection) plot.getDataset(i);
                    datasets.addAll(tsc.getSeries());
                }
            }
            maxRowCount = 0;
            for (Iterator di = datasets.iterator(); di.hasNext(); ) {
                TimeSeries ts = (TimeSeries) di.next();
                if (maxRowCount < ts.getItemCount()) {
                    maxRowCount = ts.getItemCount();
                }
            }
        }
    }
    
}
