/*
 * CorrelationDialog.java
 */

package de.pdv_systeme.JTLviz;

import java.io.FileWriter;
import java.io.IOException;
import java.text.MessageFormat;
import java.text.DecimalFormat;
import java.util.Comparator;
import java.util.Date;

import java.awt.Color;
import java.awt.Font;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.data.Range;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesDataItem;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.ExtensionFileFilter;
import org.jfree.util.StringUtils;

import de.pdv_systeme.DataCollection.AbstractDataCollection;
import de.pdv_systeme.DataCollection.DataCollection;
import de.pdv_systeme.TSChart.TSChartProperty;
import de.pdv_systeme.util.SortedUniqueListModel;
import de.pdv_systeme.util.VmsDate;

/**
 * Displays a (non-modal) correlation chart dialog
 *
 * @author  Martin Vorlnder <mv@pdv-systeme.de>
 */
public class CorrelationDialog
        extends JDialog {
    
    /** The format for the correlation coefficient */
    private final static DecimalFormat corrChartFormat = new DecimalFormat("0.000;-0.000");
    
    /** list prototype element */
    private final static String REFERENCE_ELEMENT =   Main.getResourceString("CorrelationDialog.REFERENCE_ELEMENT");
    
    /** Format for the correlation chart title */
    private final static MessageFormat titleFormat = new MessageFormat(
            Main.getResourceString("CorrelationDialog.titleFormat")
            );
    
    /** Format for the correlation list title */
    private final static MessageFormat headlineFormat = new MessageFormat(
            Main.getResourceString("CorrelationDialog.headlineFormat")
            );
    
    /** Format for the correlation text file title */
    private final static MessageFormat fileTitleFormat = new MessageFormat(
            Main.getResourceString("CorrelationDialog.fileTitleFormat")
            );
    
    private JFrame parent;
    private SortedUniqueListModel itemListModel;
    private String itemName;
    private DataCollection dc;
    private TimeSeries ySeries;
    private int yItemCount;
    
    /** Nested class for the list model of the correlation dialog list */
    private static class ListModelValue implements Comparable {
        
        /** the format for the list correlation values */
        public final static DecimalFormat corrListFormat = new DecimalFormat("0.000;(0.000)");
        
        /** the correlation coefficient */
        double r;
        /** the column name */
        String name;
        
        /** Create a new instance */
        ListModelValue(final double r, final String name) {
            this.r = r;
            this.name = name;
        }
        
        /** {@inheritDoc} */
        public String toString() {
            return corrListFormat.format(r)+  "  "+name;
        }
        
        /** {@inheritDoc} */
        public boolean equals(final Object o) {
            return
                    (o instanceof ListModelValue)
                    ? r == ((ListModelValue) o).r && name.equals(((ListModelValue) o).name)
                    : false;
        }
        
        /** {@inheritDoc} */
        public int hashCode() {
            return name.hashCode()+(int)(r*1000);
        }
        
        /** {@inheritDoc} */
        public int compareTo(final Object o)
        throws ClassCastException {
            ListModelValue lo = (ListModelValue) o;
            return (Math.abs(r) < Math.abs(lo.r))
                ? -1
                    : (Math.abs(r) > Math.abs(lo.r))
                    ? +1 
                    : name.compareTo(lo.name);
        }
    }
    
    /** Nested class that provides a comparator for ListModelValues. */
    private static class ListModelValueComparator implements Comparator {
        public int compare(Object obj1, Object obj2) {
            return ((ListModelValue) obj1).compareTo(obj2);
        }
    }
    
    /**
     * Create an instance of CorrelationDialog.
     * @param parent  the JFrame object parenting the dialog
     * @param itemName  the base item against which collections are calculated
     * @param dc  the data collection
     */
    public CorrelationDialog(
            final JFrame parent,
            final String itemName,
            final DataCollection dc) {
        super(parent, false);
        this.parent = parent;
        this.itemName = itemName;
        this.dc = dc;
        this.ySeries = dc.getData(itemName);
        this.yItemCount = ySeries.getItemCount();
        this.itemListModel = new SortedUniqueListModel(new ListModelValueComparator());
        initComponents();
        ((ChartPanel) correlationChartPanel).setPopupMenu(null);
    }
    
    /**
     * Insert an item into the list.
     * @param corr  the correlation coefficient
     * @param name  the item name
     */
    public void insert(final double corr, final String name) {
        itemListModel.add(new ListModelValue(corr, name));
    }
    
    /**
     * Signal that the list is complete.
     * This either pops up a message box (in case there are no correlations),
     * or selects the list item with the highest correlation (which in turn
     * triggers the painting of the scatter diagram).
     */
    public void signalListComplete() {
        int s = itemListModel.getSize()-1;
        if (s != -1) {
            correlationList.setSelectedIndex(s);
            correlationList.ensureIndexIsVisible(s);
        } else {
            JFrame parent = (JFrame) getParent();
            dispose();
            JOptionPane.showMessageDialog(
                    parent,
                    Main.getResourceString("No_relevant_items_found_..."),
                    Main.getResourceString("APP_NAME"),
                    JOptionPane.PLAIN_MESSAGE
                    );
        }
    }
    
    /**
     * Generate the scatter plot
     * @param xSeries  time series to plot
     * @param corrValue  the correlation
     * @return the scatter plot JFreeChart
     */
    private JFreeChart generateScatterPlot(
            final TimeSeries xSeries,
            final double corrValue
            ) {
        String xName = (String) xSeries.getKey();
        XYSeries s = new XYSeries(xName+ " / "+itemName);
        for (int i = 0; i < yItemCount; i++) {
            TimeSeriesDataItem yDataItem = ySeries.getDataItem(i);
            Number xValue = xSeries.getValue(yDataItem.getPeriod());
            if (xValue != null)
                s.add(new XYDataItem(xValue, yDataItem.getValue()), false);
        }
        Date fromTime = ySeries.getTimePeriod(0).getStart();
        Date toTime = ySeries.getTimePeriod(yItemCount-1).getStart();
        String title = titleFormat.format( new Object[] {
            dc.getNodename(),
                    VmsDate.toString(fromTime, AbstractDataCollection.PRECISION),
                    VmsDate.toString(toTime, AbstractDataCollection.PRECISION)
        });
        String[] titles = title.split(Main.NEWLINE);
        JFreeChart chart = ChartFactory.createScatterPlot(
                titles[0],                  // title
                xName,                      // xAxisLabel
                itemName,                 // yAxisLabel
                new XYSeriesCollection(s),  // dataset
                PlotOrientation.VERTICAL,   // orientation
                false,                      // legend
                true,                       // tooltips
                false                       // urls
                );
        chart.setAntiAlias(false);
        Color titleColor = TSChartProperty.chartTitleColor;
        chart.getTitle().setPaint(titleColor);
        Font titleFont = chart.getTitle().getFont();
        for (int i = 1; i < titles.length; ++i) {
            TextTitle subTitle = new TextTitle(titles[i], titleFont);
            subTitle.setPaint(titleColor);
            chart.addSubtitle(subTitle);
        }
        chart.setBackgroundPaint(TSChartProperty.chartBackgroundColor);
        
        XYPlot plot = (XYPlot) chart.getPlot();
        plot.setBackgroundPaint(TSChartProperty.plotBackgroundColor);
        plot.setDomainGridlinePaint(TSChartProperty.plotGridlineColor);
        plot.setRangeGridlinePaint(TSChartProperty.plotGridlineColor);
        
        // insert correlation value annotaion
        Range xr = plot.getDomainAxis().getRange();
        Range yr = plot.getRangeAxis().getRange();
        double x =
                (corrValue < 0)
                ? xr.getUpperBound() - xr.getLength() / 10
                : xr.getLowerBound() + xr.getLength() / 10;
        double y = yr.getLowerBound() + yr.getLength() * 9/10;
        XYTextAnnotation corrValAnnot = new XYTextAnnotation(
                Main.getFormattedResourceString("CorrelationDialog.r=",
                new Object[] {
            new Double(corrValue)
        }), x, y);
        corrValAnnot.setPaint(TSChartProperty.plotSeriesColor[0]);
        plot.addAnnotation(corrValAnnot);
        
        return chart;
    }
    
    /**
     * Generate a list of correlation values from the list model as a string
     * @return the string
     */
    private String getCorrelationListAsString() {
        Date fromTime = ySeries.getTimePeriod(0).getStart();
        Date toTime = ySeries.getTimePeriod(yItemCount-1).getStart();
        StringBuffer sb = new StringBuffer();
        sb.append(fileTitleFormat.format( new Object[] {
            itemName,
                    dc.getNodename(),
                    VmsDate.toString(fromTime, AbstractDataCollection.PRECISION),
                    VmsDate.toString(toTime, AbstractDataCollection.PRECISION)
        }));
        for (int i = itemListModel.getSize()-1; i >= 0; i--) {
            sb.append(((ListModelValue)itemListModel.getElementAt(i)).toString());
            sb.append(Main.NEWLINE);
        }
        return sb.toString();
    }
    
    /**
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        javax.swing.JPanel buttonPanel;
        javax.swing.JScrollPane correlationListScrollPane;
        java.awt.GridBagConstraints gridBagConstraints;
        javax.swing.JPanel leftPanel;

        leftPanel = new javax.swing.JPanel();
        headlineTextArea = new javax.swing.JTextArea();
        correlationListScrollPane = new javax.swing.JScrollPane();
        correlationList = new javax.swing.JList();
        buttonPanel = new javax.swing.JPanel();
        saveCorrelationListButton = new javax.swing.JButton();
        printCorrelationListButton = new javax.swing.JButton();
        exportChartButton = new javax.swing.JButton();
        printChartButton = new javax.swing.JButton();
        okButton = new javax.swing.JButton();
        correlationChartPanel = new ChartPanel(null, false);

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle(Main.getResourceString("T4_Correlation_Coefficients"));
        leftPanel.setLayout(new java.awt.BorderLayout());

        headlineTextArea.setEditable(false);
        headlineTextArea.setFont(new java.awt.Font("SansSerif", 1, 11));
        headlineTextArea.setLineWrap(true);
        headlineTextArea.setText(headlineFormat.format(new Object[] {itemName}));
        headlineTextArea.setWrapStyleWord(true);
        leftPanel.add(headlineTextArea, java.awt.BorderLayout.NORTH);

        correlationListScrollPane.setHorizontalScrollBarPolicy(javax.swing.JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        correlationList.setModel(itemListModel);
        correlationList.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        correlationList.setPrototypeCellValue(REFERENCE_ELEMENT);
        correlationList.addListSelectionListener(new javax.swing.event.ListSelectionListener() {
            public void valueChanged(javax.swing.event.ListSelectionEvent evt) {
                correlationListValueChanged(evt);
            }
        });

        correlationListScrollPane.setViewportView(correlationList);

        leftPanel.add(correlationListScrollPane, java.awt.BorderLayout.CENTER);

        buttonPanel.setLayout(new java.awt.GridBagLayout());

        saveCorrelationListButton.setText(Main.getResourceString("Save_Correlation_List"));
        saveCorrelationListButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveCorrelationListButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        buttonPanel.add(saveCorrelationListButton, gridBagConstraints);

        printCorrelationListButton.setText(Main.getResourceString("Print_Correlation_List"));
        printCorrelationListButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                printCorrelationListButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        buttonPanel.add(printCorrelationListButton, gridBagConstraints);

        exportChartButton.setText(Main.getResourceString("Export_Chart"));
        exportChartButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exportChartButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 0;
        buttonPanel.add(exportChartButton, gridBagConstraints);

        printChartButton.setText(Main.getResourceString("Print_Chart"));
        printChartButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                printChartButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        buttonPanel.add(printChartButton, gridBagConstraints);

        okButton.setText(Main.getResourceString("OK"));
        okButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                okButtonActionPerformed(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
        buttonPanel.add(okButton, gridBagConstraints);

        leftPanel.add(buttonPanel, java.awt.BorderLayout.SOUTH);

        getContentPane().add(leftPanel, java.awt.BorderLayout.WEST);

        correlationChartPanel.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        getContentPane().add(correlationChartPanel, java.awt.BorderLayout.CENTER);

        pack();
    }
    // </editor-fold>//GEN-END:initComponents
    
    private void printCorrelationListButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_printCorrelationListButtonActionPerformed
        // TODO: make printing a correlation list work
        Main.showNYI(this.parent);
        String data = getCorrelationListAsString();

        // Tips I found for printing text (Programmer FAQ 7.4 & 8.2):
        // print the list to a file and print the file
        // or open the printer as a file
    }//GEN-LAST:event_printCorrelationListButtonActionPerformed
    
    private void saveCorrelationListButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveCorrelationListButtonActionPerformed
        // Show file chooser.
        JFileChooser fileChooser = new JFileChooser();
        String extension = Main.getResourceString(".txt");
        ExtensionFileFilter filter = new ExtensionFileFilter(
                Main.getResourceString("Text_files_(*.TXT)"),
                extension);
        fileChooser.addChoosableFileFilter(filter);
        
        int result = fileChooser.showSaveDialog(this);
        
        // Process the results.
        if (result == fileChooser.APPROVE_OPTION) {
            String filename = fileChooser.getSelectedFile().getPath();
            // enforce .txt extension
            if (! StringUtils.endsWithIgnoreCase(filename, extension))
                filename = filename + extension;
            try {
                FileWriter out = new FileWriter(filename);
                out.write(getCorrelationListAsString());
                out.close();
            } catch (IOException e) {
                Main.reportError(parent, e, Main.getResourceString("I/O_Error"));
            }
        }
    }//GEN-LAST:event_saveCorrelationListButtonActionPerformed
    
    private void printChartButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_printChartButtonActionPerformed
        ((ChartPanel) correlationChartPanel).createChartPrintJob();
    }//GEN-LAST:event_printChartButtonActionPerformed
    
    private void exportChartButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exportChartButtonActionPerformed
        try {
            ((ChartPanel) correlationChartPanel).doSaveAs();
        } catch (java.io.IOException e) {
            Main.reportError(parent, e, Main.getResourceString("I/O_Error"));
        }
    }//GEN-LAST:event_exportChartButtonActionPerformed
    
    private void correlationListValueChanged(javax.swing.event.ListSelectionEvent evt) {//GEN-FIRST:event_correlationListValueChanged
        if (! evt.getValueIsAdjusting()) {
            JFreeChart chart;
            int selectedItem = correlationList.getSelectedIndex();
            if (selectedItem == -1) {
                chart = null;
            } else {
                ListModelValue val = (ListModelValue) itemListModel.getElementAt(selectedItem);
                chart = generateScatterPlot(dc.getData(val.name), val.r);
            }
            
            ((ChartPanel) correlationChartPanel).setChart(chart);
        }
    }//GEN-LAST:event_correlationListValueChanged
    
    private void okButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_okButtonActionPerformed
        dispose();
    }//GEN-LAST:event_okButtonActionPerformed
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JPanel correlationChartPanel;
    private javax.swing.JList correlationList;
    private javax.swing.JButton exportChartButton;
    private javax.swing.JTextArea headlineTextArea;
    private javax.swing.JButton okButton;
    private javax.swing.JButton printChartButton;
    private javax.swing.JButton printCorrelationListButton;
    private javax.swing.JButton saveCorrelationListButton;
    // End of variables declaration//GEN-END:variables
    
}
