/*
 * ItemListModel.java
 */

package de.pdv_systeme.JTLviz;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;

import de.pdv_systeme.DataCollection.DataCollection;
import de.pdv_systeme.util.SortedUniqueListModel;

/**
 * The performance data item list model
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class ItemListModel extends SortedUniqueListModel {
    
    /** The default sort style */
    private static final ItemListSortStyle
            DEFAULT_SORTSTYLE = ItemListSortStyle.ALPHABETICALLY;
    
    /** The default sort direction */
    private static final ItemListSortDirection
            DEFAULT_SORTDIRECTION = ItemListSortDirection.ASCENDING;
    
    /** A comparator for ItemListElements */
    private static class ItemListComparator implements Comparator {
        /** The sort style */
        private ItemListSortStyle sortStyle;
        
        /** The sort direction */
        private ItemListSortDirection sortDirection;
        
        /** The start position for alphabetical sort */
        private int startPosition;
        
        /**
         * Create a new instance of ItemListComparator.
         * @param sortStyle the sort order
         */
        public ItemListComparator(
                final ItemListSortStyle sortStyle,
                final ItemListSortDirection sortDirection
                ) {
            this.sortStyle = sortStyle;
            this.sortDirection = sortDirection;
        }
        
        /**
         * Get the sort style.
         * @return the sort style
         */
        public ItemListSortStyle getSortStyle() {
            return this.sortStyle;
        }
        
        /**
         * Get the sort direction.
         * @return the sort direction
         */
        public ItemListSortDirection getSortDirection() {
            return this.sortDirection;
        }

        /**
         * Set the start position.
         * @param startPosition the start position
         */
        public void setStartPosition(int startPosition) {
            if (startPosition < 0)
                throw new IndexOutOfBoundsException("Negative startPosition");
            this.startPosition = startPosition;
        }
        
        public int getStartPosition() {
            return this.startPosition;
        }
        
        /** {@inheritDoc} */
        public int compare(Object obj1, Object obj2) {
            ItemListElement
                    el1 = (ItemListElement) obj1,
                    el2 = (ItemListElement) obj2;
            if (sortStyle == ItemListSortStyle.ALPHABETICALLY) {
                String
                        s1 = el1.name,
                        s2 = el2.name;
                s1 = (startPosition < s1.length()) ? s1.substring(startPosition) : "";
                s2 = (startPosition < s2.length()) ? s2.substring(startPosition) : "";
                if (sortDirection == ItemListSortDirection.ASCENDING)
                    return s1.compareTo(s2);
                else
                    return s2.compareTo(s1);
            } else if (sortStyle == ItemListSortStyle.MEDIAN) {
                if (sortDirection == ItemListSortDirection.ASCENDING)
                    return el1.median.compareTo(el2.median);
                else
                    return el2.median.compareTo(el1.median);
            } else if (sortStyle == ItemListSortStyle.ORIGINAL_INDEX) {
                if (sortDirection == ItemListSortDirection.ASCENDING)
                    return el1.index - el2.index;
                else
                    return el2.index - el1.index;
            } else {
                throw new RuntimeException("Invalid 'sortOrder' parameter");
            }
        }
    }
    
    /** The item list elements */
    private class ItemListElement implements Comparable {
        String name;
        String sortKey;
        String median;
        int index;
        
        /**
         * Creates a new instance of ItemListElement
         * @param n the name
         * @param s the sort key
         * @param m the median
         * @param i the index
         */
        public ItemListElement(final String n, final String s, final String m, final int i) {
            name = n; sortKey = s; median = m; index = i;
        }
        
        /** {@inheritDoc} */
        public int compareTo(Object obj) {
            return getComparator().compare(this, obj);
        }
        
        /** {@inheritDoc} */
        public String toString() {
            return name;
        }
        
    }
    
    /** Creates a new instance of ItemListModel */
    public ItemListModel() {
        this(DEFAULT_SORTSTYLE, DEFAULT_SORTDIRECTION);
    }
    
    /**
     * Create a new instance of ItemListModel.
     * @param elements The elements to populate the list model with
     */
    public ItemListModel(final Collection elements) {
        this(elements, DEFAULT_SORTSTYLE, DEFAULT_SORTDIRECTION);
    }
    
    /**
     * Creates a new instance of ItemListModel
     * @param sortStyle the sort style
     * @param sortDirection the sort direction
     */
    public ItemListModel(
            final ItemListSortStyle sortStyle,
            final ItemListSortDirection sortDirection
            ) {
        super(new ItemListComparator(sortStyle, sortDirection));
    }
    
    /**
     * Create a new instance of ItemListModel.
     * @param elements The elements to populate the list model with
     * @param sortStyle the sort style
     * @param sortDirection the sort direction
     */
    public ItemListModel(
            final Collection elements,
            final ItemListSortStyle sortStyle,
            final ItemListSortDirection sortDirection
            ) {
        super(elements, new ItemListComparator(sortStyle, sortDirection));
    }
    
    /**
     * Set the sort style.
     * @param sortStyle the sort style
     */
    public void setSortStyle(final ItemListSortStyle sortStyle) {
        setComparator(new ItemListComparator(sortStyle, getSortDirection()));
    }
    
    /**
     * Set the alphabetical sort order with a start position.
     * @param startPosition the start position
     */
    public void setSortStyle(final int startPosition) {
        ItemListComparator comp =
                new ItemListComparator(ItemListSortStyle.ALPHABETICALLY, getSortDirection());
        comp.setStartPosition(startPosition);
        setComparator(comp);
    }
    
    /**
     * Set the sort direction.
     * @param sortDirection the sort direction
     */
    public void setSortDirection(final ItemListSortDirection sortDirection) {
        setComparator(new ItemListComparator(getSortStyle(), sortDirection));
    }
    
    /**
     * Get the sort style.
     * @return the sort style
     */
    public ItemListSortStyle getSortStyle() {
        return ((ItemListComparator) getComparator()).getSortStyle();
    }
    
    /**
     * Get the sort direction.
     * @return the sort direction
     */
    public ItemListSortDirection getSortDirection() {
        return ((ItemListComparator) getComparator()).getSortDirection();
    }
    
    /**
     * Add an item list element.
     * @param name the name
     * @param sortKey the sort key
     * @param median the median
     * @param index the index
     */
    public void add(final String name, final String sortKey, final String median, final int index) {
        add(new ItemListElement(name, sortKey, median, index));
    }
    
    /**
     * Add all item of a data collection
     * @param dc the data Collection
     * @param baseIndex the base index to use
     */
    public void addAll(final DataCollection dc, final int baseIndex) {
        ArrayList list = new ArrayList(dc.getItemCount());
        for (Iterator di = dc.getAllItemNames().iterator(); di.hasNext(); ) {
            String name = (String) di.next();
            list.add(new ItemListElement(name, name, dc.getMedian(name), baseIndex + dc.getIndex(name)));
        }
        addAll(list);
    }
    
    /**
     * Specialized implementation of {@link SortedUniqueListModel#contains}.
     * @param name the name to search for
     */
    public boolean contains(final String name) {
        return indexOf(name) >= 0;
    }
    
    /**
     * Specialized implementation of {@link SortedUniqueListModel#indexOf}.
     * @param name the name to search for
     */
    public int indexOf(final String name) {
        Iterator ei = delegate.iterator();
        for (int i = 0; ei.hasNext(); i++)
            if (ei.next().toString().equals(name))
                return i;
        return -1;
    }
    
}
