/*
 * JTLviz Main.java
 */

package de.pdv_systeme.JTLviz;

import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;

import java.awt.Component;
import java.awt.event.ItemEvent;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import org.jfree.chart.ChartPanel;
import org.jfree.data.time.RegularTimePeriod;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.time.TimeSeriesDataItem;
import org.jfree.data.time.TimeTableXYDataset;
import org.jfree.data.time.MovingAverage;
import org.jfree.data.time.TimePeriodAnchor;
import org.jfree.data.xy.XYDataset;
import org.jfree.ui.ExtensionFileFilter;
import org.jfree.ui.RefineryUtilities;
import org.jfree.util.StringUtils;

import de.pdv_systeme.DataCollection.AbstractDataCollection;
import de.pdv_systeme.DataCollection.DataCollection;
import de.pdv_systeme.DataCollection.FileDataCollection;
import de.pdv_systeme.DataCollection.MemoryUsageDataCollection;
import de.pdv_systeme.DataCollection.Operation;
import de.pdv_systeme.DataCollection.TDCDataCollection;
import de.pdv_systeme.DataCollection.TLCDataCollection;
import de.pdv_systeme.DataCollection.UnknownFileFormatException;
import de.pdv_systeme.TSChart.TSChart;
import de.pdv_systeme.TSChart.TSChartPropertyChangeListener;
import de.pdv_systeme.TSChart.TSChartPropertyChangeEvent;
import de.pdv_systeme.TSChart.TSChartProperty;
import de.pdv_systeme.TSChart.TSChartDrawingSupplier;
import de.pdv_systeme.util.AboutPanel;
import de.pdv_systeme.util.TextAreaBox;

/**
 * JTLviz is a pure Java re-implementation of TLviz.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public final class Main
        extends JFrame
        implements TSChartPropertyChangeListener {
    
    // TODO: Implement the difference between primary and secondary charts (measured vs. calculated)
    // TODO: Check behavior if TLC and ECP/TDC files are mixed
    // TODO: Implement loading of ZIP files
    
    /** This architecture's newline character */
    static final String NEWLINE = StringUtils.getLineSeparator();
    
    /**
     * The package-wide resourceBundle.
     * @see #getResourceString
     * @see #getFormattedResourceString
     */
    private static final ResourceBundle resources
            = ResourceBundle.getBundle("de.pdv_systeme.JTLviz.ResourceBundle");
    
    /** The application's name */
    static final String APP_NAME = getResourceString("APP_NAME");
    
    /** the list item selected by default */
    private static final String DEFAULT_ELEMENT = getResourceString("Main.DEFAULT_ELEMENT");
    
    /** list prototype element */
    private static final String PROTOTYPE_ELEMENT = getResourceString("Main.PROTOTYPE_ELEMENT");
    
    /** Threshold for entry of an item in the correlation item list */
    private static final double CORRELATION_THRESHOLD = 0.8;
    
    /** Maximum number of items per data collection (for sorting by index) */
    private static final int MAX_ITEMS_PER_DC = 1000;
    
    /** Action command for sort order */
    private static final String
            SORT_ORDER_ASCENDING = "Ascending",
            SORT_ORDER_DESCENDING = "Descending";
    
    /** A file filter to select CSV files */
    private static final ExtensionFileFilter
            csvFilter = new ExtensionFileFilter(
            getResourceString("CSV_files_(*.CSV)"),
            getResourceString(".csv")
            ),
            zipFilter = new ExtensionFileFilter(
            getResourceString("ZIP_files_(*.ZIP)"),
            getResourceString(".zip")
            );
    
    /** The data collections */
    private ArrayList collections; // of DataCollection
    
    /** The list model shown in the item list */
    private ItemListModel itemListModel;
    
    /** The stack of previuos list models */
    private Stack prevListModels; // of ItemListModel
    
    /** Flag whether the title still has its default value */
    private boolean isDefaultTitle;
    
    /** The Chart */
    private TSChart chart;
    
    /** The chart's properties */
    private TSChartProperty prop;
    
    /** The data collection used for correlation */
    private DataCollection correlationDataCollection;
    
    /** The last file opened/saved (for file choosers) */
    private File lastFile;
    
    /** The start position for alphabetical sort */
    private int alphaSortStartPos;
    
    /**
     * The main entry point.
     *
     * @param args  the command line arguments
     */
    public static void main(final String args[]) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                new Main(args);
            }
        });
    }
    
    /**
     * The Main constructor.
     *
     * @param loadFile  array of TLC/ECP filenames to load (may be <code>null</code>)
     */
    public Main(final String[] loadFile) {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        
        collections = new ArrayList();
        itemListModel = new ItemListModel();
        prevListModels = new Stack();
        
        prop = new TSChartProperty();
        prop.addPropertyChangeListener(this);
        
        isDefaultTitle = true;
        chart = null;
        lastFile = null;
        alphaSortStartPos = 1;
        
        initComponents();
        
        itemList.setModel(itemListModel);
        setMenuItemsEnabled(-1);
        RefineryUtilities.centerFrameOnScreen(this);
        setVisible(true);
        
        if (System.getProperty("JTLviz.memDebug") != null) {
            int maxAge = 300; /* seconds = 5 minutes */
            int interval = 2; /* seconds */
            MemoryUsageDataCollection mem
                    = new MemoryUsageDataCollection(maxAge, interval);
            addDCtoItemlist(mem);
        }
        
        if (loadFile != null) {
            for (int i = 0; i < loadFile.length; i++) {
                lastFile = new File(loadFile[i]);
                FileDataCollection dc = processFile(lastFile);
                if (dc != null)
                    addDCtoItemlist(dc);
            }
        };
    }
    
    /**
     * The default Main constructor.
     */
    public Main() {
        this(null);
    }
    
    /**
     * Show a "Not Yet Implemented" messagebox.
     *
     * @param parent the message box' parent frame
     */
    public static void showNYI(final JFrame parent) {
        JOptionPane.showMessageDialog(
                parent,
                getResourceString("NYI"),
                getResourceString("APP_NAME"),
                JOptionPane.PLAIN_MESSAGE
                );
    }
    
    /**
     * Report an exception via stderr and a message dialog.
     *
     * @param parent  the parent frame
     * @param exc  the exception to report
     * @param title  the dialog title
     */
    public static void reportError(JFrame parent, Exception exc, String title) {
        System.err.println(exc.getMessage());
        exc.printStackTrace();
        JOptionPane.showMessageDialog(parent, exc, title, JOptionPane.ERROR_MESSAGE);
    }
    
    /**
     * Report an error via stderr and a message dialog.
     *
     * @param parent  the parent frame
     * @param errMsg  the error message
     * @param title  the dialog title
     */
    public static void reportError(JFrame parent, String errMsg, String title) {
        System.err.println(errMsg);
        JOptionPane.showMessageDialog(parent, errMsg, title, JOptionPane.ERROR_MESSAGE);
    }
    
    /**
     * Get a string from the resource bundle.
     *
     * @param key  the key for the desired string
     * @return the string for the given key
     */
    static String getResourceString(String key) {
        return resources.getString(key);
    }
    
    /**
     * Get a format string from the resource bundle
     * and format it with arguments.
     *
     * @param key  the key for the desired string
     * @param arguments   an array of objects to be formatted and substituted
     * @return the string for the given key, formatted using the arguments
     */
    static String getFormattedResourceString(String key, Object[] arguments) {
        return MessageFormat.format(resources.getString(key), arguments);
    }
    
    /**
     * Enable/disable the menu items based on whether a file has been loaded,
     * whether exactly one or exactly two items are selected,
     * or whether a previous listbox exists.
     *
     * @param numItemsSelected  number of items selected (-1 if no file is opened)
     */
    private void setMenuItemsEnabled(final int numItemsSelected) {
        final boolean ALWAYS = true;
        final boolean NEVER = false;
        final boolean FILE_OPENED  = (numItemsSelected != -1);
        final boolean CHART_OPENED = (numItemsSelected > 0);
        final boolean SELECTED_ONE = (numItemsSelected == 1);
        final boolean SELECTED_TWO = (numItemsSelected == 2);
        final boolean HAVE_PREV_LISTBOX = ! prevListModels.empty();
        
        // File menu
        openMenuItem.setEnabled                   (ALWAYS);
        printPreviewMenuItem.setEnabled           (CHART_OPENED);
        printChartMenuItem.setEnabled             (CHART_OPENED);
        exportChartMenuItem.setEnabled            (CHART_OPENED);
        saveAsMenuItem.setEnabled                 (FILE_OPENED);
        propertiesMenuItem.setEnabled             (FILE_OPENED);
        closeMenuItem.setEnabled                  (FILE_OPENED);
        exitMenuItem.setEnabled                   (ALWAYS);
        
        // Options menu
        addItemMenuItem.setEnabled                (FILE_OPENED);
        chartMultipleItemsMenuItem.setEnabled     (FILE_OPENED);
        cycleDualMenuItem.setEnabled              (SELECTED_TWO);
        editChartPropertiesMenuItem.setEnabled    (CHART_OPENED);
        movingAverageMenuItem.setEnabled          (FILE_OPENED);
        ratioChartsMenuItem.setEnabled            (SELECTED_TWO);
        smoothMenuItem.setEnabled                 (SELECTED_ONE);
        candleChartMenuItem.setEnabled            (SELECTED_ONE);
        sortOrderMenuItem.setEnabled              (FILE_OPENED);
        
        // Chart Settings menu
        setTitleMenuItem.setEnabled               (FILE_OPENED);
        setSubtitleMenuItem.setEnabled            (FILE_OPENED);
        setFooterMenuItem.setEnabled              (FILE_OPENED);
        setPrimaryLineWidthMenuItem.setEnabled    (FILE_OPENED);
        setSecondaryLineWidthMenuItem.setEnabled  (FILE_OPENED);
        setMovingAveragePeriodMenuItem.setEnabled(FILE_OPENED);
        setSmoothingFactorMenuItem.setEnabled     (FILE_OPENED);
        
        // Modify Item List menu
        removeZeroItemsMenuItem.setEnabled        (FILE_OPENED);
        removeSelectedMenuItem.setEnabled         (FILE_OPENED);
        removeNotSelectedMenuItem.setEnabled      (FILE_OPENED);
        removeContainStringMenuItem.setEnabled    (FILE_OPENED);
        removeNotContainStringMenuItem.setEnabled(FILE_OPENED);
        sortMedianMenuItem.setEnabled             (FILE_OPENED);
        sortAlphabeticallyMenuItem.setEnabled     (FILE_OPENED);
        listOriginalOrderMenuItem.setEnabled      (FILE_OPENED);
        revertToPreviousListboxMenuItem.setEnabled(HAVE_PREV_LISTBOX);
        
        // Help menu
        aboutMenuItem.setEnabled                  (ALWAYS);
        viewReleaseNotesMenuItem.setEnabled       (NEVER);
    }
    
    /**
     * Push a non-empty list model onto the stack.
     * Set the "Revert to Previous Listbox" button's availability.
     */
    private void pushListModel() {
        if (itemListModel.isEmpty()) {
            revertToPreviousListboxMenuItem.setEnabled(false);
            return;
        }
        
        prevListModels.push(itemListModel);
        itemListModel =
                new ItemListModel(itemListModel, itemListModel.getSortStyle(), itemListModel.getSortDirection());
        revertToPreviousListboxMenuItem.setEnabled(true);
        itemList.setModel(itemListModel);
    }
    
    /**
     * Pop a list model from the stack.
     * Set the "Revert to Previous Listbox" button's availability.
     */
    private void popListModel() {
        if (prevListModels.empty())
            return;
        
        List selectedItems = getSelectedItems();
        itemListModel = (ItemListModel) prevListModels.pop();
        if (prevListModels.empty())
            revertToPreviousListboxMenuItem.setEnabled(false);
        itemList.setModel(itemListModel);
        setSelectedItems(selectedItems);
        itemList.revalidate();
    }
    
    /**
     * Get all selected items in the listbox.
     *
     * @return a list of all selected items
     */
    private List getSelectedItems() {
        int[] idx = itemList.getSelectedIndices();
        List selectedItems = new ArrayList(idx.length);
        for (int i = 0; i < idx.length; i++)
            selectedItems.add(itemListModel.getElementAt(idx[i]));
        return selectedItems;
    }
    
    /**
     * Set the items selected in the listbox.
     * Make sure the first selected item is visible.
     *
     * @param selectedItems the selected items
     */
    private void setSelectedItems(final List selectedItems) {
        int[] idx = new int[selectedItems.size()];
        Iterator si = selectedItems.iterator();
        for (int i = 0; si.hasNext(); i++)
            idx[i] = itemListModel.indexOf(si.next());
        itemList.setSelectedIndices(idx);
        if (idx.length > 0)
            itemList.ensureIndexIsVisible(idx[0]);
    }
    
    /**
     * Add a data collection to the item list
     *
     * @param dc the data collection
     */
    private void addDCtoItemlist(AbstractDataCollection dc) {
        collections.add(dc);
        List selectedItems = getSelectedItems();
        pushListModel();
        itemListModel.addAll(dc, MAX_ITEMS_PER_DC * collections.size());
        if (isDefaultTitle)
            setChartDefaultTitle();
        if (selectedItems.size() == 0) {
            // select default element, or first element, if default not in list
            int idx = Math.max(0, itemListModel.indexOf(DEFAULT_ELEMENT));
            itemList.setSelectedIndex(idx);
            itemList.ensureIndexIsVisible(idx);
        } else {
            setSelectedItems(selectedItems);
        }
        itemList.revalidate();
        leftPanel.setVisible(true);
    }
    
    /**
     * Load and process a data collection file.
     *
     * @param file  the data collection file
     * @return the data collection
     */
    private FileDataCollection processFile(final File file) {
        FileDataCollection dc = null;
        
        // Check file format and instantiate dc
        try {
            if (TLCDataCollection.testFormat(file)) {
                boolean useAllTimestamps =
                        (System.getProperty("JTLviz.useAllTimestamps") != null);
                dc = new TLCDataCollection(file, useAllTimestamps);
            } else if (TDCDataCollection.testFormat(file)) {
                dc = new TDCDataCollection(file);
            } else {
                throw new UnknownFileFormatException(file.getPath());
            }
        } catch (Exception e) {
            Main.reportError(this, e, getResourceString("File_Format_Error"));
            return null;
        }
        
        // Load file
        String filename = file.getPath();
        LoadTask loadTask = null;
        try {
            Object[] filenameArg = new Object[] {filename};
            
            statusLine.setText(
                    getFormattedResourceString("Counting_number_of_rows_of_data_in_", filenameArg
                    ));
            int lc = dc.getLinecount();
            
            // Look for a collection to get timestamps from
            FileDataCollection timeDc = null;
            if (System.getProperty("JTLviz.useIndividualTimestamps") == null) {
                for (Iterator di = collections.iterator(); di.hasNext(); ) {
                    DataCollection d = (DataCollection) di.next();
                    if (d instanceof FileDataCollection &&
                            ((FileDataCollection) d).getLinecount() == lc) {
                        timeDc = (FileDataCollection) d;
                        break;
                    }
                }
            }
            
            statusLine.setText(
                    getFormattedResourceString("Extracting_data_from_", filenameArg
                    ));
            loadTask = new LoadTask(this, dc, timeDc);
            loadTask.go();
        } catch (Exception e) {
            Main.reportError(this, e, getResourceString("I/O_Error"));
            loadTask = null;
        }
        
        // If the load hasn't finished, restore the status line and return
        if (loadTask == null || ! loadTask.isDone()) {
            if (collections.size() != 0) {
                DataCollection last = (DataCollection) collections.get(collections.size()-1);
                if (last instanceof FileDataCollection) {
                    statusLine.setText(((FileDataCollection) last).getFilename());
                    return null;
                }
            }
            statusLine.setText(getResourceString("Main.DEFAULT_STATUSLINE"));
            return null;
        }
        
        // Process file data
        statusLine.setText(filename);
        return dc;
    }
    
    /**
     * Generate the chart from the selected list model items.
     * <p>
     * Side effects:
     * <ul>
     *   <li>Ensure the first selected item is visible in the item list
     *   <li>Enable/disable the Correlate button
     *   <li>Enable/disable the menu items that depend on the number of selected items.
     * </ul>
     */
    private void generateChart() {
        int seriesCount;
        if (collections.isEmpty()) {
            seriesCount = -1;
            correlationDataCollection = null;
            chart = null;
        } else {
            int[] idx = itemList.getSelectedIndices();
            seriesCount = 0;
            for (int i = 0; i < idx.length; ++i) {
                String name = itemListModel.getElementAt(idx[i]).toString();
                for (Iterator ci = collections.iterator(); ci.hasNext(); ) {
                    DataCollection dc = (DataCollection) ci.next();
                    if (dc.contains(name))
                        seriesCount++;
                }
            }
            if (seriesCount == 0) {
                correlationDataCollection = null;
                chart = null;
            } else {
                itemList.ensureIndexIsVisible(idx[0]);
                try {
                    chart = new TSChart((ChartPanel) chartPanel, prop);
                    chart.setStackedString(getResourceString("Values_are_STACKED"));
                } catch (Exception e) {
                    Main.reportError(this, e, getResourceString("Error_creating_TimeSeriesChart"));
                }
            }
        }
        
        correlateButton.setEnabled(seriesCount == 1);
        setMenuItemsEnabled(seriesCount);
        ((ChartPanel) chartPanel).setChart(chart);
    }
    
    /**
     * Generate a dataset from the selected list model items.
     *
     * @return the dataset
     */
    private XYDataset generateDataset() {
        List selectedItems = getSelectedItems();
        
        if (prop.isStackedLines()) {
            TimeTableXYDataset dataset = new TimeTableXYDataset();
            dataset.setDomainIsPointsInTime(true);
            dataset.setXPosition(TimePeriodAnchor.END);
            
            int collNum = 1;
            for (Iterator ci = collections.iterator(); ci.hasNext(); collNum++) {
                DataCollection dc = (DataCollection) ci.next();
                for (Iterator item = selectedItems.iterator(); item.hasNext(); ) {
                    String name = item.next().toString();
                    if (dc.contains(name)) {
                        TimeSeries origTs = dc.getData(name);
                        // Create a copy of the series under a new name
                        String newSeriesName = getFormattedResourceString("LegendItem", new Object[] {
                            origTs.getKey(), new Integer(collNum)
                        });
                        for (int itemNum = 0; itemNum < origTs.getItemCount(); itemNum++) {
                            TimeSeriesDataItem dataItem = origTs.getDataItem(itemNum);
                            dataset.add(
                                    dataItem.getPeriod(),
                                    dataItem.getValue(),
                                    newSeriesName,
                                    itemNum == origTs.getItemCount()-1
                                    );
                        }
                        correlationDataCollection = dc;
                    }
                }
            }
            
            return dataset;
        } else { // ! prop.isStackedLines()
            TimeSeriesCollection dataset = new TimeSeriesCollection();
            dataset.setDomainIsPointsInTime(true);
            
            int collNum = 1;
            for (Iterator ci = collections.iterator(); ci.hasNext(); collNum++) {
                DataCollection dc = (DataCollection) ci.next();
                for (Iterator ii = selectedItems.iterator(); ii.hasNext(); ) {
                    String name = ii.next().toString();
                    if (dc.contains(name)) {
                        TimeSeries origTs = dc.getData(name);
                        // Create a copy of the series under a new name
                        String newSeriesName = getFormattedResourceString("LegendItem", new Object[] {
                            origTs.getKey(), new Integer(collNum)
                        });
                        TimeSeries newTs = new TimeSeries(
                                newSeriesName,
                                origTs.getTimePeriodClass()
                                );
                        for (int itemNum = 0; itemNum < origTs.getItemCount(); itemNum++) {
                            TimeSeriesDataItem dataItem = origTs.getDataItem(itemNum);
                            newTs.add(
                                    dataItem,
                                    itemNum == origTs.getItemCount()-1
                                    );
                        }
                        dataset.addSeries(newTs);
                        correlationDataCollection = dc;
                    }
                }
            }
            
            if (prop.isRatioChart() && dataset.getSeriesCount() == 2) {
                TimeSeries
                        ts1 = dataset.getSeries(0),
                        ts2 = dataset.getSeries(1);
                String
                        ts1Name = (String) ts1.getKey(),
                        ts2Name = (String) ts2.getKey();
                String
                        ts12Name = getFormattedResourceString("LegendItem_Ratio", new Object[] {
                    ts1Name, ts2Name
                }),
                        ts21Name = getFormattedResourceString("LegendItem_Ratio", new Object[] {
                    ts2Name, ts1Name
                });
                TimeSeries
                        ts12 = new TimeSeries(ts12Name, ts1.getTimePeriodClass()),
                        ts21 = new TimeSeries(ts21Name, ts1.getTimePeriodClass());
                for (int item = 0; item < ts1.getItemCount(); item++) {
                    TimeSeriesDataItem t1 = ts1.getDataItem(item);
                    RegularTimePeriod time = t1.getPeriod();
                    TimeSeriesDataItem t2 = ts2.getDataItem(time);
                    Number
                            d1 = t1.getValue(),
                            d2 = t2.getValue();
                    Number
                            r12 = null,
                            r21 = null;
                    if (d1 != null && d2 != null) {
                        if (d2.doubleValue() != 0.0)
                            r12 = new Double(d1.doubleValue() / d2.doubleValue());
                        if (d1.doubleValue() != 0.0)
                            r21 = new Double(d2.doubleValue() / d1.doubleValue());
                    }
                    ts12.add(time, r12, item == ts1.getItemCount()-1);
                    ts21.add(time, r21, item == ts1.getItemCount()-1);
                }
                dataset.addSeries(ts12);
                dataset.addSeries(ts21);
            }
            
            if (prop.isMovingAverage()) {
                int origSeriesCount = dataset.getSeriesCount();
                for (int series = 0; series < origSeriesCount; series++) {
                    TimeSeries origTs = dataset.getSeries(series);
                    String newSeriesName =
                            getFormattedResourceString("LegendItem_Moving_Average",
                            new Object[] {origTs.getKey(), new Integer(prop.getMovingAveragePeriod())}
                    );
                    TimeSeries newTs = MovingAverage.createMovingAverage(
                            origTs,
                            newSeriesName,
                            60*prop.getMovingAveragePeriod(),
                            prop.getMovingAveragePeriod()-1);
                    dataset.addSeries(newTs);
                }
            }
            
            return dataset;
        }
    }
    
    /**
     * Set the chart's default title.
     */
    private void setChartDefaultTitle() {
        if (collections.size() != 0) {
            Iterator ci = collections.iterator();
            String nodename = ((DataCollection) ci.next()).getNodename();
            StringBuffer titleBuffer = new StringBuffer(nodename);
            while (ci.hasNext()) {
                nodename = ((DataCollection) ci.next()).getNodename();
                titleBuffer.append(", ");
                titleBuffer.append(nodename);
            }
            prop.setTitle(getFormattedResourceString("Node(s):_", new Object[] {titleBuffer}));
        }
    }
    
    /**
     * Show a dialog requesting a title string.
     *
     * @param type  a string specifying the kind of title
     * @param defaultText  the default to fill in
     * @return  the title entered (or <code>null</code> if dialog was canceled)
     */
    private String doSetTitleDialog(final String type, final String defaultText) {
        return doStringDialog(getFormattedResourceString("Set_", new Object[] {type}),
                getFormattedResourceString("Please_set_the_", new Object[] {type}), defaultText);
    }
    
    /**
     * Show a string entry dialog.
     *
     * @param title  the dialog title
     * @param message  the message to display
     * @param defaultText  the default to fill in
     * @return the string entered (or <code>null</code> if dialog was canceled)
     */
    private String doStringDialog(
            final String title,
            final String message,
            final String defaultText
            ) {
        JTextField textField = new JTextField(defaultText, 40);
        JOptionPane optionPane = new JOptionPane(new Object[] {message, textField},
                JOptionPane.PLAIN_MESSAGE,
                JOptionPane.OK_CANCEL_OPTION
                );
        JDialog stringDialog = optionPane.createDialog(this, title);
        RefineryUtilities.centerDialogInParent(stringDialog);
        stringDialog.show();
        Object selectedValue = optionPane.getValue();
        if (selectedValue != null
                && selectedValue instanceof Integer
                && ((Integer) selectedValue).intValue() == JOptionPane.OK_OPTION)
            return textField.getText();
        else
            return null;
    }
    
    /**
     * Set a new item list sort style.
     *
     * @param sortStyle the sort style
     */
    private void changeItemListOrder(final ItemListSortStyle sortStyle) {
        List selectedItems = getSelectedItems();
        pushListModel();
        if (sortStyle == ItemListSortStyle.ALPHABETICALLY)
            itemListModel.setSortStyle(alphaSortStartPos);
        else
            itemListModel.setSortStyle(sortStyle);
        setSelectedItems(selectedItems);
        itemList.revalidate();
    }
    
    /**
     * Set a new item list sort direction.
     *
     * @param sortDirection the sort direction
     */
    private void changeItemListOrder(final ItemListSortDirection sortDirection) {
        List selectedItems = getSelectedItems();
        pushListModel();
        itemListModel.setSortDirection(sortDirection);
        setSelectedItems(selectedItems);
        itemList.revalidate();
    }
    
    /**
     * Receives notification of a TimeSeriesChartPropertyChange event.
     *
     * @param event  the event.
     */
    public void propertyChanged(TSChartPropertyChangeEvent event) {
        if (chart == null) {
            generateChart();
        } else {
            String name = event.getPropertyName();
            if (name == TSChartProperty.TITLE_PROPERTY) {
                chart.setTitle((String) event.getNewValue());
            } else if (name == TSChartProperty.SUBTITLE_PROPERTY
                    || name == TSChartProperty.FOOTERTEXT_PROPERTY) {
                chart.generateSubtitles();
            } else if (name == TSChartProperty.PRIMARYLINEWIDTH_PROPERTY
                    || name == TSChartProperty.SECONDARYLINEWIDTH_PROPERTY) {
                chart.getPlot().setDrawingSupplier(new TSChartDrawingSupplier(prop));
                // the above statement really should regenerate the chart
                generateChart();
            } else if (name == TSChartProperty.STACKEDLINES_PROPERTY
                    || name == TSChartProperty.RATIOCHART_PROPERTY
                    || name == TSChartProperty.MOVINGAVERAGE_PROPERTY
                    || name == TSChartProperty.MOVINGAVERAGEPERIOD_PROPERTY
                    || name == TSChartProperty.SMOOTHCHART_PROPERTY
                    || name == TSChartProperty.SMOOTHINGFACTOR_PROPERTY
                    || name == TSChartProperty.CANDLECHART_PROPERTY) {
                prop.setDataset(generateDataset());
            } else if (name == TSChartProperty.DATASET_PROPERTY
                    || name == TSChartProperty.DUALCHART_PROPERTY
                    || name == TSChartProperty.DUALYAXIS_PROPERTY) {
                generateChart();
            } else {
                throw new RuntimeException("Unknown TSChart property "+name);
            }
        }
    }
    
    /**
     * Nested class for the item list cell renderer, to display tool tips
     */
    class ItemListCellRenderer extends DefaultListCellRenderer {
        /** {@inheritDoc} */
        public Component getListCellRendererComponent(
                JList list,              // the list
                Object value,            // value to display
                int index,               // cell index
                boolean isSelected,      // is the cell selected
                boolean cellHasFocus     // the list and the cell have the focus
                ) {
            super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
            String name = value.toString();
            for (Iterator ci = collections.iterator(); ci.hasNext(); ) {
                DataCollection dc = (DataCollection) ci.next();
                if (dc.contains(name)) {
                    // TODO: Wrap long descriptions
                    setToolTipText(dc.getDescription(name));
                    break;
                }
            }
            return this;
        }
    }
    
    
    /**
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        javax.swing.JPanel buttonsPanel;
        javax.swing.JScrollPane columnListScrollPane;
        javax.swing.JPanel correlateZoomPanel;
        javax.swing.JPanel rightPanel;
        javax.swing.JPanel stackPanel;

        leftPanel = new javax.swing.JPanel();
        columnListScrollPane = new javax.swing.JScrollPane();
        itemList = new javax.swing.JList();
        buttonsPanel = new javax.swing.JPanel();
        stackPanel = new javax.swing.JPanel();
        stackButton = new javax.swing.JToggleButton();
        unstackButton = new javax.swing.JToggleButton();
        correlateZoomPanel = new javax.swing.JPanel();
        correlateButton = new javax.swing.JButton();
        undoZoomButton = new javax.swing.JButton();
        rightPanel = new javax.swing.JPanel();
        chartPanel = new ChartPanel(
            null,   // chart
            false,  // properties
            false,  // save
            false,  // print
            false,  // zoom
            true    // tooltips
        );
        statusLine = new javax.swing.JLabel();
        menuBar = new javax.swing.JMenuBar();
        fileMenu = new javax.swing.JMenu();
        openMenuItem = new javax.swing.JMenuItem();
        printPreviewMenuItem = new javax.swing.JMenuItem();
        printChartMenuItem = new javax.swing.JMenuItem();
        exportChartMenuItem = new javax.swing.JMenuItem();
        saveAsMenuItem = new javax.swing.JMenuItem();
        propertiesMenuItem = new javax.swing.JMenuItem();
        closeMenuItem = new javax.swing.JMenuItem();
        exitMenuItem = new javax.swing.JMenuItem();
        optionsMenu = new javax.swing.JMenu();
        addItemMenuItem = new javax.swing.JMenuItem();
        chartMultipleItemsMenuItem = new javax.swing.JMenuItem();
        cycleDualMenuItem = new javax.swing.JMenuItem();
        editChartPropertiesMenuItem = new javax.swing.JMenuItem();
        movingAverageMenuItem = new javax.swing.JCheckBoxMenuItem();
        ratioChartsMenuItem = new javax.swing.JCheckBoxMenuItem();
        smoothMenuItem = new javax.swing.JCheckBoxMenuItem();
        candleChartMenuItem = new javax.swing.JCheckBoxMenuItem();
        sortOrderMenuItem = new javax.swing.JMenu();
        sortOrderAscendingMenuItem = new javax.swing.JMenuItem();
        sortOrderDescendingMenuItem = new javax.swing.JMenuItem();
        chartSettingsMenu = new javax.swing.JMenu();
        setTitleMenuItem = new javax.swing.JMenuItem();
        setSubtitleMenuItem = new javax.swing.JMenuItem();
        setFooterMenuItem = new javax.swing.JMenuItem();
        setPrimaryLineWidthMenuItem = new javax.swing.JMenuItem();
        setSecondaryLineWidthMenuItem = new javax.swing.JMenuItem();
        setMovingAveragePeriodMenuItem = new javax.swing.JMenuItem();
        setSmoothingFactorMenuItem = new javax.swing.JMenuItem();
        modifyItemlistMenu = new javax.swing.JMenu();
        removeZeroItemsMenuItem = new javax.swing.JMenuItem();
        removeSelectedMenuItem = new javax.swing.JMenuItem();
        removeNotSelectedMenuItem = new javax.swing.JMenuItem();
        removeContainStringMenuItem = new javax.swing.JMenuItem();
        removeNotContainStringMenuItem = new javax.swing.JMenuItem();
        sortMedianMenuItem = new javax.swing.JMenuItem();
        sortAlphabeticallyMenuItem = new javax.swing.JMenuItem();
        listOriginalOrderMenuItem = new javax.swing.JMenuItem();
        revertToPreviousListboxMenuItem = new javax.swing.JMenuItem();
        helpMenu = new javax.swing.JMenu();
        aboutMenuItem = new javax.swing.JMenuItem();
        viewReleaseNotesMenuItem = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle(Main.getResourceString("Main.APP_TITLE"));
        leftPanel.setLayout(new java.awt.BorderLayout(2, 2));

        leftPanel.setVisible(false);
        itemList.setCellRenderer(new ItemListCellRenderer());
        itemList.setPrototypeCellValue(PROTOTYPE_ELEMENT);
        itemList.addListSelectionListener(new javax.swing.event.ListSelectionListener() {
            public void valueChanged(javax.swing.event.ListSelectionEvent evt) {
                itemListValueChanged(evt);
            }
        });

        columnListScrollPane.setViewportView(itemList);

        leftPanel.add(columnListScrollPane, java.awt.BorderLayout.CENTER);

        buttonsPanel.setLayout(new java.awt.BorderLayout(2, 2));

        stackPanel.setLayout(new java.awt.GridLayout(0, 1, 2, 2));

        stackPanel.setBorder(new javax.swing.border.LineBorder(new java.awt.Color(0, 0, 0), 1, true));
        stackButton.setText(getResourceString("Stack"));
        stackButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                stackButtonActionPerformed(evt);
            }
        });

        stackPanel.add(stackButton);

        unstackButton.setSelected(true);
        unstackButton.setText(getResourceString("UnStack"));
        unstackButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                unstackButtonActionPerformed(evt);
            }
        });

        stackPanel.add(unstackButton);

        buttonsPanel.add(stackPanel, java.awt.BorderLayout.CENTER);

        correlateZoomPanel.setLayout(new java.awt.GridLayout(0, 1, 2, 2));

        correlateButton.setText(getResourceString("Correlate"));
        correlateButton.setDefaultCapable(false);
        correlateButton.setEnabled(false);
        correlateButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                correlateButtonActionPerformed(evt);
            }
        });

        correlateZoomPanel.add(correlateButton);

        undoZoomButton.setText(getResourceString("Undo_Zoom/Scroll"));
        undoZoomButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                undoZoomButtonActionPerformed(evt);
            }
        });

        correlateZoomPanel.add(undoZoomButton);

        buttonsPanel.add(correlateZoomPanel, java.awt.BorderLayout.EAST);

        leftPanel.add(buttonsPanel, java.awt.BorderLayout.SOUTH);

        getContentPane().add(leftPanel, java.awt.BorderLayout.WEST);

        rightPanel.setLayout(new java.awt.BorderLayout());

        chartPanel.setBorder(new javax.swing.border.BevelBorder(javax.swing.border.BevelBorder.RAISED));
        rightPanel.add(chartPanel, java.awt.BorderLayout.CENTER);

        getContentPane().add(rightPanel, java.awt.BorderLayout.CENTER);

        statusLine.setText(Main.getResourceString("Main.DEFAULT_STATUSLINE"));
        getContentPane().add(statusLine, java.awt.BorderLayout.SOUTH);

        fileMenu.setMnemonic(getResourceString("File").charAt(0));
        fileMenu.setText(getResourceString("File"));
        openMenuItem.setText(getResourceString("Open"));
        openMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                openMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(openMenuItem);

        printPreviewMenuItem.setText(getResourceString("PrintPreview"));
        printPreviewMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                printPreviewMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(printPreviewMenuItem);

        printChartMenuItem.setText(getResourceString("PrintChart"));
        printChartMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                printChartMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(printChartMenuItem);

        exportChartMenuItem.setText(getResourceString("ExportChart"));
        exportChartMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exportChartMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(exportChartMenuItem);

        saveAsMenuItem.setText(getResourceString("Save_as_..."));
        saveAsMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveAsMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(saveAsMenuItem);

        propertiesMenuItem.setText(getResourceString("Properties"));
        propertiesMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                propertiesMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(propertiesMenuItem);

        closeMenuItem.setText(getResourceString("Close"));
        closeMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                closeMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(closeMenuItem);

        exitMenuItem.setText(getResourceString("Exit"));
        exitMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitMenuItemActionPerformed(evt);
            }
        });

        fileMenu.add(exitMenuItem);

        menuBar.add(fileMenu);

        optionsMenu.setMnemonic(getResourceString("Options").charAt(0));
        optionsMenu.setText(getResourceString("Options"));
        addItemMenuItem.setText(getResourceString("Add_New_Item_to_List_Box"));
        addItemMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                addItemMenuItemActionPerformed(evt);
            }
        });

        optionsMenu.add(addItemMenuItem);

        chartMultipleItemsMenuItem.setText(getResourceString("Chart_Multiple_Items"));
        chartMultipleItemsMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                chartMultipleItemsMenuItemActionPerformed(evt);
            }
        });

        optionsMenu.add(chartMultipleItemsMenuItem);

        cycleDualMenuItem.setText(getResourceString("Cycle_Dual_Y_or_Dual_Charts"));
        cycleDualMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                cycleDualMenuItemActionPerformed(evt);
            }
        });

        optionsMenu.add(cycleDualMenuItem);

        editChartPropertiesMenuItem.setText(getResourceString("Edit_Chart_Properties"));
        editChartPropertiesMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                editChartPropertiesMenuItemActionPerformed(evt);
            }
        });

        optionsMenu.add(editChartPropertiesMenuItem);

        movingAverageMenuItem.setText(Main.getResourceString("Toggle_Moving_Average"));
        movingAverageMenuItem.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(java.awt.event.ItemEvent evt) {
                movingAverageMenuItemStateChanged(evt);
            }
        });

        optionsMenu.add(movingAverageMenuItem);

        ratioChartsMenuItem.setText(Main.getResourceString("Generate_Ratio_Charts"));
        ratioChartsMenuItem.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(java.awt.event.ItemEvent evt) {
                ratioChartsMenuItemStateChanged(evt);
            }
        });

        optionsMenu.add(ratioChartsMenuItem);

        smoothMenuItem.setText(Main.getResourceString("Smooth/Unsmooth"));
        smoothMenuItem.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(java.awt.event.ItemEvent evt) {
                smoothMenuItemStateChanged(evt);
            }
        });

        optionsMenu.add(smoothMenuItem);

        candleChartMenuItem.setText(Main.getResourceString("Add/Remove_Candle_Chart"));
        candleChartMenuItem.addItemListener(new java.awt.event.ItemListener() {
            public void itemStateChanged(java.awt.event.ItemEvent evt) {
                candleChartMenuItemStateChanged(evt);
            }
        });

        optionsMenu.add(candleChartMenuItem);

        sortOrderMenuItem.setText(getResourceString("Set_Sort_Order"));
        sortOrderAscendingMenuItem.setText(getResourceString("Ascending"));
        sortOrderAscendingMenuItem.setActionCommand(SORT_ORDER_ASCENDING);
        sortOrderAscendingMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortOrderActionPerformed(evt);
            }
        });

        sortOrderMenuItem.add(sortOrderAscendingMenuItem);

        sortOrderDescendingMenuItem.setText(getResourceString("Descending"));
        sortOrderDescendingMenuItem.setActionCommand(SORT_ORDER_DESCENDING);
        sortOrderDescendingMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortOrderActionPerformed(evt);
            }
        });

        sortOrderMenuItem.add(sortOrderDescendingMenuItem);

        optionsMenu.add(sortOrderMenuItem);

        menuBar.add(optionsMenu);

        chartSettingsMenu.setMnemonic(getResourceString("Chart_Settings").charAt(0));
        chartSettingsMenu.setText(getResourceString("Chart_Settings"));
        setTitleMenuItem.setText(getResourceString("Set_Chart_Title"));
        setTitleMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setTitleMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setTitleMenuItem);

        setSubtitleMenuItem.setText(getResourceString("Set_Chart_Sub-Title"));
        setSubtitleMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setSubtitleMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setSubtitleMenuItem);

        setFooterMenuItem.setText(getResourceString("Set_Chart_Footer"));
        setFooterMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setFooterMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setFooterMenuItem);

        setPrimaryLineWidthMenuItem.setText(getResourceString("Set_Primary_Chart_Line_Width"));
        setPrimaryLineWidthMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setPrimaryLineWidthMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setPrimaryLineWidthMenuItem);

        setSecondaryLineWidthMenuItem.setText(getResourceString("Set_Secondary_Chart_Line_Width"));
        setSecondaryLineWidthMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setSecondaryLineWidthMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setSecondaryLineWidthMenuItem);

        setMovingAveragePeriodMenuItem.setText(getResourceString("Set_Moving_Average_Period_Value"));
        setMovingAveragePeriodMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setMovingAveragePeriodMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setMovingAveragePeriodMenuItem);

        setSmoothingFactorMenuItem.setText(getResourceString("Set_Smoothing_Factor"));
        setSmoothingFactorMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                setSmoothingFactorMenuItemActionPerformed(evt);
            }
        });

        chartSettingsMenu.add(setSmoothingFactorMenuItem);

        menuBar.add(chartSettingsMenu);

        modifyItemlistMenu.setMnemonic(getResourceString("Modify_Item_List").charAt(0));
        modifyItemlistMenu.setText(getResourceString("Modify_Item_List"));
        removeZeroItemsMenuItem.setText(getResourceString("Remove_Items_with_Zero_Values"));
        removeZeroItemsMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeZeroItemsMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(removeZeroItemsMenuItem);

        removeSelectedMenuItem.setText(getResourceString("Remove_Selected_Item(s)"));
        removeSelectedMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeSelectedMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(removeSelectedMenuItem);

        removeNotSelectedMenuItem.setText(getResourceString("Remove_Item(s)_NOT_Selected"));
        removeNotSelectedMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeNotSelectedMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(removeNotSelectedMenuItem);

        removeContainStringMenuItem.setText(getResourceString("Remove_Item(s)_containing_string"));
        removeContainStringMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeContainStringMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(removeContainStringMenuItem);

        removeNotContainStringMenuItem.setText(getResourceString("Remove_Item(s)_NOT_containing_string"));
        removeNotContainStringMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                removeNotContainStringMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(removeNotContainStringMenuItem);

        sortMedianMenuItem.setText(getResourceString("Sort_in_Median_order"));
        sortMedianMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortMedianMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(sortMedianMenuItem);

        sortAlphabeticallyMenuItem.setText(getResourceString("Sort_Items_in_Alphabetical_Order"));
        sortAlphabeticallyMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sortAlphabeticallyMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(sortAlphabeticallyMenuItem);

        listOriginalOrderMenuItem.setText(getResourceString("List_Items_in_Original_Order"));
        listOriginalOrderMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                listOriginalOrderMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(listOriginalOrderMenuItem);

        revertToPreviousListboxMenuItem.setText(getResourceString("Revert_to_Previous_List_Box"));
        revertToPreviousListboxMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                revertToPreviousListboxMenuItemActionPerformed(evt);
            }
        });

        modifyItemlistMenu.add(revertToPreviousListboxMenuItem);

        menuBar.add(modifyItemlistMenu);

        helpMenu.setMnemonic(getResourceString("Help").charAt(0));
        helpMenu.setText(getResourceString("Help"));
        aboutMenuItem.setText(Main.getFormattedResourceString("About_", new Object[] {APP_NAME}));
        aboutMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                aboutMenuItemActionPerformed(evt);
            }
        });

        helpMenu.add(aboutMenuItem);

        viewReleaseNotesMenuItem.setText(getResourceString("View_JTLviz_Release_Notes"));
        viewReleaseNotesMenuItem.setEnabled(false);
        helpMenu.add(viewReleaseNotesMenuItem);

        menuBar.add(helpMenu);

        setJMenuBar(menuBar);

        pack();
    }
    // </editor-fold>//GEN-END:initComponents
    
    private void candleChartMenuItemStateChanged(java.awt.event.ItemEvent evt) {//GEN-FIRST:event_candleChartMenuItemStateChanged
        Object o = evt.getItemSelectable();
        if (o == candleChartMenuItem)
            prop.setCandleChart(evt.getStateChange() == ItemEvent.SELECTED);
        // TODO: implement candle chart
        showNYI(this);
    }//GEN-LAST:event_candleChartMenuItemStateChanged
    
    private void smoothMenuItemStateChanged(java.awt.event.ItemEvent evt) {//GEN-FIRST:event_smoothMenuItemStateChanged
        Object o = evt.getItemSelectable();
        if (o == smoothMenuItem)
            prop.setSmoothChart(evt.getStateChange() == ItemEvent.SELECTED);
        // TODO: implement chart smoothing
        showNYI(this);
    }//GEN-LAST:event_smoothMenuItemStateChanged
    
    private void ratioChartsMenuItemStateChanged(java.awt.event.ItemEvent evt) {//GEN-FIRST:event_ratioChartsMenuItemStateChanged
        Object o = evt.getItemSelectable();
        if (o == ratioChartsMenuItem)
            prop.setRatioChart(evt.getStateChange() == ItemEvent.SELECTED);
    }//GEN-LAST:event_ratioChartsMenuItemStateChanged
    
    private void movingAverageMenuItemStateChanged(java.awt.event.ItemEvent evt) {//GEN-FIRST:event_movingAverageMenuItemStateChanged
        Object o = evt.getItemSelectable();
        if (o == movingAverageMenuItem)
            prop.setMovingAverage(evt.getStateChange() == ItemEvent.SELECTED);
    }//GEN-LAST:event_movingAverageMenuItemStateChanged
    
    private void listOriginalOrderMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_listOriginalOrderMenuItemActionPerformed
        changeItemListOrder(ItemListSortStyle.ORIGINAL_INDEX);
    }//GEN-LAST:event_listOriginalOrderMenuItemActionPerformed
    
    private void sortAlphabeticallyMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sortAlphabeticallyMenuItemActionPerformed
        boolean invalidEntry;
        do {
            invalidEntry = false;
            String startPosString = doStringDialog(
                    getResourceString("Set_Sort_Key_Start_Position"),
                    getResourceString("Please_enter_Sort_Key_Start_Position_(1-10)"),
                    Integer.toString(alphaSortStartPos)
                    );
            if (startPosString != null) {
                int sp = 0;
                try {
                    sp = Integer.parseInt(startPosString);
                    invalidEntry = (sp < 1 || sp > 10);
                } catch (Exception e) {
                    invalidEntry = true;
                }
                if (! invalidEntry) {
                    alphaSortStartPos = sp;
                    changeItemListOrder(ItemListSortStyle.ALPHABETICALLY);
                }
            }
        } while (invalidEntry);
    }//GEN-LAST:event_sortAlphabeticallyMenuItemActionPerformed
    
    private void sortOrderActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sortOrderActionPerformed
        if (evt.getActionCommand() == SORT_ORDER_ASCENDING)
            changeItemListOrder(ItemListSortDirection.ASCENDING);
        else // evt.getActionCommand() == SORT_ORDER_DESCENDING
            changeItemListOrder(ItemListSortDirection.DESCENDING);
    }//GEN-LAST:event_sortOrderActionPerformed
    
    private void sortMedianMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sortMedianMenuItemActionPerformed
        if (System.getProperty("JTLviz.withMedian") != null)
            changeItemListOrder(ItemListSortStyle.MEDIAN);
        else
            showNYI(this);
    }//GEN-LAST:event_sortMedianMenuItemActionPerformed
    
    private void setSmoothingFactorMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setSmoothingFactorMenuItemActionPerformed
        boolean invalidEntry;
        do {
            invalidEntry = false;
            String smoothFactor = doStringDialog(
                    getResourceString("Set_Smoothing_Factor"),
                    getResourceString("Please_enter_Smoothing_Factor_(1-100)"),
                    String.valueOf(prop.getSmoothingFactor())
                    );
            if (smoothFactor != null) {
                try {
                    prop.setSmoothingFactor(Integer.parseInt(smoothFactor));
                } catch (Exception e) {
                    invalidEntry = true;
                }
            }
        } while (invalidEntry);
    }//GEN-LAST:event_setSmoothingFactorMenuItemActionPerformed
    
    private void setMovingAveragePeriodMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setMovingAveragePeriodMenuItemActionPerformed
        boolean invalidEntry;
        do {
            invalidEntry = false;
            String movAvg = doStringDialog(
                    getResourceString("Moving_Average_Period"),
                    getResourceString("Please_enter_PERIOD_value_to_use"),
                    String.valueOf(prop.getMovingAveragePeriod())
                    );
            if (movAvg != null) {
                try {
                    prop.setMovingAveragePeriod(Integer.parseInt(movAvg));
                } catch (Exception e) {
                    invalidEntry = true;
                }
            }
        } while (invalidEntry);
    }//GEN-LAST:event_setMovingAveragePeriodMenuItemActionPerformed
    
    private void setSecondaryLineWidthMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setSecondaryLineWidthMenuItemActionPerformed
        boolean invalidEntry;
        do {
            invalidEntry = false;
            String lineWidth = doStringDialog(
                    getResourceString("Set_Secondary_Chart_Line_Width"),
                    getResourceString("Please_enter_Secondary_Chart_Line_Width_(1-5)"),
                    String.valueOf(prop.getSecondaryLineWidth())
                    );
            if (lineWidth != null) {
                try {
                    prop.setSecondaryLineWidth(Integer.parseInt(lineWidth));
                } catch (Exception e) {
                    invalidEntry = true;
                }
            }
        } while (invalidEntry);
    }//GEN-LAST:event_setSecondaryLineWidthMenuItemActionPerformed
    
    private void setPrimaryLineWidthMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setPrimaryLineWidthMenuItemActionPerformed
        boolean invalidEntry;
        do {
            invalidEntry = false;
            String lineWidth = doStringDialog(
                    getResourceString("Set_Primary_Chart_Line_Width"),
                    getResourceString("Please_enter_Primary_Chart_Line_Width_(1-5)"),
                    String.valueOf(prop.getPrimaryLineWidth())
                    );
            if (lineWidth != null) {
                try {
                    prop.setPrimaryLineWidth(Integer.parseInt(lineWidth));
                } catch (Exception e) {
                    invalidEntry = true;
                }
            }
        } while (invalidEntry);
    }//GEN-LAST:event_setPrimaryLineWidthMenuItemActionPerformed
    
    private void saveAsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveAsMenuItemActionPerformed
        // Show file chooser.
        JFileChooser fileChooser = new JFileChooser(lastFile);
        fileChooser.addChoosableFileFilter(csvFilter);
        
        int result = fileChooser.showSaveDialog(this);
        
        // Process the results.
        if (result == fileChooser.APPROVE_OPTION) {
            lastFile = fileChooser.getSelectedFile();
            String filename = lastFile.getPath();
            // enforce .csv extension
            if (false) {
                String extension = getResourceString(".csv");
                if (! StringUtils.endsWithIgnoreCase(filename, extension))
                    filename = filename + extension;
            }
            DataCollection last = (DataCollection) collections.get(collections.size()-1);
            if (! (last instanceof AbstractDataCollection))
                return;
            AbstractDataCollection adc = (AbstractDataCollection) last;
            Set items = new TreeSet();
            for (Iterator li = itemListModel.iterator(); li.hasNext(); ) {
                String name = li.next().toString();
                if (adc.contains(name))
                    items.add(name);
            }
            try {
                adc.writeData(new File(filename), items);
            } catch (IOException e) {
                Main.reportError(this, e, getResourceString("I/O_Error"));
            }
        }
    }//GEN-LAST:event_saveAsMenuItemActionPerformed
    
    private void printPreviewMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_printPreviewMenuItemActionPerformed
        // TODO: implement "Print Preview"
        showNYI(this);
    }//GEN-LAST:event_printPreviewMenuItemActionPerformed
    
    private void revertToPreviousListboxMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_revertToPreviousListboxMenuItemActionPerformed
        popListModel();
    }//GEN-LAST:event_revertToPreviousListboxMenuItemActionPerformed
    
    private void removeNotContainStringMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeNotContainStringMenuItemActionPerformed
        String currStr = itemList.getSelectedValue().toString();
        if (currStr == null)
            currStr = "";
        String containString = doStringDialog(
                getResourceString("Remove_Items_NOT_containing_string"),
                getResourceString("Please_enter_search_string"),
                currStr);
        if (containString != null && ! containString.equals("")) {
            List selectedItems = getSelectedItems();
            pushListModel();
            for (Iterator li = itemListModel.iterator(); li.hasNext(); )
                if (li.next().toString().indexOf(containString) == -1)
                    li.remove();
            setSelectedItems(selectedItems);
            itemList.revalidate();
        }
    }//GEN-LAST:event_removeNotContainStringMenuItemActionPerformed
    
    private void removeContainStringMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeContainStringMenuItemActionPerformed
        String currStr = itemList.getSelectedValue().toString();
        if (currStr == null)
            currStr = "";
        String containString = doStringDialog(
                getResourceString("Remove_Items_containing_string"),
                getResourceString("Please_enter_search_string"),
                currStr);
        if (containString != null && ! containString.equals("")) {
            List selectedItems = getSelectedItems();
            pushListModel();
            for (Iterator li = itemListModel.iterator(); li.hasNext(); )
                if (li.next().toString().indexOf(containString) != -1)
                    li.remove();
            setSelectedItems(selectedItems);
            itemList.revalidate();
        }
    }//GEN-LAST:event_removeContainStringMenuItemActionPerformed
    
    private void chartMultipleItemsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_chartMultipleItemsMenuItemActionPerformed
        String currStr = itemList.getSelectedValue().toString();
        if (currStr == null)
            currStr = "";
        String matchString = doStringDialog(
                getResourceString("Chart_Multiple_Items"),
                getResourceString("Please_enter_search_string"),
                currStr);
        if (matchString != null && ! matchString.equals("")) {
            int[] matches = new int[itemListModel.getSize()];
            int count = 0;
            Iterator li = itemListModel.iterator();
            for (int i = 0; li.hasNext(); i++)
                if (li.next().toString().indexOf(matchString) != -1)
                    matches[count++] = i;
            if (count != 0) {
                int[] match = new int[count];
                System.arraycopy(matches, 0, match, 0, count);
                itemList.setSelectedIndices(match);
            }
        }
    }//GEN-LAST:event_chartMultipleItemsMenuItemActionPerformed
    
    private void cycleDualMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_cycleDualMenuItemActionPerformed
        if (prop.isDualChart()) {
            prop.setDualChart(false);
            prop.setDualYAxis(false);
        } else {
            if (prop.isDualYAxis()) {
                prop.setDualChart(true);
                prop.setDualYAxis(false);
            } else {
                prop.setDualYAxis(true);
            }
        }
    }//GEN-LAST:event_cycleDualMenuItemActionPerformed
    
    private void removeNotSelectedMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeNotSelectedMenuItemActionPerformed
        List selectedItems = getSelectedItems();
        pushListModel();
        itemListModel.retainAll(selectedItems);
        setSelectedItems(selectedItems);
        itemList.revalidate();
    }//GEN-LAST:event_removeNotSelectedMenuItemActionPerformed
    
    private void removeSelectedMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeSelectedMenuItemActionPerformed
        List selectedItems = getSelectedItems();
        pushListModel();
        itemListModel.removeAll(selectedItems);
        itemList.clearSelection();
        itemList.revalidate();
    }//GEN-LAST:event_removeSelectedMenuItemActionPerformed
    
    private void addItemMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_addItemMenuItemActionPerformed
        AddItemDialog addItemDialog = new AddItemDialog(this, itemListModel);
        RefineryUtilities.centerDialogInParent(addItemDialog);
        addItemDialog.show();
        if (addItemDialog.isValidItem()) {
            String name = addItemDialog.getNewItemName();
            Operation op = addItemDialog.getOperation();
            
            Object source1 = addItemDialog.getSourceItem(AddItemDialog.SOURCE_ITEM1);
            Object source2 = addItemDialog.getSourceItem(AddItemDialog.SOURCE_ITEM2);
            String s1 = (source1 instanceof String) ? (String) source1 : null;
            String s2 = (source2 instanceof String) ? (String) source2 : null;
            Double d1 = (source1 instanceof Double) ? (Double) source1 : null;
            Double d2 = (source2 instanceof Double) ? (Double) source2 : null;
            
            boolean isAdded = false;
            if (d1 != null && d2 != null) {
                DataCollection dc = (DataCollection) collections.get(0);
                dc.addCalculatedSeries(name, op, d1.doubleValue(), d2.doubleValue());
                isAdded = true;
            } else if (s1 != null) {
                for (Iterator ci1 = collections.iterator(); ci1.hasNext(); ) {
                    DataCollection dc1 = (DataCollection) ci1.next();
                    if (dc1.contains(s1)) {
                        if (d2 != null) {
                            dc1.addCalculatedSeries(name, op, dc1.getData(s1), d2.doubleValue());
                            isAdded = true;
                        } else { // s2 != null
                            for (Iterator ci2 = collections.iterator(); ci2.hasNext(); ) {
                                DataCollection dc2 = (DataCollection) ci2.next();
                                if (dc2.contains(s2)) {
                                    dc1.addCalculatedSeries(name, op, dc1.getData(s1), dc2.getData(s2));
                                    isAdded = true;
                                }
                            }
                        }
                    }
                }
            } else { // d1 != null && s2 != null
                for (Iterator ci2 = collections.iterator(); ci2.hasNext(); ) {
                    DataCollection dc2 = (DataCollection) ci2.next();
                    if (dc2.contains(s2)) {
                        dc2.addCalculatedSeries(name, op, d1.doubleValue(), dc2.getData(s2));
                        isAdded = true;
                    }
                }
            }
            
            if (isAdded) {
                pushListModel();
                itemListModel.add(name, "", "", 0);
                int idx = itemListModel.indexOf(name);
                itemList.setSelectedIndex(idx);
                itemList.ensureIndexIsVisible(idx);
                itemList.revalidate();
            }
        }
        addItemDialog.dispose();
    }//GEN-LAST:event_addItemMenuItemActionPerformed
    
    private void editChartPropertiesMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_editChartPropertiesMenuItemActionPerformed
        ChartPropertiesDialog propDialog = new ChartPropertiesDialog(this, chart);
        RefineryUtilities.centerDialogInParent(propDialog);
        propDialog.show();
    }//GEN-LAST:event_editChartPropertiesMenuItemActionPerformed
    
    private void stackButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_stackButtonActionPerformed
        unstackButton.setSelected(false);
        prop.setStackedLines(true);
    }//GEN-LAST:event_stackButtonActionPerformed
    
    private void unstackButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_unstackButtonActionPerformed
        stackButton.setSelected(false);
        prop.setStackedLines(false);
    }//GEN-LAST:event_unstackButtonActionPerformed
    
    private void undoZoomButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_undoZoomButtonActionPerformed
        ((ChartPanel) chartPanel).restoreAutoBounds();
    }//GEN-LAST:event_undoZoomButtonActionPerformed
    
    private void printChartMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_printChartMenuItemActionPerformed
        // TODO: fix chart printing
        ((ChartPanel) chartPanel).createChartPrintJob();
    }//GEN-LAST:event_printChartMenuItemActionPerformed
    
    private void correlateButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_correlateButtonActionPerformed
        String corrColumn = itemList.getSelectedValue().toString();
        // TODO: This only correlates items in correlationDataCollection
        DataCollection dc = correlationDataCollection;
        CorrelationDialog corrDialog = new CorrelationDialog(
                this,       // parent
                corrColumn, // column
                dc          // data collection
                );
        RefineryUtilities.centerDialogInParent(corrDialog);
        corrDialog.show();
        for (Iterator li = itemListModel.iterator(); li.hasNext(); ) {
            String name = li.next().toString();
            if (! name.equals(corrColumn) && dc.contains(name)) {
                double r = dc.correlation(corrColumn, name);
                if (! Double.isNaN(r) && Math.abs(r) >= CORRELATION_THRESHOLD) {
                    corrDialog.insert(r, name);
                }
            }
        }
        corrDialog.signalListComplete();
    }//GEN-LAST:event_correlateButtonActionPerformed
    
    private void exportChartMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exportChartMenuItemActionPerformed
        try {
            ((ChartPanel) chartPanel).doSaveAs();
        } catch (IOException e) {
            Main.reportError(this, e, getResourceString("I/O_Error"));
        }
    }//GEN-LAST:event_exportChartMenuItemActionPerformed
    
    private void propertiesMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_propertiesMenuItemActionPerformed
        JTextArea textArea = new TextAreaBox(
                this,
                getResourceString("Properties"),
                "",
                30, 60
                ).getTextArea();
        
        Iterator ci = collections.iterator();
        for (int collNum = 1; ci.hasNext(); collNum++) {
            DataCollection dc = (DataCollection) ci.next();
            
            String header = getFormattedResourceString("Properties_Header", new Object[] {
                new Integer(collNum),
                        (dc instanceof FileDataCollection)
                        ? ((FileDataCollection)dc).getFilename()
                        : getResourceString("N/A"),
                        dc.getFormat(),
                        dc.getNodename()
            });
            textArea.append(header);
            
            String[] info = dc.getInfo();
            for (int i = 0; i < info.length; i++) {
                textArea.append(
                        getFormattedResourceString("______info", new Object[] {info[i]})
                        );
            }
            textArea.append(NEWLINE);
            
            textArea.append(getFormattedResourceString("______Column_count:_", new Object[] {
                new Integer(dc.getItemCount())
            }));
            
            textArea.append(getFormattedResourceString("______Row_count:_", new Object[] {
                new Integer(dc.getRecordCount((String) dc.getAllItemNames().get(0)))
            }));
            textArea.append(NEWLINE);
        }
    }//GEN-LAST:event_propertiesMenuItemActionPerformed
    
    private void removeZeroItemsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeZeroItemsMenuItemActionPerformed
        JTextArea textArea = new TextAreaBox(
                this,
                getResourceString("Remove_Items_with_Zero_Values"),
                getResourceString("The_following_items_have_been_removed:"),
                20, 40
                ).getTextArea();
        
        // Remove zero items from itemListModel
        List selectedItems = getSelectedItems();
        pushListModel();
        for (Iterator li = itemListModel.iterator(); li.hasNext(); ) {
            Object item = li.next();
            String name = item.toString();
            boolean isZero = true;
            for (Iterator ci = collections.iterator(); ci.hasNext(); ) {
                DataCollection dc = (DataCollection) ci.next();
                if (dc.contains(name)
                && (dc.getDataMin(name) != 0.0 || dc.getDataMax(name) != 0.0)) {
                    isZero = false;
                    break;
                }
            }
            if (isZero) {
                li.remove();
                if (selectedItems.contains(item))
                    selectedItems.remove(item);
                textArea.append(name);
                textArea.append(NEWLINE);
            }
        }
        
        setSelectedItems(selectedItems);
        itemList.revalidate();
    }//GEN-LAST:event_removeZeroItemsMenuItemActionPerformed
    
    private void aboutMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_aboutMenuItemActionPerformed
        String appTitle = getFormattedResourceString("About_", new Object[] {APP_NAME});
        
        String appVersion = getFormattedResourceString("APP_VERSION_STRING", new Object[] {
            getResourceString("APP_VERSION"), getResourceString("TLVIZ_VERSION")
        });
        
        // Build app info
        List appInfoList = new ArrayList();
        boolean found = true;
        for (int i = 0; found; i++) {
            try {
                appInfoList.add(getResourceString("Main.APP_INFO_"+i));
            } catch (MissingResourceException e) {
                found = false;
            }
        }
        String [] appInfo = (String[]) appInfoList.toArray(new String[appInfoList.size()]);
        
        // Create an info dialog
        JDialog aboutDialog = new JDialog(
                this,       // owner
                appTitle,   // title
                true        // modal?
                );
        aboutDialog.getContentPane().add(new AboutPanel(
                APP_NAME,
                appVersion,
                getResourceString("APP_COPYRIGHT"),
                appInfo
                ));
        aboutDialog.pack();
        RefineryUtilities.centerDialogInParent(aboutDialog);
        aboutDialog.show();
    }//GEN-LAST:event_aboutMenuItemActionPerformed
    
    private void setFooterMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setFooterMenuItemActionPerformed
        String newFooter = doSetTitleDialog(getResourceString("Footer_Text"), prop.getFooterText());
        if (newFooter != null) {
            prop.setFooterText(newFooter);
        }
    }//GEN-LAST:event_setFooterMenuItemActionPerformed
    
    private void setSubtitleMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setSubtitleMenuItemActionPerformed
        String newSubTitle = doSetTitleDialog(getResourceString("Sub-Title"), prop.getSubTitle());
        if (newSubTitle != null) {
            prop.setSubTitle(newSubTitle);
        }
    }//GEN-LAST:event_setSubtitleMenuItemActionPerformed
    
    private void setTitleMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_setTitleMenuItemActionPerformed
        String newTitle = doSetTitleDialog(getResourceString("Title"), prop.getTitle());
        if (newTitle != null) {
            if (isDefaultTitle && ! newTitle.equals(prop.getTitle()))
                isDefaultTitle = false;
            prop.setTitle(newTitle);
        }
    }//GEN-LAST:event_setTitleMenuItemActionPerformed
    
    private void itemListValueChanged(javax.swing.event.ListSelectionEvent evt) {//GEN-FIRST:event_itemListValueChanged
        if (! evt.getValueIsAdjusting()) {
            prop.setDataset(generateDataset());
        }
    }//GEN-LAST:event_itemListValueChanged
    
    private void closeMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_closeMenuItemActionPerformed
        // Save the selected items
        List selectedItems = getSelectedItems();
        
        // Clear itemListModel and prevListModels
        itemListModel.clear();
        prevListModels.clear();
        
        // "close" last file
        collections.remove(collections.size()-1);
        
        if (collections.isEmpty()) {
            statusLine.setText(getResourceString("Main.DEFAULT_STATUSLINE"));
            leftPanel.setVisible(false);
            setMenuItemsEnabled(-1);
            correlateButton.setEnabled(false);
            chart = null;
            ((ChartPanel) chartPanel).setChart(chart);
        } else {
            DataCollection last = (DataCollection) collections.get(collections.size()-1);
            statusLine.setText(
                    last instanceof FileDataCollection
                    ? ((FileDataCollection) last).getFilename()
                    : getResourceString("Main.DEFAULT_STATUSLINE")
                    );
            
            if (isDefaultTitle)
                setChartDefaultTitle();
            
            // Re-populate itemListModel
            Iterator ci = collections.iterator();
            for (int baseIndex = MAX_ITEMS_PER_DC; ci.hasNext(); baseIndex += MAX_ITEMS_PER_DC)
                itemListModel.addAll((DataCollection) ci.next(), baseIndex);
            
            // Remove selectedItemNames no longer in list model
            for (Iterator si = selectedItems.listIterator(); si.hasNext(); )
                if (! itemListModel.contains(si.next()))
                    si.remove();
            
            setSelectedItems(selectedItems);
        }
        itemList.revalidate();
    }//GEN-LAST:event_closeMenuItemActionPerformed
    
    private void openMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openMenuItemActionPerformed
        // Show file chooser.
        JFileChooser fileChooser = new JFileChooser(lastFile);
        fileChooser.setDialogTitle(getResourceString("Open_data_collection_file"));
        fileChooser.addChoosableFileFilter(zipFilter);
        fileChooser.addChoosableFileFilter(csvFilter);
        
        int result = fileChooser.showOpenDialog(this);
        
        // Process the results.
        if (result == fileChooser.APPROVE_OPTION) {
            lastFile = fileChooser.getSelectedFile();
            FileDataCollection dc = processFile(lastFile);
            addDCtoItemlist(dc);
        }
    }//GEN-LAST:event_openMenuItemActionPerformed
    
    private void exitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitMenuItemActionPerformed
        System.exit(0);
    }//GEN-LAST:event_exitMenuItemActionPerformed
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem aboutMenuItem;
    private javax.swing.JMenuItem addItemMenuItem;
    private javax.swing.JCheckBoxMenuItem candleChartMenuItem;
    private javax.swing.JMenuItem chartMultipleItemsMenuItem;
    private javax.swing.JPanel chartPanel;
    private javax.swing.JMenu chartSettingsMenu;
    private javax.swing.JMenuItem closeMenuItem;
    private javax.swing.JButton correlateButton;
    private javax.swing.JMenuItem cycleDualMenuItem;
    private javax.swing.JMenuItem editChartPropertiesMenuItem;
    private javax.swing.JMenuItem exitMenuItem;
    private javax.swing.JMenuItem exportChartMenuItem;
    private javax.swing.JMenu fileMenu;
    private javax.swing.JMenu helpMenu;
    private javax.swing.JList itemList;
    private javax.swing.JPanel leftPanel;
    private javax.swing.JMenuItem listOriginalOrderMenuItem;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JMenu modifyItemlistMenu;
    private javax.swing.JCheckBoxMenuItem movingAverageMenuItem;
    private javax.swing.JMenuItem openMenuItem;
    private javax.swing.JMenu optionsMenu;
    private javax.swing.JMenuItem printChartMenuItem;
    private javax.swing.JMenuItem printPreviewMenuItem;
    private javax.swing.JMenuItem propertiesMenuItem;
    private javax.swing.JCheckBoxMenuItem ratioChartsMenuItem;
    private javax.swing.JMenuItem removeContainStringMenuItem;
    private javax.swing.JMenuItem removeNotContainStringMenuItem;
    private javax.swing.JMenuItem removeNotSelectedMenuItem;
    private javax.swing.JMenuItem removeSelectedMenuItem;
    private javax.swing.JMenuItem removeZeroItemsMenuItem;
    private javax.swing.JMenuItem revertToPreviousListboxMenuItem;
    private javax.swing.JMenuItem saveAsMenuItem;
    private javax.swing.JMenuItem setFooterMenuItem;
    private javax.swing.JMenuItem setMovingAveragePeriodMenuItem;
    private javax.swing.JMenuItem setPrimaryLineWidthMenuItem;
    private javax.swing.JMenuItem setSecondaryLineWidthMenuItem;
    private javax.swing.JMenuItem setSmoothingFactorMenuItem;
    private javax.swing.JMenuItem setSubtitleMenuItem;
    private javax.swing.JMenuItem setTitleMenuItem;
    private javax.swing.JCheckBoxMenuItem smoothMenuItem;
    private javax.swing.JMenuItem sortAlphabeticallyMenuItem;
    private javax.swing.JMenuItem sortMedianMenuItem;
    private javax.swing.JMenuItem sortOrderAscendingMenuItem;
    private javax.swing.JMenuItem sortOrderDescendingMenuItem;
    private javax.swing.JMenu sortOrderMenuItem;
    private javax.swing.JToggleButton stackButton;
    private javax.swing.JLabel statusLine;
    private javax.swing.JButton undoZoomButton;
    private javax.swing.JToggleButton unstackButton;
    private javax.swing.JMenuItem viewReleaseNotesMenuItem;
    // End of variables declaration//GEN-END:variables
    
}
