/*
 * TSChart.java
 */

package de.pdv_systeme.TSChart;

import java.util.Iterator;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;

import java.awt.Color;
import java.awt.Paint;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.PeriodAxis;
import org.jfree.chart.axis.PeriodAxisLabelInfo;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.axis.AxisLocation;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.plot.CombinedDomainXYPlot;
import org.jfree.chart.renderer.xy.StackedXYAreaRenderer;
import org.jfree.chart.renderer.xy.StandardXYItemRenderer;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.LegendItem;
import org.jfree.chart.LegendItemCollection;
import org.jfree.chart.LegendItemSource;
import org.jfree.chart.block.BlockBorder;
import org.jfree.data.Range;
import org.jfree.data.time.Day;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeTableXYDataset;
import org.jfree.ui.HorizontalAlignment;
import org.jfree.ui.RectangleInsets;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.VerticalAlignment;

import de.pdv_systeme.util.VmsDate;

/**
 * Draw a chart where the domain axis is a TimeSeries.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class TSChart
        extends JFreeChart
        implements LegendItemSource {
    
    // TODO: make legend items' checkboxes work to select visibility of line
    
    // TODO: make the date axis work with date and time on separate lines
    /** Flag: Draw a separate axis for the days (doesn't work yet) */
    private static final boolean WITH_DAY_AXIS = false;
    
    /** The chart's panel */
    private ChartPanel chartPanel;
    
    /** The chart's properties */
    private TSChartProperty prop;
    
    /* The legend items */
    private LegendItemCollection legendItems = new LegendItemCollection();
    
    private static final Font TITLE_FONT = DEFAULT_TITLE_FONT;
    
    /* String displayed if prop.isStackedLines() */
    private String stackedString = "Values are STACKED";
    
    /**
     * Create a new instance of TSChart.
     * @param parent  this chart's panel
     * @param prop  the chart's properties
     */
    public TSChart(ChartPanel parent, TSChartProperty prop) {
        
        super(
                prop.getTitle(), TITLE_FONT,
                isDualPlot(prop) ? new CombinedDomainXYPlot() : new XYPlot(),
                false        // draw legend?
                );
        
        this.chartPanel = parent;
        this.prop = prop;
        
        setAntiAlias(false);
        generatePlot();
        
        if (getTitle() != null)
            getTitle().setPaint(prop.chartTitleColor);
        setBackgroundPaint(prop.chartBackgroundColor);
        
        generateSubtitles();
        
        // replace default Mouse/MouseMotionListener
        chartPanel.removeMouseListener(chartPanel);
        chartPanel.removeMouseMotionListener(chartPanel);
        MyMouseListener myMouseListener = new MyMouseListener();
        chartPanel.addMouseListener(myMouseListener);
        chartPanel.addMouseMotionListener(myMouseListener);
    }
    
    private static boolean isDualPlot(TSChartProperty prop) {
        return prop.isDualChart() && ! prop.isDualYAxis() && ! prop.isStackedLines();
    }
    
    /**
     * Generate a plot.
     * As a side effect, populates the legendItems.
     */
    private void generatePlot() {
        legendItems = new LegendItemCollection();
        if (isDualPlot(prop)) {
            generateDualPlot();
        } else {
            generateSinglePlot();
        }
    }
    
    /**
     * Generate a dual plot.
     * As a side effect, populates the legendItems.
     */
    private void generateDualPlot() {
        CombinedDomainXYPlot plot = (CombinedDomainXYPlot) getPlot();
        TSChartDrawingSupplier ds = new TSChartDrawingSupplier(prop);
        
        plot.setDrawingSupplier(ds);
        plot.setAxisOffset(new RectangleInsets(5.0, 5.0, 5.0, 5.0));
        plot.setGap(10.0);
        
        generateDomainAxis(plot);
        
        XYToolTipGenerator toolTipGenerator = generateToolTipGenerator();
        
        /*
         * For every series, generate a subplot
         */
        TimeSeriesCollection tsc = (TimeSeriesCollection) prop.getDataset();
        for (int series = 0; series < tsc.getSeriesCount(); series++) {
            TimeSeries ts = tsc.getSeries(series);
            
            XYPlot subplot = new XYPlot();
            subplot.setBackgroundPaint(prop.plotBackgroundColor);
            subplot.setDomainGridlinePaint(prop.plotGridlineColor);
            subplot.setRangeGridlinePaint(prop.plotGridlineColor);
            
            subplot.setDataset(new TimeSeriesCollection(ts));
            
            StandardXYItemRenderer renderer = new StandardXYItemRenderer(
                    StandardXYItemRenderer.LINES,
                    toolTipGenerator);
            renderer.setDrawSeriesLineAsPath(true);
            renderer.setSeriesPaint(0, ds.getNextPaint());
            subplot.setRenderer(renderer);
            
            NumberAxis valueAxis = new NumberAxis();
            valueAxis.setAutoRangeIncludesZero(false);
            subplot.setRangeAxis(valueAxis);
            subplot.setRangeAxisLocation(series == 0
                    ? AxisLocation.BOTTOM_OR_LEFT
                    : AxisLocation.TOP_OR_LEFT);
            
            Paint seriesPaint = renderer.getSeriesPaint(0);
            addLegendItem(ts.getKey(), seriesPaint, true);
            
            plot.add(subplot);
        }
    }
    
    /**
     * Generate a single plot.
     * As a side effect, populates the legendItems.
     */
    private void generateSinglePlot() {
        XYPlot plot = getXYPlot();
        plot.setBackgroundPaint(prop.plotBackgroundColor);
        plot.setDomainGridlinePaint(prop.plotGridlineColor);
        plot.setRangeGridlinePaint(prop.plotGridlineColor);
        plot.setDrawingSupplier(new TSChartDrawingSupplier(prop));
        plot.setAxisOffset(new RectangleInsets(5.0, 5.0, 5.0, 5.0));
        
        generateDomainAxis(plot);
        
        XYToolTipGenerator toolTipGenerator = generateToolTipGenerator();
        
        if (prop.isStackedLines()) {
            plot.setDataset(prop.getDataset());
            if (WITH_DAY_AXIS) {
                plot.mapDatasetToDomainAxis(0, 1);
            }
            
            StackedXYAreaRenderer renderer = new StackedXYAreaRenderer(
                    StackedXYAreaRenderer.AREA,
                    toolTipGenerator,
                    null);
            plot.setRenderer(renderer);
            
            NumberAxis valueAxis = new NumberAxis();
            valueAxis.setAutoRangeIncludesZero(false);
            plot.setRangeAxis(valueAxis);
            plot.mapDatasetToRangeAxis(0, 0);
            
            TimeTableXYDataset tsc = (TimeTableXYDataset) prop.getDataset();
            for (int series = 0; series < tsc.getSeriesCount(); series++) {
                Paint seriesPaint = renderer.getSeriesPaint(series);
                addLegendItem(tsc.getSeriesKey(series), seriesPaint, true);
            }
        } else {
            /*
             * Split XYDataset into single datasets
             * to get control of the range axis mapping and the colors
             */
            TimeSeriesCollection tsc = (TimeSeriesCollection) prop.getDataset();
            for (int series = 0; series < tsc.getSeriesCount(); series++) {
                TimeSeries ts = tsc.getSeries(series);
                
                plot.setDataset(series, new TimeSeriesCollection(ts));
                if (WITH_DAY_AXIS) {
                    plot.mapDatasetToDomainAxis(series, 1);
                }
                
                StandardXYItemRenderer renderer = new StandardXYItemRenderer(
                        StandardXYItemRenderer.LINES,
                        toolTipGenerator);
                renderer.setDrawSeriesLineAsPath(true);
                plot.setRenderer(series, renderer);
                
                Paint seriesPaint = renderer.getSeriesPaint(0);
                if (series == 0 || prop.isDualYAxis()) {
                    NumberAxis valueAxis = new NumberAxis();
                    valueAxis.setAutoRangeIncludesZero(false);
                    if (prop.isDualYAxis()) {
                        valueAxis.setAxisLinePaint(seriesPaint);
                        valueAxis.setTickLabelPaint(seriesPaint);
                        valueAxis.setLabelPaint(seriesPaint);
                    }
                    plot.setRangeAxis(series, valueAxis);
                    plot.mapDatasetToRangeAxis(series, series);
                } else {
                    plot.mapDatasetToRangeAxis(series, 0);
                }
                
                addLegendItem(ts.getKey(), seriesPaint, true);
            }
        }
    }
    
    /** Create the domain axis */
    private void generateDomainAxis(XYPlot plot) {
        DateAxis timeAxis = new DateAxis();
        timeAxis.setLowerMargin(0.01);  // reduce the default margins
        timeAxis.setUpperMargin(0.01);
        timeAxis.setDateFormatOverride(new SimpleDateFormat(
                WITH_DAY_AXIS
                ? VmsDate.TIME_SECONDS_STRING
                : VmsDate.TIME_SECONDS_STRING+" "+System.getProperty("line.separator")+"("+VmsDate.DATE_STRING+")",
                Locale.US
                ));
        plot.setDomainAxis(timeAxis);
        
        if (WITH_DAY_AXIS) {
            PeriodAxis dateAxis = new PeriodAxis("");
            PeriodAxisLabelInfo[] info = new PeriodAxisLabelInfo[1];
            info[0] = new PeriodAxisLabelInfo(Day.class, VmsDate.dateFormatDate);
            dateAxis.setLabelInfo(info);
            dateAxis.setAxisLineVisible(false);
            plot.setDomainAxis(1, dateAxis);
            plot.setDomainAxisLocation(1, plot.getDomainAxisLocation());
        }
    }
    
    /**
     * Create a tooltip generator.
     * @return the tooltip generator.
     */
    private XYToolTipGenerator generateToolTipGenerator() {
        return new StandardXYToolTipGenerator(
                StandardXYToolTipGenerator.DEFAULT_TOOL_TIP_FORMAT,
                VmsDate.dateFormatSeconds,
                NumberFormat.getInstance()
                );
    }
    
    
    /**
     * Set the string displayed if chart is stacked
     * @param stackedString  the string to display
     */
    public void setStackedString(String stackedString) {
        this.stackedString = stackedString;
        generateSubtitles();
    }
    
    /**
     * Generate the chart's subtitles (sub-title, stacked string, footer, legend).
     */
    public void generateSubtitles() {
        clearSubtitles();
        
        Color titleColor = prop.chartTitleColor;
        String titleFontName = TITLE_FONT.getName();
        int titleFontStyle = TITLE_FONT.getStyle();
        int titleFontSize = TITLE_FONT.getSize();
        
        String subTitleText = prop.getSubTitle();
        if (subTitleText != null) {
            TextTitle subTitle = new TextTitle(
                    subTitleText,
                    new Font(titleFontName, titleFontStyle, titleFontSize * 2/3),
                    titleColor,
                    RectangleEdge.TOP,
                    HorizontalAlignment.CENTER,
                    VerticalAlignment.CENTER,
                    TextTitle.DEFAULT_PADDING
                    );
            addSubtitle(subTitle);
        }
        
        // generate sub-titles below the chart (including the legend)
        // in reverse order  (i.e. from the bottom towards the chart)
        String footerText = prop.getFooterText();
        if (footerText != null) {
            TextTitle footerTitle = new TextTitle(
                    footerText,
                    new Font(titleFontName, Font.ITALIC, titleFontSize / 2),
                    titleColor,
                    RectangleEdge.BOTTOM,
                    HorizontalAlignment.CENTER,
                    VerticalAlignment.CENTER,
                    TextTitle.DEFAULT_PADDING
                    );
            addSubtitle(footerTitle);
        }
        
        if (prop.isStackedLines()) {
            TextTitle stackedTitle = new TextTitle(
                    stackedString,
                    new Font(titleFontName, Font.ITALIC, titleFontSize / 2),
                    Color.RED,
                    RectangleEdge.BOTTOM,
                    HorizontalAlignment.CENTER,
                    VerticalAlignment.CENTER,
                    TextTitle.DEFAULT_PADDING
                    );
            addSubtitle(stackedTitle);
        }
        
        // create legend
        TSChartLegendTitle legend = new TSChartLegendTitle(this);
        legend.setMargin(new RectangleInsets(1.0, 1.0, 1.0, 1.0));
        legend.setBorder(new BlockBorder());
        legend.setBackgroundPaint(Color.white);
        legend.setPosition(RectangleEdge.BOTTOM);
        addLegend(legend); // == addSubtitle(legend);
    }
    
    /**
     * Add a legend item.
     * @param key
     * @param labelPaint
     * @param isVisible
     */
    private void addLegendItem(
            Comparable key,
            Paint labelPaint,
            boolean isVisible
            ) {
        String label = (String) key;
        LegendItem item = new TSChartLegendItem(
                label, label, isVisible, TSChartLabelBlock.DEFAULT_FONT, labelPaint);
        if (item != null) {
            legendItems.add(item);
        }
    }
    
    /*
     * Implementation of the LegendItemSource interface
     */
    /** {@inheritDoc} */
    public LegendItemCollection getLegendItems() {
        if (legendItems.getItemCount() == 0) {
            generatePlot();
        }
        return legendItems;
    }
    
    
    /**
     * Mouse/MouseMotion listener.
     * <p>
     * This serves two purposes:
     * <ul>
     *  <li>dragging (with the right button pressed) the mouse over the chart
     *  <li>clicking with the left button in a legend item.
     * </ul>
     * All events - except for those above - are passed on to the chart panel
     * (as we disables its Mouse and MouseMotion listener).
     */
    private class MyMouseListener implements MouseListener, MouseMotionListener {
        
        private final boolean debugMouseEvents
                = (System.getProperty("JTLviz.debugMouseEvents") != null);
        
        /** Latest mouse coordinate */
        private int lastx, lasty;
        
        /**
         * Shift the domain axis
         * @param plot  the plot
         * @param shift  the factor by which to shift
         */
        private void shiftDomainAxis(final XYPlot plot, final double shift) {
            for (int i = 0; i < plot.getDomainAxisCount(); i++) {
                ValueAxis dom = plot.getDomainAxis(i);
                Range xr = dom.getRange();
                dom.setRange(Range.shift(xr, shift*xr.getLength(), true));
            }
        }
        
        /**
         * Shift the range axis
         * @param plot  the plot
         * @param shift  the factor by which to shift
         */
        private void shiftValueAxis(final XYPlot plot, final double shift) {
            for (int i = 0; i < plot.getRangeAxisCount(); i++) {
                ValueAxis rng = plot.getRangeAxis(i);
                Range yr = rng.getRange();
                rng.setRange(Range.shift(yr, shift*yr.getLength(), true));
            }
        }
        
        /**
         * Returns true iff button 3 is pressed.
         * @return true iff button 3 is pressed
         */
        private boolean isButton3Down(final MouseEvent e) {
            return ((e.getModifiersEx() & MouseEvent.BUTTON3_DOWN_MASK) == MouseEvent.BUTTON3_DOWN_MASK);
        }

        /** {@inheritDoc} */
        public void mouseClicked(final MouseEvent e) {
            if (debugMouseEvents && e.getButton() == MouseEvent.BUTTON1) {
                setNotify(true);  // force a redraw
                ChartRenderingInfo info = chartPanel.getChartRenderingInfo();
                if (info != null) {
                    EntityCollection entities = info.getEntityCollection();
                    if (entities != null) {
                        Insets insets = chartPanel.getInsets();
                        ChartEntity entity = entities.getEntity(
                                e.getX() - insets.left,
                                e.getY() - insets.top
                                );
                        if (entity != null)
                            System.err.println(e.paramString()+": "+entity.toString());
                        else
                            System.err.println(e.paramString()+": Entity is null");
                    } else {
                        System.err.println(e.paramString()+": EntityCollection is null");
                    }
                } else {
                    System.err.println(e.paramString()+": ChartRenderingInfo is null");
                }
            } else {
                chartPanel.mouseClicked(e);
            }
        }
        
        /** {@inheritDoc} */
        public void mousePressed(final MouseEvent e) {
            if (isButton3Down(e)) {
                lastx = e.getX();
                lasty = e.getY();
            } else {
                chartPanel.mousePressed(e);
            }
        }
        
        /** {@inheritDoc} */
        public void mouseReleased(final MouseEvent e) {
            chartPanel.mouseReleased(e);
        }
        
        /** {@inheritDoc} */
        public void mouseEntered(final MouseEvent e) {
            chartPanel.mouseEntered(e);
        }
        
        /** {@inheritDoc} */
        public void mouseExited(final MouseEvent e) {
            chartPanel.mouseExited(e);
        }
        
        /** {@inheritDoc} */
        public void mouseDragged(final MouseEvent e) {
            if (isButton3Down(e)) {
                int
                        currx = e.getX(),
                        curry = e.getY();
                double
                        ddx = (lastx - currx) / 400.0,
                        ddy = (curry - lasty) / 400.0;
                XYPlot plot = getXYPlot();
                if (plot instanceof CombinedDomainXYPlot) {
                    // shift all subplots
                    CombinedDomainXYPlot cdplot = (CombinedDomainXYPlot) plot;
                    shiftDomainAxis(cdplot, ddx);
                    for (Iterator si = cdplot.getSubplots().iterator(); si.hasNext(); ) {
                        shiftValueAxis((XYPlot) si.next(), ddy);
                    }
                } else { // plot instanceof XYPlot
                    shiftDomainAxis(plot, ddx);
                    shiftValueAxis(plot, ddy);
                }
                
                lastx = currx;
                lasty = curry;
            } else {
                chartPanel.mouseDragged(e);
            }
        }
        
        /** {@inheritDoc} */
        public void mouseMoved(final MouseEvent e) {
            chartPanel.mouseMoved(e);
        }
        
        /* This was the ItemListener callback of the legend items' checkboxes */
//        private void _itemStateChanged(ItemEvent e) {
//            JCheckBox checkBox = (JCheckBox) e.getSource();
//            Boolean isSelected = Boolean.valueOf(e.getStateChange() == ItemEvent.SELECTED);
//            String name = checkBox.getName();
//            int idx = prop.getDataset().indexOf(name);
//            if (isDualPlot(prop)) {
//                CombinedDomainXYPlot plot = (CombinedDomainXYPlot) getPlot();
//                XYPlot subplot = (XYPlot) plot.getSubplots().get(idx);
//                StandardXYItemRenderer renderer = (StandardXYItemRenderer) subplot.getRenderer();
//                renderer.setSeriesVisible(0, isSelected);
//            } else if (prop.isStackedLines()) {
//                XYPlot plot = getXYPlot();
//                StackedXYAreaRenderer renderer = (StackedXYAreaRenderer) plot.getRenderer(idx);
//                //  - not possible with StackedXYAreaRenderer to set visibility -
//            } else {
//                XYPlot plot = getXYPlot();
//                StandardXYItemRenderer renderer = (StandardXYItemRenderer) plot.getRenderer(idx);
//                renderer.setSeriesVisible(0, isSelected);
//            }
//        }
        
    }
    
}
