/*
 * TSChartLabelBlock.java
 */

package de.pdv_systeme.TSChart;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;

import org.jfree.chart.block.AbstractBlock;
import org.jfree.chart.block.Block;
import org.jfree.chart.block.BlockResult;
import org.jfree.chart.block.EntityBlockParams;
import org.jfree.chart.block.RectangleConstraint;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.StandardEntityCollection;
import org.jfree.text.TextBlock;
import org.jfree.text.TextBlockAnchor;
import org.jfree.text.TextUtilities;
import org.jfree.ui.Size2D;

/**
 * An alternative to LabelBlock with the ability to set the label font color.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class TSChartLabelBlock
        extends AbstractBlock // should really extend LabelBlock
        implements Block {
    
    /** The default font. */
    public static final Font DEFAULT_FONT = new Font("SansSerif", Font.PLAIN, 10);
    
    /** The default paint. */
    public static final Paint DEFAULT_PAINT = Color.black;
    
    // TODO: properly (de-)serialize member variables
    /**
     * The text for the label - retained in case the label needs
     * regenerating (for example, to change the font).
     */
    private String text;
    
    /** The label. */
    private TextBlock label;
    
    /** The font. */
    private Font font;
    
    /** The paint. */
    private Paint paint;
    
    /** The tool tip text (can be <code>null</code>). */
    private String toolTipText;
    
    /** The URL text (can be <code>null</code>). */
    private String urlText;
    
    /**
     * Creates a new label block.
     *
     * @param text  the label text.
     */
    public TSChartLabelBlock(String text) {
        this(text, DEFAULT_FONT, DEFAULT_PAINT);
    }
    
    /**
     * Creates a new label block.
     *
     * @param text  the label text.
     * @param font  the font.
     * @param paint  the paint.
     */
    public TSChartLabelBlock(String text, Font font, Paint paint) {
        this.text = text;
        this.label = TextUtilities.createTextBlock(text, font, paint);
        this.font = font;
        this.paint = paint;
        this.toolTipText = null;
        this.urlText = null;
    }
    
    /**
     * Returns the font.
     *
     * @return The font.
     */
    public Font getFont() {
        return this.font;
    }
    
    /**
     * Sets the font and regenerates the label.
     *
     * @param font  the font.
     */
    public void setFont(Font font) {
        this.font = font;
        this.label = TextUtilities.createTextBlock(
                this.text, font, this.paint
                );
    }
    
    /**
     * Returns the paint.
     *
     * @return The paint.
     */
    public Paint getPaint() {
        return this.paint;
    }
    
    /**
     * Sets the paint and regenerates the label.
     *
     * @param paint  the paint.
     */
    public void setPaint(Paint paint) {
        this.paint = paint;
        this.label = TextUtilities.createTextBlock(
                this.text, this.font, paint
                );
    }
    
    /**
     * Returns the tool tip text.
     *
     * @return The tool tip text (possibly <code>null</code>).
     */
    public String getToolTipText() {
        return this.toolTipText;
    }
    
    /**
     * Sets the tool tip text.
     *
     * @param text  the text (<code>null</code> permitted).
     */
    public void setToolTipText(String text) {
        this.toolTipText = text;
    }
    
    /**
     * Returns the URL text.
     *
     * @return The URL text (possibly <code>null</code>).
     */
    public String getURLText() {
        return this.urlText;
    }
    
    /**
     * Sets the URL text.
     *
     * @param text  the text (<code>null</code> permitted).
     */
    public void setURLText(String text) {
        this.urlText = text;
    }
    
    /**
     * Arranges the contents of the block, within the given constraints, and
     * returns the block size.
     *
     * @param g2  the graphics device.
     * @param constraint  the constraint (<code>null</code> not permitted).
     *
     * @return The block size (in Java2D units, never <code>null</code>).
     */
    public Size2D arrange(Graphics2D g2, RectangleConstraint constraint) {
        RectangleConstraint contentConstraint = toContentConstraint(constraint);
        g2.setFont(this.font);
        Size2D s = this.label.calculateDimensions(g2);
        return new Size2D(
                calculateTotalWidth(s.getWidth()),
                calculateTotalHeight(s.getHeight())
                );
    }
    
    /** {@inheritDoc} */
    public void draw(Graphics2D g2, Rectangle2D area) {
        draw(g2, area, null);
    }
    
    /**
     * Draws the block within the specified area.
     *
     * @param g2  the graphics device.
     * @param area  the area.
     * @param params  the {@link org.jfree.chart.block.EntityBlockParams EntityBlockParams}.
     *
     * @return a {@link org.jfree.chart.block.BlockResult BlockResult}.
     */
    public Object draw(Graphics2D g2, Rectangle2D area, Object params) {
        area = trimMargin(area);
        drawBorder(g2, area);
        area = trimBorder(area);
        area = trimPadding(area);
        
        // check if we need to collect chart entities from the container
        EntityBlockParams ebp = null;
        StandardEntityCollection sec = null;
        Shape entityArea = null;
        if (params instanceof EntityBlockParams) {
            ebp = (EntityBlockParams) params;
            if (ebp.getGenerateEntities()) {
                sec = new StandardEntityCollection();
                // TODO:  this transformation doesn't work always.  Fix!
                entityArea = g2.getTransform().createTransformedShape(area);
            }
        }
        
        g2.setPaint(this.paint);
        g2.setFont(this.font);
        this.label.draw(
                g2, (float) area.getX(), (float) area.getY(),
                TextBlockAnchor.TOP_LEFT
                );
        
        BlockResult result = null;
        if (ebp != null && sec != null) {
            // this should really be a LegendItemEntity
            ChartEntity entity = new ChartEntity(
                    entityArea, this.toolTipText, this.urlText
                    );
            sec.add(entity);
            result = new BlockResult();
            result.setEntityCollection(sec);
        }
        return result;
    }
    
}
