/*
 * TSChartLegendItem.java
 */

package de.pdv_systeme.TSChart;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.ItemListener;
import java.awt.geom.Line2D;

import org.jfree.chart.LegendItem;

/**
 * A subclass of LegendItem to gain control of the color and font.
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class TSChartLegendItem extends LegendItem {
    
    /**
     * The shape must be non-null for a LegendItem - if no shape is required,
     * use this.
     */
    private static final Shape UNUSED_SHAPE = new Line2D.Float();
    
    /**
     * The stroke must be non-null for a LegendItem - if no stroke is required,
     * use this.
     */
    private static final Stroke UNUSED_STROKE = new BasicStroke(0.0f);
    
    /**
     * The paint must be non-null for a LegendItem - if no paint is required,
     * use this.
     */
    private static final Paint UNUSED_PAINT = Color.black;

    /** The label font. */
    private transient Font labelFont;
    
    /** The label paint. */
    private transient Paint labelPaint;
    
    /** Flag: is the associated series visible? */
    private transient boolean seriesVisible;
    
    /**
     * Creates a new legend item.
     *
     * @param label  the label (<code>null</code> not permitted).
     * @param description  the description (not currently used,
     *        <code>null</code> permitted).
     * @param seriesVisible  a flag that controls whether or not the associated
     *                     series is visible.
     * @param labelFont  the label font (<code>null</code> not permitted).
     * @param labelPaint  the label paint (<code>null</code> not permitted).
     */
    public TSChartLegendItem(
            String label,
            String description,
            boolean seriesVisible,
            Font labelFont,
            Paint labelPaint
            ) {
        super(
                label,
                description,
                label,          // toolTipText
                null,           // urlText
                false,          // shapeVisble
                UNUSED_SHAPE,   // shape
                false,          // shapeFilled
                UNUSED_PAINT,   // fillPaint
                false,          // shapeOutlineVisible
                UNUSED_PAINT,   // outlinePaint
                UNUSED_STROKE,  // outlineStroke
                false,          // lineVisible
                UNUSED_SHAPE,   // line
                UNUSED_STROKE,  // lineStroke
                UNUSED_PAINT    // linePaint
                );
        if (labelFont == null) {
            throw new IllegalArgumentException("Null 'labelFont' argument.");
        }
        if (labelPaint == null) {
            throw new IllegalArgumentException("Null 'labelPaint' argument.");
        }
        this.seriesVisible = seriesVisible;
        this.labelFont = labelFont;
        this.labelPaint = labelPaint;
    }
    
    /**
     * Returns a flag that indicates whether or not the line is visible.
     * 
     * @return A boolean.
     */
    public boolean isSeriesVisible() {
        return this.seriesVisible;
    }
    
    /**
     * Returns the label font.
     *
     * @return The label font (never <code>null</code>).
     */
    public Font getLabelFont() {
        return this.labelFont;
    }
    
    /**
     * Returns the label paint.
     *
     * @return The label paint (never <code>null</code>).
     */
    public Paint getLabelPaint() {
        return this.labelPaint;
    }
    
    /**
     * Tests this item for equality with an arbitrary object.
     *
     * @param obj  the object (<code>null</code> permitted).
     *
     * @return A boolean.
     */
    public boolean equals(Object obj) {
        if (! super.equals(obj)) {
            return false;
        }
        TSChartLegendItem that = (TSChartLegendItem) obj;
        if (this.seriesVisible != that.seriesVisible) {
            return false;
        }
        if (!this.labelFont.equals(that.labelFont)) {
            return false;
        }
        if (!this.labelPaint.equals(that.labelPaint)) {
            return false;
        }
        return true;
    }
    
}
