/*
 * TSChartProperty.java
 */

package de.pdv_systeme.TSChart;

import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;

import javax.swing.event.EventListenerList;

import org.jfree.data.xy.XYDataset;


/**
 * Properties of TSChart
 *
 * @author Martin Vorlnder <mv@pdv-systeme.de>
 */
public class TSChartProperty {
    
    /** Name of the bound property */
    public static final String
            DATASET_PROPERTY             = "dataset",
            TITLE_PROPERTY               = "title",
            SUBTITLE_PROPERTY            = "subTitle",
            FOOTERTEXT_PROPERTY          = "footerText",
            PRIMARYLINEWIDTH_PROPERTY    = "primaryLineWidth",
            SECONDARYLINEWIDTH_PROPERTY  = "secondaryLineWidth",
            MOVINGAVERAGEPERIOD_PROPERTY = "movingAveragePeriod",
            SMOOTHINGFACTOR_PROPERTY     = "smoothingFactor",
            DUALCHART_PROPERTY           = "dualChart",
            DUALYAXIS_PROPERTY           = "dualYAxis",
            MOVINGAVERAGE_PROPERTY       = "movingAverage",
            SMOOTHCHART_PROPERTY         = "smoothChart",
            STACKEDLINES_PROPERTY        = "stackedLines",
            RATIOCHART_PROPERTY          = "ratioChart",
            CANDLECHART_PROPERTY         = "candleChart";
    
    /** the charted dataset */
    private XYDataset dataset = null;
    
    /** chart title */
    private String title = null;
    
    /** chart sub-title */
    private String subTitle = null;
    
    /** chart footer text */
    private String footerText = null;
    
    /** line width of the primary chart line */
    private int primaryLineWidth = 1; // originally 3
    
    /** line width of the secondary chart line */
    private int secondaryLineWidth = 3;
    
    /** the moving average period (in seconds?) */
    private int movingAveragePeriod = 10;
    
    /** the smoothing factor */
    private int smoothingFactor = 50;
    
    /** flag: draw dual chart? */
    private boolean dualChart = false;
    
    /** flag: draw dual Y axis? */
    private boolean dualYAxis = false;
    
    /** flag: draw moving average? */
    private boolean movingAverage = false;
    
    /** flag: smooth chart? */
    private boolean smoothChart = false;
    
    /** flag: draw stacked lines? */
    private boolean stackedLines = false;
    
    /** flag: draw ratio chart? */
    private boolean ratioChart = false;
    
    /** flag: draw candle chart? */
    private boolean candleChart = false;
    
    
    /** The text color of the chart titles */
    public static final Color chartTitleColor = Color.BLUE;
    
    /** The background color of charts */
    public static final Color chartBackgroundColor = Color.WHITE;
    
    /** The background color of plots */
    public static final Color plotBackgroundColor = Color.LIGHT_GRAY;
    
    /** The color of plot gridlines */
    public static final Color plotGridlineColor = Color.WHITE;
    
    /** The colors the series lines are painted with */
    public static final Color[] plotSeriesColor = new Color[] {
        new Color(0xFE, 0x00, 0x00),
                new Color(0x00, 0x00, 0xFE),
                new Color(0x00, 0x80, 0x01),
                new Color(0x01, 0x00, 0x80),
                new Color(0x80, 0x00, 0x00),
                new Color(0xFF, 0x00, 0xFE),
                new Color(0x00, 0x00, 0x00),
                new Color(0x81, 0x00, 0x7F),
                new Color(0xC7, 0x90, 0xC9),
                new Color(0xFA, 0xB8, 0x47),
                new Color(0x7F, 0x80, 0x00),
                new Color(0x40, 0x00, 0x3F),
                new Color(0x00, 0xDC, 0x00),
                new Color(0xC0, 0xC0, 0x00),
                new Color(0x90, 0xDC, 0x91),
                new Color(0x8F, 0x90, 0xC0)
    };
    
    /** The strokes the series lines are painted with */
    public static final Stroke[] plotSeriesStroke = new Stroke[1];
    
    
    /** The list of property change event listeners */
    protected EventListenerList listenerList = new EventListenerList();
    
    /** Create a new instance of TSChartProperty */
    public TSChartProperty() {
        initPlotSeriesStroke();
    }
    
    /**
     * Adds a listener to the list that's notified each time a change
     * to a property occurs.
     *
     * @param l  the <code>TSChartPropertyChangeListener</code> to be added
     */
    public void addPropertyChangeListener(TSChartPropertyChangeListener l) {
        listenerList.add(TSChartPropertyChangeListener.class, l);
    }
    
    /**
     * Removes a listener from the list that's notified each time a
     * change to a property occurs.
     *
     * @param l  the <code>TSChartPropertyChangeListener</code> to be removed
     */
    public void removePropertyChangeListener(TSChartPropertyChangeListener l) {
        listenerList.remove(TSChartPropertyChangeListener.class, l);
    }
    
    /**
     * Returns an array of all the listeners
     * registered on this <code>TSChartProperty</code>.
     *
     * @return all of this model's <code>TSChartPropertyChangeListener</code>s,
     *         or an empty array if no list data listeners
     *         are currently registered
     *
     * @see #addPropertyChangeListener
     * @see #removePropertyChangeListener
     *
     * @since 1.4
     */
    public TSChartPropertyChangeListener[] getPropertyChangeListeners() {
        return (TSChartPropertyChangeListener[]) listenerList.getListeners(
                TSChartPropertyChangeListener.class);
    }
    
    /**
     * Notify all listeners of a property change event.
     * @param source  the source of the event
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    protected void firePropertyChanged(
            Object source,
            String propertyName,
            Object oldValue,
            Object newValue
            ) {
        TSChartPropertyChangeListener[] listeners = getPropertyChangeListeners();
        TSChartPropertyChangeEvent e = null;
        
        for (int i = 0; i < listeners.length; i++) {
            if (e == null) {
                e = new TSChartPropertyChangeEvent(source, propertyName, oldValue, newValue);
            }
            listeners[i].propertyChanged(e);
        }
    }
    
    /**
     * Notify all listeners of a property change event for an integer property.
     * @param source  the source of the event
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    protected void firePropertyChanged(
            Object source,
            String propertyName,
            int oldValue,
            int newValue
            ) {
        firePropertyChanged(source, propertyName, new Integer(oldValue), new Integer(newValue));
    }
    
    /**
     * Notify all listeners of a property change event for an boolean property.
     * @param source  the source of the event
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    protected void firePropertyChanged(
            Object source,
            String propertyName,
            boolean oldValue,
            boolean newValue
            ) {
        firePropertyChanged(source, propertyName, Boolean.valueOf(oldValue), Boolean.valueOf(newValue));
    }
    
    
    /**
     * Gets the charted dataset.
     *
     * @return The dataset
     */
    public XYDataset getDataset() {
        return dataset;
    }
    
    /**
     * Sets the charted dataset, and notifies listeners.
     * If the dataset doesn't contain exactly two series,
     * it also resets the dualChart, dualYAxis, and ratioChart properties.
     *
     * @param dataset
     */
    public void setDataset(XYDataset dataset) {
        XYDataset oldDataset = this.dataset;
        this.dataset = dataset;
        if (dataset != null && dataset.getSeriesCount() != 2) {
            dualChart = false;
            dualYAxis = false;
            ratioChart = false;
        }
        if (oldDataset == null || ! oldDataset.equals(dataset)) {
            firePropertyChanged(this, DATASET_PROPERTY, oldDataset, dataset);
        }
    }
    
    /**
     * Gets the chart's title.
     *
     * @return The title
     */
    public String getTitle() {
        return title;
    }
    
    /**
     * Sets the chart's title, and notifies listeners.
     *
     * @param title
     */
    public void setTitle(final String title) {
        String oldTitle = this.title;
        this.title = title;
        if (oldTitle == null || ! oldTitle.equals(title)) {
            firePropertyChanged(this, TITLE_PROPERTY, oldTitle, title);
        }
    }
    
    /**
     * Gets the chart's sub-title.
     *
     * @return The sub-title
     */
    public String getSubTitle() {
        return subTitle;
    }
    
    /**
     * Sets the chart's sub-title, and notifies listeners.
     *
     * @param subTitle
     */
    public void setSubTitle(final String subTitle) {
        String oldSubTitle = this.subTitle;
        this.subTitle = subTitle;
        if (oldSubTitle == null || ! oldSubTitle.equals(subTitle)) {
            firePropertyChanged(this,  SUBTITLE_PROPERTY, oldSubTitle, subTitle);
        }
    }
    
    /**
     * Gets the chart's footer text.
     *
     * @return The footer text
     */
    public String getFooterText() {
        return footerText;
    }
    
    /**
     * Sets the chart's footer text, and notifies listeners.
     *
     * @param footerText
     */
    public void setFooterText(final String footerText) {
        String oldFooterText = this.footerText;
        this.footerText = footerText;
        if (oldFooterText == null || ! oldFooterText.equals(footerText)) {
            firePropertyChanged(this, FOOTERTEXT_PROPERTY, oldFooterText, footerText);
        }
    }
    
    /**
     * Gets the chart's primary line width.
     *
     * @return The line width
     */
    public int getPrimaryLineWidth() {
        return primaryLineWidth;
    }
    
    /**
     * Sets the chart's primary line width (1-5), and notifies listeners.
     *
     * @param primaryLineWidth
     * @throws IllegalArgumentException if <code>primaryLineWidth</code>
     *      is less than 1 or greater than 5
     */
    public void setPrimaryLineWidth(final int primaryLineWidth) {
        if (primaryLineWidth < 1 || primaryLineWidth > 5) {
            throw new IllegalArgumentException();
        }
        int oldPrimaryLineWidth = this.primaryLineWidth;
        this.primaryLineWidth = primaryLineWidth;
        initPlotSeriesStroke();
        if (oldPrimaryLineWidth != primaryLineWidth) {
            firePropertyChanged(this, PRIMARYLINEWIDTH_PROPERTY, oldPrimaryLineWidth, primaryLineWidth);
        }
    }
    
    /**
     * Gets the chart's secondary line width.
     *
     * @return The line width
     */
    public int getSecondaryLineWidth() {
        return secondaryLineWidth;
    }
    
    /**
     * Sets the chart's secondary chart line width (1-5), and notifies listeners.
     *
     * @param secondaryLineWidth
     * @throws IllegalArgumentException if <code>secondaryLineWidth</code>
     *      is less than 1 or greater than 5
     */
    public void setSecondaryLineWidth(final int secondaryLineWidth) {
        if (secondaryLineWidth < 1 || secondaryLineWidth > 5) {
            throw new IllegalArgumentException();
        }
        int oldSecondaryLineWidth = this.secondaryLineWidth;
        this.secondaryLineWidth = secondaryLineWidth;
        if (oldSecondaryLineWidth != secondaryLineWidth) {
            firePropertyChanged(this, SECONDARYLINEWIDTH_PROPERTY, oldSecondaryLineWidth, secondaryLineWidth);
        }
    }
    
    /**
     * Gets the chart's moving average period (in seconds).
     *
     * @return The period
     */
    public int getMovingAveragePeriod() {
        return movingAveragePeriod;
    }
    
    /**
     * Sets the chart' moving average period (in seconds), and notifies listeners.
     *
     * @param movingAveragePeriod
     */
    public void setMovingAveragePeriod(final int movingAveragePeriod) {
        int oldMovingAveragePeriod = this.movingAveragePeriod;
        this.movingAveragePeriod = movingAveragePeriod;
        if (oldMovingAveragePeriod != movingAveragePeriod) {
            firePropertyChanged(this, MOVINGAVERAGEPERIOD_PROPERTY, oldMovingAveragePeriod, movingAveragePeriod);
        }
    }
    
    /**
     * Gets the chart's smoothing factor.
     *
     * @return The factor
     */
    public int getSmoothingFactor() {
        return smoothingFactor;
    }
    
    /**
     * Sets the smoothing factor (1-100), and notifies listeners.
     *
     * @param smoothingFactor
     * @throws IllegalArgumentException if <code>smoothingFactor</code>
     *      is less than 1 or greater than 100
     */
    public void setSmoothingFactor(final int smoothingFactor) {
        if (smoothingFactor < 1 || smoothingFactor > 100)
            throw new IllegalArgumentException();
        int oldSmoothingFactor = this.smoothingFactor;
        this.smoothingFactor = smoothingFactor;
        if (oldSmoothingFactor != smoothingFactor) {
            firePropertyChanged(this, SMOOTHINGFACTOR_PROPERTY, oldSmoothingFactor, smoothingFactor);
        }
    }
    
    /**
     * Gets the dual chart flag.
     *
     * @return The flag
     */
    public boolean isDualChart() {
        return dualChart;
    }
    
    /**
     * Sets the dual chart flag, and notifies listeners.
     *
     * @param dualChart
     */
    public void setDualChart(final boolean dualChart) {
        boolean oldDualChart = this.dualChart;
        this.dualChart = dualChart;
        if (oldDualChart != dualChart) {
            firePropertyChanged(this, DUALCHART_PROPERTY, oldDualChart, dualChart);
        }
    }
    
    /**
     * Gets the dual Y axis flag.
     *
     * @return The flag
     */
    public boolean isDualYAxis() {
        return dualYAxis;
    }
    
    /**
     * Sets the dual Y axis flag, and notifies listeners.
     *
     * @param dualYAxis
     */
    public void setDualYAxis(final boolean dualYAxis) {
        boolean oldDualYAxis = this.dualYAxis;
        this.dualYAxis = dualYAxis;
        if (oldDualYAxis != dualYAxis) {
            firePropertyChanged(this, DUALYAXIS_PROPERTY, oldDualYAxis, dualYAxis);
        }
    }
    
    /**
     * Gets the moving average flag.
     *
     * @return The flag
     */
    public boolean isMovingAverage() {
        return movingAverage;
    }
    
    /**
     * Sets the moving average flag, and notifies listeners.
     *
     * @param movingAverage
     */
    public void setMovingAverage(final boolean movingAverage) {
        boolean oldMovingAverage = this.movingAverage;
        this.movingAverage = movingAverage;
        if (oldMovingAverage != movingAverage) {
            firePropertyChanged(this, MOVINGAVERAGE_PROPERTY, oldMovingAverage, movingAverage);
        }
    }
    
    /**
     * Gets the smooth chart flag.
     *
     * @return The flag
     */
    public boolean isSmoothChart() {
        return smoothChart;
    }
    
    /**
     * Sets the smooth chart flag, and notifies listeners.
     *
     * @param smoothChart
     */
    public void setSmoothChart(final boolean smoothChart) {
        boolean oldSmoothChart = this.smoothChart;
        this.smoothChart = smoothChart;
        if (oldSmoothChart != smoothChart) {
            firePropertyChanged(this, SMOOTHCHART_PROPERTY, oldSmoothChart, smoothChart);
        }
    }
    
    /**
     * Gets the stacked lines flag.
     *
     * @return The flag
     */
    public boolean isStackedLines() {
        return stackedLines;
    }
    
    /**
     * Sets the stacked lines flag, and notifies listeners.
     *
     * @param stackedLines
     */
    public void setStackedLines(final boolean stackedLines) {
        boolean oldStackedLines = this.stackedLines;
        this.stackedLines = stackedLines;
        if (oldStackedLines != stackedLines) {
            firePropertyChanged(this, STACKEDLINES_PROPERTY, oldStackedLines, stackedLines);
        }
    }
    
    /**
     * Gets the ratio chart flag.
     *
     * @return  The flag
     */
    public boolean isRatioChart() {
        return ratioChart;
    }
    
    /**
     * Sets the ratio chart flag, and notifies listeners.
     *
     * @param ratioChart
     */
    public void setRatioChart(final boolean ratioChart) {
        boolean oldRatioChart = this.ratioChart;
        this.ratioChart = ratioChart;
        if (oldRatioChart != ratioChart) {
            firePropertyChanged(this, RATIOCHART_PROPERTY, oldRatioChart, ratioChart);
        }
    }
    
    /**
     * Gets the candle chart flag, and notifies listeners.
     *
     * @return  The flag
     */
    public boolean isCandleChart() {
        return candleChart;
    }
    
    /**
     * Sets the candle chart flag.
     *
     * @param candleChart
     */
    public void setCandleChart(final boolean candleChart) {
        boolean oldCandleChart = this.candleChart;
        this.candleChart = candleChart;
        if (oldCandleChart != candleChart) {
            firePropertyChanged(this, CANDLECHART_PROPERTY, oldCandleChart, candleChart);
        }
    }
    

    /**
     * Initialize the plotSeriesStroke element,
     * using the primaryLineWidth.
     */
    private void initPlotSeriesStroke() {
        plotSeriesStroke[0] = 
        new BasicStroke(primaryLineWidth, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_BEVEL);
    }
    
}
