/*
 * MyCollections.java
 */

package de.pdv_systeme.util;

import java.util.List;

/**
 * Supplements {@link java.util.Collections}
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class MyCollections {
    
    // Suppresses default constructor, ensuring non-instantiability.
    private MyCollections() {
    }
    
    /**
     * Selects the k-th element of a list.
     * @param list the list.
     * @param k the index.
     * @return the k-th element.
     * @throws ClassCastException if the list contains elements that are not
     *	       <i>mutually comparable</i> (for example, strings and integers).
     */
    public static Object select(List list, int k) {
        Object[] a = list.toArray();
        return select(a, k, 0, a.length-1);
    }
    
    /**
     * Returns the median of the list.
     * @param list the list.
     * @return the median.
     * @throws ClassCastException if the list contains elements that are not
     *	       <i>mutually comparable</i> (for example, strings and integers).
     */
    public static Object median(List list) {
        return select(list, 1+(list.size()>>1));
    }
    
    /**
     * Select the k-th element in a sub-array of objects
     * @param a the array.
     * @param k the selection index.
     * @param lwb the lower bound.
     * @param upb the upper bound.
     * @return the k-th element.
     * @throws ClassCastException if the array contains elements that are not
     * 	       <i>mutually comparable</i>.
     */
    private static Object select(Object[] a, int k, int lwb, int upb) {
        k += lwb-1;
        if (k < lwb || k > upb)
            throw new IndexOutOfBoundsException("Selection ("+String.valueOf(k)+
                    ") out of bounds ["+String.valueOf(lwb)+","+String.valueOf(upb)+"]");
        
        while (upb >= k && k >= lwb) {
            int i = lwb;
            int j = upb;
            Comparable temp = (Comparable) a[k];
            a[k] = a[lwb];
            a[lwb] = temp;
            while (i < j) {
                while (temp.compareTo(a[j]) < 0)
                    --j;
                a[i] = a[j];
                while (i < j && temp.compareTo(a[i]) >= 0)
                    ++i;
                a[j] = a[i];
            }
            a[i] = temp;
            if (k < i)
                upb = i-1;
            else
                lwb = i+1;
        }
        return a[k];
    }
}
