/*
 * SortedUniqueListModel.java
 */

package de.pdv_systeme.util;

import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractListModel;
import javax.swing.ListModel;

/**
 * This class builds on the {@link javax.swing.AbstractListModel} class
 * and the {@link java.util.Collection} interface
 * so that items are unique and sorted.
 *
 * @author Martin Vorlnder <mv@pdv-systeme.de>
 */
public class SortedUniqueListModel
        extends AbstractListModel
        implements ListModel, Collection, Cloneable {
    
    /** Storage for the list model items. */
    protected ArrayList delegate;
    
    /** The comparator used to sort the items. */
    protected Comparator comparator;
    
    /** Create a new instance of SortedUniqueListModel. */
    public SortedUniqueListModel() {
        this.delegate = new ArrayList();
        this.comparator = Collator.getInstance();
    }
    
    /**
     * Create a new instance of SortedUniqueListModel.
     * @param comparator the comparator to use
     */
    public SortedUniqueListModel(final Comparator comparator) {
        this.delegate = new ArrayList();
        this.comparator = comparator;
    }
    
    /**
     * Create a new instance of SortedUniqueListModel.
     * @param elements The elements to populate the list model with
     */
    public SortedUniqueListModel(final Collection elements) {
        this.delegate = new ArrayList(elements);
        this.comparator = Collator.getInstance();
    }
    
    /**
     * Create a new instance of SortedUniqueListModel.
     * @param elements The elements to populate the list model with
     * @param comparator the comparator to use
     */
    protected SortedUniqueListModel(final Collection elements, final Comparator comparator) {
        this.delegate = new ArrayList(elements);
        this.comparator = comparator;
    }
    
    /**
     * Set the comparator.
     * @param comparator the new comparator
     */
    public void setComparator(final Comparator comparator) {
        this.comparator = comparator;
        Collections.sort(delegate, comparator);
        fireContentsChanged(this, 0, delegate.size()-1);
    }
    
    /**
     * Get the comparator.
     * @return the comparator
     */
    public Comparator getComparator() {
        return this.comparator;
    }
    
    /*
     * Implementation of the ListModel interface
     */
    /** {@inheritDoc} */
    public int getSize() {
        return delegate.size();
    }
    
    /** {@inheritDoc} */
    public Object getElementAt(final int index) {
        return delegate.get(index);
    }
    
    /*
     * Implementation of the Collection interface
     * (including ListDataListener notification).
     */
    /** {@inheritDoc} */
    public Iterator iterator() {
        return delegate.iterator();
    }
    
    /** {@inheritDoc} */
    public boolean add(final Object o) {
        return insert((Comparable) o); // notifies the listeners
    }
    
    /** {@inheritDoc} */
    public boolean addAll(final Collection c) {
        int oldSize = delegate.size();
        boolean hasChanged = false;
        for (Iterator ci = c.iterator(); ci.hasNext(); )
            if (insert((Comparable) ci.next(), false))
                hasChanged = true;
        if (hasChanged)
            fireContentsChanged(this, 0,  oldSize-1);
        return hasChanged;
    }
    
    /** {@inheritDoc} */
    public void clear() {
        int oldSize = delegate.size();
        delegate.clear();
        fireIntervalRemoved(this, 0, oldSize-1);
    }
    
    /** {@inheritDoc} */
    public boolean remove(final Object o) {
        int index = delegate.indexOf(o);
        boolean hasChanged = delegate.remove(o);
        if (hasChanged)
            fireIntervalRemoved(this, index, index);
        return hasChanged;
    }
    
    /** {@inheritDoc} */
    public boolean removeAll(final Collection c) {
        int oldSize = delegate.size();
        boolean hasChanged = delegate.removeAll(c);
        if (hasChanged)
            fireContentsChanged(this, 0, oldSize-1);
        return hasChanged;
    }
    
    /** {@inheritDoc} */
    public boolean retainAll(final Collection c) {
        int oldSize = delegate.size();
        boolean hasChanged = delegate.retainAll(c);
        fireContentsChanged(this, 0, oldSize-1);
        return hasChanged;
    }
    
    /** {@inheritDoc} */
    public boolean contains(final Object o) {
        return delegate.contains(o);
    }
    
    /** {@inheritDoc} */
    public boolean containsAll(final Collection c) {
        return delegate.containsAll(c);
    }
    
    /** {@inheritDoc} */
    public boolean isEmpty() {
        return delegate.size() == 0;
    }
    
    /** {@inheritDoc} */
    public int size() {
        return delegate.size();
    }
    
    /** {@inheritDoc} */
    public Object[] toArray() {
        return delegate.toArray();
    }
    
    /** {@inheritDoc} */
    public Object[] toArray(Object[] a) {
        return delegate.toArray(a);
    }
    
    
    /**
     * Returns the index in this list of the first occurrence of the specified
     * element, or -1 if this list does not contain this element.
     * More formally, returns the lowest index <tt>i</tt> such that
     * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt>,
     * or -1 if there is no such index.
     *
     * @param o element to search for.
     * @return the index in this list of the first occurrence of the specified
     * 	       element, or -1 if this list does not contain this element.
     */
    public int indexOf(final Object o) {
        return delegate.indexOf(o);
    }
    
    
    /**
     * Insert a comparable object.
     * @param o  the object to be inserted
     * @return <tt>true</tt> if this SortedUniqueListModel changed as a result of the call.
     */
    protected boolean insert(final Comparable o) {
        return insert(o, true);
    }
    
    /**
     * Insert a comparable object.
     * @param o  the object to be inserted
     * @param notify  if the listeners are to be notified
     * @return <tt>true</tt> if this SortedUniqueListModel changed as a result of the call.
     */
    protected boolean insert(final Comparable o, final boolean notify) {
        if (delegate.isEmpty()
        || comparator.compare(o, delegate.get(delegate.size()-1)) > 0) {
            delegate.add(o);
            if (notify)
                fireIntervalAdded(this, 0, 0);
            return true;
        } else {
            int index = Collections.binarySearch(delegate, o, comparator);
            if (index < 0) {
                index = -index-1;
                delegate.add(index, o);
                if (notify)
                    fireIntervalAdded(this, index, index);
                return true;
            }
        }
        return false;
        /*
         * Old implementation (when the class was based on DefaultListModel,
         * which isn't a Collection):
         
        if (isEmpty()) {
            super.add(o);
            fireIntervalAdded(this, 0, 0);
            return true;
        }
         
        // Do a binary search for the insertion point
        int lo = 0;
        int hi = size() - 1;
        while (lo <= hi) {
            int med = (lo + hi) >> 1;
            int cmp = ((Comparable) get(med)).compareTo(o);
            if (cmp < 0)
                lo = med + 1;
            else if (cmp > 0)
                hi = med - 1;
            else
                return false; // o already in list;
        }
        super.insertElementAt(o, lo);
        fireIntervalAdded(this, lo, lo);
        return true;
         */
    }
    
}
