/*
 * VmsDate.java
 */

package de.pdv_systeme.util;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Utility routines for handling VMS dates
 *
 * @author <a href="mailto:mv@pdv-systeme.de">Martin Vorlnder</a>
 */
public class VmsDate {
    
    /** Output precision specifier for {@link #toString(Object,int) toString} */
    public static final int
            PRECISION_DATE = 1,
            PRECISION_MINUTE = 2,
            PRECISION_SECOND = 3,
            PRECISION_CENTISECOND = 4,
            PRECISION_MILLISECOND = 5;
    
    /** Date string for VMS timestamps */
    public static final String DATE_STRING = "dd-MMM-yyyy";
    
    /** Time string for VMS timestamps (to the minute) */
    public static final String TIME_MINUTES_STRING = "HH:mm";
    
    /** Time string for VMS timestamps (to the second) */
    public static final String TIME_SECONDS_STRING = "HH:mm:ss";
    
    /** Time string for VMS timestamps (to the millisecond) */
    public static final String TIME_MILLIS_STRING = "HH:mm:ss.SSS";
    
    /** Date format for VMS timestamps (date only, no time) */
    public static final SimpleDateFormat dateFormatDate
            = new SimpleDateFormat(DATE_STRING, Locale.US);
    
    /** Date format for VMS timestamps (to the minute) */
    public static final SimpleDateFormat dateFormatMinutes
            = new SimpleDateFormat(DATE_STRING+" "+TIME_MINUTES_STRING, Locale.US);
    
    /** Date format for VMS timestamps (to the second) */
    public static final SimpleDateFormat dateFormatSeconds
            = new SimpleDateFormat(DATE_STRING+" "+TIME_SECONDS_STRING, Locale.US);
    
    /** Date format for VMS timestamps (to the millisecond) */
    public static final SimpleDateFormat dateFormatMillis
            = new SimpleDateFormat(DATE_STRING+" "+TIME_MILLIS_STRING, Locale.US);
    
    /** Pattern string for a VMS date's month field */
    public static final String MONTH_PATTERN = "JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC";
    
    /** Pattern string for a VMS date */
    public static final String DATE_PATTERN =
            "^"+                        // Beginning of string
            "( ?\\d|\\d{2})"+           // day: 1 digit with optional leading whitespace or 2 digits
            "-"+                        // separator '-'
            "("+MONTH_PATTERN+")"+      // month
            "-"+                        // separator '-'
            "(\\d{4})"+                 // year: 4 digits
            "(?:"+              // optionally (1):
            "(?:\\s+|:|-)"+             // separator multiple whitespace, ':', or -'
            "(\\d{2})"+                 // hour: 2 digits
            "(?:"+              // optionally (2):
            ":" +                       // separator ':'
            "(\\d{2})"+                 // minute: 2 digits
            "(?:"+              // optionally (3):
            ":"+                        // separator ':'
            "(\\d{2})"+                 // second: 2 digits
            "(?:"+              // optionally (4):
            "\\."+                      // separator '.'
            "(\\d{2})"+                 // centiseconds: 2 digits
            ")?"+               // end option 4
            ")?"+               // end option 3
            ")?"+               // end option 2
            ")?"+               // end option 1
            "$";                        // End of string
    
    /** Pattern for a VMS date */
    public static final Pattern datePattern
            = Pattern.compile(DATE_PATTERN, Pattern.CASE_INSENSITIVE);

    /** @see #convertToMillis */
    private static Calendar workingCalendar = null;
    
    /** Class is not to be instantiated */
    private VmsDate() {}
    
    /**
     * Convert time to milliseconds (in the default timezone)
     * @param year
     * @param month
     * @param day
     * @param hour
     * @param minute
     * @param second
     * @param millisecond
     * @return the time in milliseconds
     */
    private static synchronized long convertToMillis(
            final int year, final int month, final int day,
            final int hour, final int minute, final int second, final int millisecond) {
        if (workingCalendar == null) {
            workingCalendar = Calendar.getInstance();
        }
        workingCalendar.clear();
        workingCalendar.set(year, month, day, hour, minute, second);
        return workingCalendar.getTimeInMillis()+millisecond;
    }
    
    /**
     * Parse a VMS date string.
     * @param s  the string representing the VMS date or time
     * @param date  if <code>s</code> only represents the time part, this supplies the date
     * @return the date
     * @throws ParseException if the string does not represent a valid VMS date
     */
    public static Date parse(final String s, final String date)
    throws ParseException {
        Date d = null;
        try {
            d = parse(s);
        } catch (ParseException e1) {
            d = parse(date+" "+s);
        }
        return d;
    }
    
    /**
     * Parse a VMS date string.
     * @param s  the string representing the VMS date
     * @return the date
     * @throws ParseException if the string does not represent a valid VMS date
     */
    public static Date parse(final String s)
    throws ParseException {
        Matcher m = datePattern.matcher(s);
        if (! m.matches()) {
            throw new ParseException(s, 0);
        }
        int day, month, year, hour, minute, second, centi;
        try {
            day    = Integer.parseInt(m.group(1).trim());
            month  = MONTH_PATTERN.indexOf(m.group(2).toUpperCase()) / 4; // zero based!
            year   = Integer.parseInt(m.group(3));
            hour   = (m.group(4) != null) ? Integer.parseInt(m.group(4)) : 0;
            minute = (m.group(5) != null) ? Integer.parseInt(m.group(5)) : 0;
            second = (m.group(6) != null) ? Integer.parseInt(m.group(6)) : 0;
            centi  = (m.group(7) != null) ? Integer.parseInt(m.group(7)) : 0;
        } catch (NumberFormatException e) {
            // As we parsed 's' to be a valid VMS timestamp, this shouldn't happpen
            throw new ParseException(s, 0);
        }
        return new Date(convertToMillis(year, month, day, hour, minute, second, centi*10));
    }
    
    /**
     * Returns a String object representing this <code>VmsDate</code>'s value.
     * @param date the VmsDate object
     * @return the <code>date</code>'s string representation
     * @throws ClassCastException if the <code>date</code> isn't a <code>Date</code> object
     */
    public static String toString(final Object date) {
        return toString(date, PRECISION_CENTISECOND);
    }
    
    /**
     * Returns a String object representing this <code>VmsDate</code>'s value.
     * @param date the VmsDate object
     * @param specifier specifies the requested output precision (one of PRECISION_*)
     * @return the <code>date</code>'s string representation
     * @throws ClassCastException if the <code>date</code> isn't a <code>Date</code> object
     * @throws IllegalArgumentException if the <code>specifier</code> is out of range
     */
    public static String toString(final Object date, final int specifier) {
        switch (specifier) {
            case PRECISION_DATE :
                return dateFormatDate.format(date);
            case PRECISION_MINUTE :
                return dateFormatMinutes.format(date);
            case PRECISION_SECOND :
                return dateFormatSeconds.format(date);
            case PRECISION_CENTISECOND :
                String ds = dateFormatMillis.format(date);
                return ds.substring(0, ds.length()-1); // cut off millisecond
            case PRECISION_MILLISECOND :
                return dateFormatMillis.format(date);
            default :
                throw new IllegalArgumentException("Invalid precision "+specifier);
        }
    }
    
}
